/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DATA_H
#define DATA_H

#include <QObject>
#include <QList>
#include <QLoggingCategory>
#include <QPair>
#include <types/message.h>
#include <types/dialog.h>
#include <types/user.h>
#include <types/chat.h>
#include <types/decryptedmessage.h>
#include <secret/secretchatmessage.h>
#include "models/dialogitem.h"
#include "models/contactitem.h"
#include "models/messageitem.h"
#include "models/messagesview.h"
#include "models/dialogsview.h"
#include "models/contactsview.h"
#include "incomingfile.h"
#include "dbmanager.h"

Q_DECLARE_LOGGING_CATEGORY(TG_PLUGIN_PROFILING)

class Data : public QObject
{
    Q_OBJECT
public:
    explicit Data(QObject *parent = 0);
    ~Data();

    DbManager &dbManager();

    // updates
    bool insertOrUpdateState(const UpdatesState &state);
    bool insertState(const UpdatesState &state);
    bool updateState(const UpdatesState &state);
    UpdatesState getState();

    // dialog operations
    bool existsDialog(qint32 dialogId);
    bool insertDialogs(const QList<Dialog> &dialogs, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users);
    bool insertOrUpdateDialog(const Dialog &dialog);
    bool insertDialog(const Dialog &dialog);
    bool updateDialog(const Dialog &dialog);
    bool deleteDialog(qint32 dialogId);
    bool dropMessagesView(qint32 dialogId);
    DialogItem getDialog(qint32 dialogId);
    qint32 getDialogIdFromMessageId(qint32 msgId);
    bool dialogIsEmpty(qint32 dialogId);
    QList<DialogItem> getDialogs();
    bool decreaseUnreadCount(qint32 dialogId, qint32 count);
    bool increaseUnreadCount(qint32 dialogId, qint32 count);
    qint32 getUnreadCount(qint32 dialogId);

    // user operations
    bool existsUser(qint32 userId);
    bool insertOrUpdateUser(const User &user);
    bool insertUser(const User &user);
    bool updateUser(const User &user);
    bool updateUserName(qint32 userId, const QString &firstName, const QString &lastName);
    bool updateUserIsContact(qint32 userId, bool isContact, bool hasPhone);
    qint32 insertOrUpdateUserThumbnailFileLocation(const User &user);
    qint32 insertOrUpdateUserBigPhotoFileLocation(const User &user);
    qint32 insertOrUpdateChatThumbnailFileLocation(const Chat &chat);
    qint32 insertOrUpdateChatBigPhotoFileLocation(const Chat &chat);
    qint32 insertOrUpdateFileLocation(qint32 id, const FileLocation &serverLocation, bool isUser, bool isThumbnail);
    bool deleteUsers();
    bool hasChangedUsername(const User &user);
    ContactItem getUser(qint32 id);
    QList<ContactItem> getUsers();
    QList<qint32> getUserSecretChats(qint32 userId);

    // contacts operations
    bool isContact(qint32 userId);
    bool updateContactsLink(const ContactsLink &link);
    bool updateContactsLink(qint32 userId, const ContactsMyLink &myLink, const ContactsForeignLink &foreignLink);

    // chat operations
    bool existsChat(qint32 chatId);
    bool insertOrUpdateChat(const Chat &chat);
    bool insertChat(const Chat &chat);
    bool updateChat(const Chat &chat);
    bool updateChatTitle(qint32 chatId, const QString &newTitle);
    void updateChatParticipants(const ChatParticipants &participants);
    bool deleteChat(qint32 chatId);
    bool deleteChatPhoto(qint32 chatId);
    DialogItem getChat(qint32 id);

    // secret chat operations
    bool insertSecretChat(qint32 chatId, qint32 date, qint32 peerId, qint32 state);
    bool updateSecretChat(qint32 chatId, qint32 state);
    bool insertOrUpdateSecretChat(qint32 chatId, qint32 date, qint32 peerId, qint32 state);
    DialogItem getSecretChat(qint32 id);
    bool insertSecretChatMessage(const SecretChatMessage &secretChatMessage);
    bool existsSecretChatMessage(const SecretChatMessage &SecretChatMessage);
    bool insertDecryptedMessage(qint32 chatId, const DecryptedMessage &message, qint32 date, bool outgoing, qint64 mediaId = 0, qint64 requestId = 0);
    bool appendOutgoingDecryptedMessage(qint64 requestId, qint32 chatId, const DecryptedMessage &message, const QString &mediaLocalPath = "", const QString &thumbnailLocalPath = "");
    QList<qint32> selectDecryptedMessagesToMarkAsRead(qint32 chatId, qint32 maxDate, bool outgoing);
    qint32 getDecryptedMessagesMaxDate(qint32 chatId, bool outgoing = false);
    void deleteSecretChatHistory(qint32 dialogId, bool deleteChat = false);
    bool markSecretChatAsRead(qint32 chatId);
    qint32 insertAttachment(qint32 messageId, const EncryptedFile &attachment, const DecryptedMessageMedia &media);
    bool insertLocalMedia(qint64 mediaId, MessageMedia::MessageMediaType mediaType, const QString &mediaLocalPath, const QString &thumbnailLocalPath);

    // message operations
    bool insertOrUpdateMessage(const Message &message);
    bool insertMessage(const Message &message, qint64 randomId = 0, bool sent = true, const QString &mediaLocalPath = "", const QString &thumbnailLocalPath = "");
    bool appendOutgoingMessage(qint32 recipientId, qint64 randomId, const Message &message, const QString &mediaLocalPath = "", qint64 uploadFileId = 0, const QString &thumbnailLocalPath = "");
    bool updateMessage(const Message &message);
    bool updateMessage(qint32 msgId, const Message &message);
    bool updateMessage(qint32 oldMessageId, qint32 newMessageId, qint32 date);
    bool updateMessageId(qint64 randomId, qint32 id);
    bool deleteMessages(qint32 dialogId);
    bool deleteMessages(const QList<qint32> msgIds);
    bool isMessageIncomingAndUnread(qint32 msgId);
    bool markChatAsRead(qint32 chatId);
    bool markMessagesAsRead(const QList<qint32> &messages);
    qint32 getMessageId(qint64 randomId);
    MessageItem getMessage(qint32 messageId);
    MessageItem topMessage(qint32 dialogId);

    QString getThumbnail(qint32 id, bool isUser);
    qint32 insertOrUpdateUserThumbnail(qint32 id, const QString &localPath, const FileLocation &serverLocation);
    qint32 insertOrUpdateUserBigPhoto(qint32 id, const QString &localPath, const FileLocation &serverLocation);
    qint32 insertOrUpdateChatThumbnail(qint32 id, const QString &localPath, const FileLocation &serverLocation);
    qint32 insertOrUpdateChatBigPhoto(qint32 id, const QString &localPath, const FileLocation &serverLocation);

    qint32 existsPhoto(qint64 photoId);
    qint32 insertPhoto(const Photo &photo);
    qint32 updatePhoto(qint32 id, qint32 oldPhotoId, const Photo &photo);
    qint32 deletePhoto(qint32 photoId);
    bool hasThumbnailDownloaded(qint32 photoId);
    void downloadThumbnail(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType);
    void downloadPhoto(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType);
    bool downloadPhoto(qint32 messageId);
    bool downloadUserBigPhoto(qint32 userId);
    bool downloadChatBigPhoto(qint32 chatId);

    PhotoSize getPhotoSize(qint32 id);
    qint32 getPhotoMsgId(qint32 id);

    qint32 existsVideo(qint64 videoId);
    qint32 insertVideo(const Video &video);
    qint32 updateVideo(qint64 oldVideoId, const Video &video);
    qint32 updateVideo(qint64 videoId, const QString &localPath);
    QString getVideoLocalPath(qint64 videoId);
    bool downloadVideo(qint32 messageId);
    qint32 insertOrUpdatePhotoSize(qint64 oldMediaId, qint64 newMediaId, const PhotoSize &photoSize);
    FileLocation getFileLocation(qint32 fileLocationId);
    bool deleteFileLocation(qint32 fileLocationId);
    qint32 insertFileLocation(const QString &localPath);
    qint32 insertFileLocation(const FileLocation &fileLocation);
    qint32 updateFileLocation(qint32 fileLocationId, const QString &localPath);
    qint32 updateFileLocation(qint32 fileLocationId, const FileLocation &fileLocation);
    qint32 insertPhotoSize(qint64 photoId, const PhotoSize &photoSize, qint32 fileLocationId);
    qint32 updatePhotoSize(qint64 oldPhotoId, qint64 newPhotoId, const PhotoSize &photoSize);
    qint32 updatePhotoSizeLocation(qint64 photoId, const QString &type, const qint32 fileLocationId);
    qint32 getPhotoSizeFileLocationId(qint64 photoId, const QString &type);

    qint32 insertOrUpdateMessageAction(const Message &message);
    qint32 insertMessageAction(const Message &message);
    qint32 updateMessageAction(const Message &message);
    bool executeAction(const Message &message);

    qint64 getDownloadingFileId(qint32 id);

    void setMutingInProgress(qint64 requestId, qint32 peerId, bool muting);
    QMap<qint64, qint32> mChatsToClearHistory;
    QMap<qint32, qint32> mChatsToDelete;

public:

Q_SIGNALS:

    void unreadCountChanged(qint32 unreadCount);

public Q_SLOTS:
    void setSecretChatMessageTTL(qint32 chatId, qint32 ttlSeconds);

protected:
    void updateChats(const QList<Chat> &chats);
    void updateUsers(const QList<User> &users);
    void updateChatPhoto(qint32 chatId, const Photo &photo);
    void updateChatFull(const ChatFull &chatFull);

Q_SIGNALS:
    void userInfo(const ContactItem &user);

    void contactsAdded(const QList<ContactItem> &contacts);
    void dialogsAdded(const QList<DialogItem> &dialogs, qint32 totalCount = -1);
    void chatUpdated(const qint32 chatId);
    void contactThumbnailUpdated(qint32 userId, const QString &thumbnail);
    void contactBigPhotoUpdated(qint32 userId, const QString &bigPhoto);
    void chatBigPhotoUpdated(qint32 chatId, const QString &bigPhoto);
    void dialogThumbnailUpdated(qint32 dialogId, const QString &thumbnail);
    void contactNameUpdated(qint32 userId, const QString &firstName, const QString lastName);
    void dialogTitleUpdated(qint32 dialogId, const QString &title);
    void dialogMuted(qint32 dialogId, bool muted);
    void dialogsMuted(NotifyPeer::NotifyPeerType notifyType, bool muted);

    void messagesAdded(const QList<MessageItem> &messages, qint32 totalCount);
    void messageAdded(const MessageItem &message);
    void outgoingMessageAdded(qint32 recipientId, const MessageItem &message);
    void messageUpdated(qint32 msgId, const MessageItem &message);

    void chatMarkedAsRead(qint32 chatId);
    void chatHistoryDeleted(qint32 chatId);
    void chatDeleted(qint32 chatId);
    void messagesMarkedAsRead(const QList<qint32> &msgIds);
    void messagesDeleted(const QList<qint32> &msgIds);

    void messagePhotoUpdated(qint32 msgId, const QString &localPath, bool isThumbnail);
    void messageVideoUpdated(qint32 msgId, const QString &videoLocalPath);

    void downloadingUpdated(qint32 id, bool downloading);
    void downloadedPercentageUpdated(qint32 id, qint8 downloadedPercentage);

    void updatesTooLong(qint32 pts, qint32 date, qint32 qts);

    void secretChatRequested(qint32 chatId, qint32 date, qint32 creatorId, qint64 accessHash);
    void secretChatCreated(qint32 chatId);
    void secretChatDiscarded(qint32 chatId);
    void secretChatUpdated(qint32 chatId);

    void userTyping(qint32 dialogId, const QString &whoIsTyping = QString());

private:
    // correlates file id with incoming file object with the rest of needed data to download the file.
    QMap<qint64, IncomingFile *> mFiles;
    // correlates id (messageId, userId, etc..) with fileId while downloading a file
    QMap<qint32, qint64> mActiveDownloads;
    // reference to service
    QObject *mServiceRef;
    DbManager mDbManager;

    qint32 mLastUnreadCount;

    // time mesurement map for file downloading. Correlates fileId with a QTime object for calculate total download time
    QMap<qint64, QTime *> mFileTimes;
    qint32 mFilesAvgTime;
    qint32 mFilesCount;
    qint32 mFilesTotalTime;

    // upload fileId -> messageId correlation to emit upload progress
    QMap<qint64, qint32> mActiveUploads;

    // map to hold if an updatePhoto action must be executed when photo is downloaded. msgId -> bool
    QMap<qint32, bool> mMessageIdsToUpdatePhoto;

    // maps to hold the size priorities for downloading one file size or another
    QMap<QString,qint8> mPhotoSmallPriorities;
    QMap<QString,qint8> mPhotoBigPriorities;

    // maps for mute/unmute. Correlates requestId with peerId
    QMap<qint64, qint32> mMuteInProgress;
    QMap<qint64, qint32> mUnmuteInProgress;

    QPair<QString, FileLocation> getFullFileLocation(qint32 id, bool isUser, bool isThumbnail);
    QString getThumbnailLocalPath(qint64 photoId);
    QString getPhotoLocalPath(qint64 photoId, bool isThumbnail = false);
    QList<qint32> getMediaMessages(qint64 mediaId);
    qint32 insertOrUpdateProfilePhoto(qint32 id, bool isUser, bool isThumbnail, const QString &localPath, const FileLocation &serverLocation);

    // returns true if queried server to get the photo file
    bool downloadProfilePhoto(qint32 id, bool isUser, bool isThumbnail);
    bool getProfilePhotoFileFromServer(const FileLocation &serverLocation, qint32 id, bool isUser, bool isThumbnail);
    bool downloadPhotoFileFromServer(const FileLocation &serverLocation, qint32 id, bool isUser, bool isThumbnail);
    void selectSizeAndDownloadPhoto(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType, bool thumbnail = true);
    void downloadPhotoSize(qint32 id, qint64 mediaId, qint32 date, IncomingFile::IncomingFileType incomingFileType,
                           bool thumbnail, const PhotoSize &photoSize, qint32 fileLocationId);
    void downloadPhotoFile(qint32 id, qint64 mediaId, qint32 date, IncomingFile::IncomingFileType incomingFileType, bool thumbnail,
                           const QString &type, qint32 size, qint32 dcId, qint32 localId, qint64 secret, qint64 volumeId, qint32 fileLocationId);
    void downloadPhotoAttachment(qint32 messageId, qint64 mediaId, qint32 date, bool thumbnail, const QString &type,
                                 qint32 size, qint32 dcId, qint64 accessHash, const QByteArray &key, const QByteArray &iv, qint32 fileLocationId);
    /** Download directly only one image using priorities to know what size
         Priorities for thumbnail sizes are m, b, s, a, x, c, y, d, w
         Priorities for big photo sizes are w, d, y, x, c, m, b, a, s

        Type 	Image filter 	Size limits, px
        s           box             100x100
        m           box             320x320
        x           box             800x800
        y           box             1280x1280
        w           box             2560x2560
        a           crop            160x160
        b           crop            320x320
        c           crop            640x640
        d           crop            1280x1280

        source: https://core.telegram.org/constructor/photoSize#thumbnail-type-and-its-sizes
    */
    PhotoSize selectThumbnailSize(const QList<PhotoSize> &photoSizes);
    PhotoSize selectBigPhotoSize(const QList<PhotoSize> &photoSizes);
    PhotoSize selectSize(const QList<PhotoSize> &photoSizes, bool thumbnail);

    // process notifications to update db and models
    void updateChatParticipantsFromUpdate(const Update &update);
    bool processUpdate(const Update &update, const QList<User> &users = QList<User>(), const QList<Chat> &chats =  QList<Chat>());
    void processUpdateUserName(const Update &update);
    void processUpdateUserPhoto(const Update &update);
    bool needUpdateDifference(qint32 seq, UpdatesState clientState);

    void notifySecretChatsThumbnailUpdated(qint32 userId, const QString &filePath);
    void notifySecretChatsTitleUpdated(qint32 userId, const QString &title);

    // settings
    bool muteDialog(qint32 dialogId, bool muted);
    bool muteDialogs(NotifyPeer::NotifyPeerType peerType, bool muted);

    void surfaceUnreadCount();

    bool exists(qint32 id, bool isUser);

private Q_SLOTS:    
    void onUpdatesGetStateAnswer(qint64 id, qint32 pts, qint32 qts, qint32 date, qint32 seq, qint32 unreadCount);
    void onUpdatesGetDifferenceAnswer(qint64 id, qint32 date, qint32 seq);
    void onUpdatesGetDifferenceAnswer(qint64 id, const QList<Message> &messages, const QList<SecretChatMessage> &secretChatMessages, const QList<Update> &otherUpdates, const QList<Chat> &chats, const QList<User> &users, const UpdatesState &state, bool isIntermediateState);

    void onAccountUpdateNotifySettingsAnswer(qint64 id, bool ok);
    void onMessagesGetDialogsAnswer(qint64 requestId, qint32 totalCount, const QList<Dialog> &dialogs, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users);
    void onUploadGetFileAnswer(qint64 fileId, const StorageFileType &type, qint32 mtime, QByteArray bytes, qint32 partId, qint32 downloaded, qint32 total);
    void onUploadCancelFileAnswer(qint64 fileId, bool cancelled);
    void onUploadSendFileAnswer(qint64 fileId, qint32 partId, qint32 downloaded, qint32 total);
    void onUsersGetUsersAnswer(qint64 id, QList<User> users);
    void onMessagesGetFullChatAnswer(qint64 id, ChatFull chatFull, QList<Chat> chats, QList<User> users);
    void onContactsImportContactsAnswer(qint64 id, QList<ImportedContact> importedContacts, QList<qint64> retryContacts, QList<User> users);
    void onContactsGetContactsAnswer(qint64 requestId, bool modified, const QList<Contact> &contacts, const QList<User> &users);
    void onMessagesGetHistoryAnswer(qint64 requestId, qint32 totalCount, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users);
    void onMessagesReadHistoryAnswer(qint64 id, qint32 pts, qint32 seq, qint32 offset);
    void onMessagesDeleteHistoryAnswer(qint64 id, qint32 pts, qint32 seq, qint32 offset);
    void onMessagesSendMessageAnswer(qint64 requestId, qint32 msgId, qint32 date, qint32 pts, qint32 seq, const QList<ContactsLink> &links);
    void onMessagesDeleteMessagesAnswer(qint64, const QList<qint32> msgIds);
    void onMessagesSendPhotoAnswer(qint64 id, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq);
    void onMessagesSendVideoAnswer(qint64 id, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq);
    void onMessagesCreateChatAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq);
    void onMessagesAddChatUserAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq);
    void onMessagesDeleteChatUserAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq);
    void onUpdateShortMessage(qint32 id, qint32 fromId, const QString &message, qint32 pts, qint32 date, qint32 seq);
    void onUpdateShortChatMessage(qint32 id, qint32 fromId, qint32 chatId, const QString &message, qint32 pts, qint32 date, qint32 seq);
    void onUpdateShort(const Update &update, qint32 date);
    void onUpdatesCombined(const QList<Update> &updates, const QList<User> &users, const QList<Chat> &chats, qint32 date, qint32 seqStart, qint32 seq);
    void onUpdates(const QList<Update> &udts, const QList<User> &users, const QList<Chat> &chats, qint32 date, qint32 seq);
    void onUpdatesTooLong();
    void onUpdateSecretChatMessage(const SecretChatMessage &secretChatMessage, qint32 qts);

    void onMessagesCreateEncryptedChatAnswer(qint32 chatId, qint32 date, qint32 peerId, qint64 accessHash);
    void onMessagesEncryptedChatRequested(qint32 chatId, qint32 date, qint32 peerId, qint64 accessHash);
    void onMessagesEncryptedChatCreated(qint32 chatId, qint32 date, qint32 peerId, qint64 accessHash);
    void onMessagesEncryptedChatDiscarded(qint32 chatId);
    void onMessagesSendEncryptedAnswer(qint64 id, qint32 date, EncryptedFile encryptedFile);
    void onMessagesSendEncryptedFileAnswer(qint64, qint32, EncryptedFile);

    void processSecretChatServiceMessage(const SecretChatMessage &secretChatMessage);
};

#endif // DATA_H
