/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "dbmanager.h"
#include "dbschema.h"

#include <QDir>
#include <QFile>
#include <QtSql/QSqlDriver>
#include <QtSql/QSqlError>
#include <QtSql/QSqlQuery>

Q_LOGGING_CATEGORY(LOG_DATABASE, "tg.plugin.db")

const QString DATABASE_NAME = QString("telegram.sqlite");

// Database version indentifiers go here.
const qint32 VERSION_INITIAL = 1; // Must start with 1!
const qint32 VERSION_MESSAGE_ACTIONS_REMOVE_PHOTO_ID = 2;
const qint32 VERSION_MESSAGES_RANDOMID = 3;
const qint32 VERSION_GROUP_PROFILE = 4;
const qint32 VERSION_DIALOGS_MUTED = 5;
const qint32 VERSION_DIALOGS_DATE = 6;
const qint32 VERSION_SECRET_CHATS = 7;
const qint32 VERSION_READ_HISTORY = 8;
const qint32 VERSION_ENCRYPTED_MEDIA_KEYS = 9;
const qint32 VERSION_ENCRYPTED_TTL = 10;
const qint32 VERSION_PROFILE_BIG_PHOTO = 11;

// Update this line when creating new upgrade version.
const qint32 DATABASE_VERSION = VERSION_PROFILE_BIG_PHOTO;


DbManager::DbManager(QObject *parent) :
    QObject(parent),
    mTransactionsDepth(0) {
}

DbManager::~DbManager() {
    close();
}

bool DbManager::initialize(const QString &path) {
    QDir dir(path);
    if (!dir.exists()) {
        qCCritical(LOG_DATABASE) << "Database path does not exist: " << path;
        return false;
    }

    mDatabasePath = dir.absoluteFilePath(DATABASE_NAME);
    // Avoid re-creating default connection.
    if (!mDatabase.contains()) {
        mDatabase = QSqlDatabase::addDatabase("QSQLITE");
    }
    mDatabase.setDatabaseName(mDatabasePath);

    if (!createOrUpgrade(DATABASE_VERSION)) {
        return false;
    }

    Q_EMIT initialized();
    return true;
}

bool DbManager::createOrUpgrade(qint32 version) {
    bool databaseExists = QFile(mDatabasePath).exists();
    if (!mDatabase.open()) {
        qCCritical(LOG_DATABASE) << "Failed to open database!";
        return false;
    }

    QSqlQuery query(mDatabase);
    query.exec("PRAGMA foreign_keys = true");
    query.exec("PRAGMA temp_store = MEMORY");

    qint32 oldVersion = getDatabaseVersion();
    if (!databaseExists || oldVersion == 0) {
        if (oldVersion == 0) purge(); // For early adopters.
        qCDebug(LOG_DATABASE) << "Creating database";
        create();
    } else if (oldVersion == -1 || oldVersion > version) {
        qCWarning(LOG_DATABASE) << "Wiping database, bad version:" << oldVersion;
        purge();
        create();
    } else if (oldVersion < version) {
        qCWarning(LOG_DATABASE) << "Upgrading database: "
                << oldVersion << " -> " << version;
        upgrade(oldVersion, version);
    } else {
        qCDebug(LOG_DATABASE) << "Database is up to date:" << version;
    }
    return mDatabase.open();
}

qint32 DbManager::getDatabaseVersion() {
    if (mDatabase.isOpen()) {
        QSqlQuery query(mDatabase);
        query.prepare("PRAGMA user_version");
        if (!query.exec()) return -1;
        bool hasVersion = query.next();
        if (!hasVersion) return 0;
        return query.value(0).toInt();
    } else {
        qCCritical(LOG_DATABASE) << "Can't get database version when database closed!";
    }
    return -1;
}

void DbManager::setDatabaseVersion(qint32 version) {
    if (mDatabase.isOpen()) {
        QSqlQuery query(mDatabase);
        query.exec(QString("PRAGMA user_version = ") + QString::number(version));
    } else {
        qCCritical(LOG_DATABASE) << "Can't set database version when database closed!";
    }
}

void DbManager::close() {
    if (mDatabase.isOpen()) {
        mDatabase.close();
        qCDebug(LOG_DATABASE) << "Closed database.";
    }
}

void DbManager::reopen() {
    if (!mDatabase.isOpen()) {
        mDatabase.open();
        qCDebug(LOG_DATABASE) << "Reopen database.";
    }
}

QSqlDatabase DbManager::database() const {
    return mDatabase;
}

bool DbManager::executeSqlQueries(const QStringList &queries) {
    QSqlQuery query(mDatabase);
    beginTransaction();
    Q_FOREACH(const QString sql, queries) {
        if (!query.exec(sql)) {
            qCCritical(LOG_DATABASE) << "Failed to execute query: "
                << query.lastQuery() << " " << query.lastError();
            rollbackTransaction();
            return false;
        }
    }
    finishTransaction();
    return true;
}

bool DbManager::create() {
    QStringList queries;

    queries << Schema::CREATE_USERS;
    queries << Schema::CREATE_DIALOGS;
    queries << Schema::CREATE_CHATS;
    queries << Schema::CREATE_MESSAGES;
    queries << Schema::CREATE_THUMBNAILS;
    queries << Schema::CREATE_STATE;
    queries << Schema::CREATE_MEDIA_PHOTOS;
    queries << Schema::CREATE_PHOTO_SIZES;
    queries << Schema::CREATE_MEDIA_VIDEOS;
    queries << Schema::CREATE_MESSAGE_ACTIONS;
    queries << UpgradeGroupChat::CREATE_PARTICIPANTS;
    queries << UpgradeSecretChats::CREATE_SECRET_CHATS;
    queries << Schema::CREATE_DIALOGS_VIEW;
    queries << Schema::CREATE_CONTACTS_VIEW;

    bool result = executeSqlQueries(queries);
    if (result) {
        setDatabaseVersion(DATABASE_VERSION);
        qCWarning(LOG_DATABASE) << "Database created.";
    } else {
        qCCritical(LOG_DATABASE) << "Failed to create database!";
    }
    return result;
}

bool DbManager::purge() {
    qCDebug(LOG_DATABASE) << "Purging database..";

    QStringList queries;

    // get all view names and drop all them
    QSqlQuery selectQuery(mDatabase);
    selectQuery.prepare("SELECT name FROM sqlite_master WHERE type='view'");
    if (!selectQuery.exec()) {
        qCCritical(LOG_DATABASE) << "Failed to purge database!" << selectQuery.lastError() << selectQuery.lastQuery();
        return false;
    }
    while (selectQuery.next()) {
        QString name = selectQuery.value("name").toString();
        queries << QString("DROP VIEW %1").arg(name);
    }

    queries << UpgradeSecretChats::DROP_SECRET_CHATS;
    queries << UpgradeGroupChat::DROP_PARTICIPANTS;
    queries << Schema::DROP_MESSAGE_ACTIONS;
    queries << Schema::DROP_MEDIA_VIDEOS;
    queries << Schema::DROP_PHOTO_SIZES;
    queries << Schema::DROP_MEDIA_PHOTOS;
    queries << Schema::DROP_STATE;
    queries << Schema::DROP_THUMBNAILS;
    queries << Schema::DROP_MESSAGES;
    queries << Schema::DROP_CHATS;
    queries << Schema::DROP_DIALOGS;
    queries << Schema::DROP_USERS;

    bool result = executeSqlQueries(queries);
    if (result) {
        qCWarning(LOG_DATABASE) << "Database purged.";
    } else {
        qCCritical(LOG_DATABASE) << "Failed to purge database!";
    }
    return result;
}

// Upgrade methods go here:

// --- end of upgrade methods

void DbManager::upgrade(qint32 oldVersion, qint32 newVersion) {
    qint32 version = oldVersion;
    switch (version) {
        case VERSION_INITIAL:
            version = VERSION_INITIAL;
            if (!upgradeMessageActionsRemovePhotoId()) break;
            // fall-through
        case VERSION_MESSAGE_ACTIONS_REMOVE_PHOTO_ID:
            version = VERSION_MESSAGE_ACTIONS_REMOVE_PHOTO_ID;
            if (!upgradeMessagesRandomId()) break;
            // fall-through
        case VERSION_MESSAGES_RANDOMID:
            version = VERSION_MESSAGES_RANDOMID;
            if (!upgradeGroupChat()) break;
            // fall-through
        case VERSION_GROUP_PROFILE:
            version = VERSION_GROUP_PROFILE;
            if (!upgradeDialogsMuted()) break;
            // fall-through
        case VERSION_DIALOGS_MUTED:
            version = VERSION_DIALOGS_MUTED;
            if (!upgradeDialogsDate()) break;
            // fall-through
        case VERSION_DIALOGS_DATE:
            version = VERSION_DIALOGS_DATE;
            if (!upgradeSecretChats()) break;
            // fall-through
        case VERSION_SECRET_CHATS:
            version = VERSION_SECRET_CHATS;
            if (!upgradeReadHistory()) break;
            // fall-through
        case VERSION_READ_HISTORY:
            version = VERSION_READ_HISTORY;
            if (!upgradeEncryptedMediaKeys()) break;
            // fall-through
        case VERSION_ENCRYPTED_MEDIA_KEYS:
            version = VERSION_ENCRYPTED_MEDIA_KEYS;
            if (!upgradeEncryptedTtl()) break;
            // fall-through
        case VERSION_ENCRYPTED_TTL:
            version = VERSION_ENCRYPTED_TTL;
            if (!upgradeProfilesBigPhotos()) break;
            // fall-through
        case VERSION_PROFILE_BIG_PHOTO:
            version = VERSION_PROFILE_BIG_PHOTO;

        // Upgrade cases go here.
    }
    if (version != newVersion) {
        qCWarning(LOG_DATABASE) << "Failed to upgrade db, have to purge data!"
                << version << newVersion;
        purge();
        create();
        version = newVersion;
    } else {
        qCDebug(LOG_DATABASE) << "Database updated to version:" << version;
    }
    setDatabaseVersion(version);
    Q_EMIT upgraded(version);
}

bool DbManager::upgradeMessageActionsRemovePhotoId() {
    qCDebug(LOG_DATABASE) << "Upgrading database; deleting column photoId from table messageActions";

    QStringList queries;
    queries << MessageActionsRemovePhotoId::CREATE_NEW_MESSAGE_ACTIONS;
    queries << MessageActionsRemovePhotoId::COPY_MESSAGE_ACTIONS_TO_NEW_TABLE;
    queries << Schema::DROP_MESSAGE_ACTIONS;
    queries << MessageActionsRemovePhotoId::RENAME_NEW_MESSAGE_ACTIONS;

    return executeSqlQueries(queries);
}

bool DbManager::upgradeMessagesRandomId() {
    qCDebug(LOG_DATABASE) << "Upgrading database; adding randomId column to messages table";

    QStringList queries;
    queries << MessagesRandomId::ALTER_MESSAGES;

    return executeSqlQueries(queries);
}

bool DbManager::upgradeGroupChat() {
    qCDebug(LOG_DATABASE) << "Upgrading database: adding group chat";

    QStringList queries;
    queries << UpgradeGroupChat::ALTER_CHATS_ADD_DATE;
    queries << UpgradeGroupChat::ALTER_CHATS_ADD_LEFT;
    queries << UpgradeGroupChat::ALTER_CHATS_ADD_VERSION;
    queries << UpgradeGroupChat::CREATE_PARTICIPANTS;

    return executeSqlQueries(queries);
}

bool DbManager::upgradeDialogsMuted() {
    qCDebug(LOG_DATABASE) << "Upgrading database: adding muted field to dialogs table";

    QStringList queries;
    queries << UpgradeDialogsMuted::ALTER_DIALOGS_ADD_MUTED;

    return executeSqlQueries(queries);
}

bool DbManager::upgradeDialogsDate() {

    qCDebug(LOG_DATABASE) << "Upgrading database: adding date field to dialogs table";

    QStringList queries;
    queries << UpgradeDialogsDate::ALTER_DIALOGS_ADD_DATE;
    queries << UpgradeDialogsDate::DELETE_DIALOGS_VIEW;
    queries << Schema::CREATE_DIALOGS_VIEW;

    //set an initial value for dialogs date the date of the last added message to the dialog
    QString sql = QString("SELECT id from dialogs");
    QSqlQuery query(mDatabase);
    query.prepare(sql);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "impossible to upgrade dialogs database" << query.lastError() << query.lastQuery();
        return false;
    }

    QSqlQuery internalQuery(mDatabase);
    while (query.next()) {
        qint32 dialogId = query.value("id").toInt();

        internalQuery.prepare("SELECT date FROM messages WHERE (toIsChat=1 and toId=:dialogIdParam1) "
                              "OR (toIsChat=0 AND (toId=:dialogIdParam2 OR fromId=:dialogIdParam3)) "
                              "ORDER BY DATE ASC LIMIT 1");
        internalQuery.bindValue(":dialogIdParam1", dialogId);
        internalQuery.bindValue(":dialogIdParam2", dialogId);
        internalQuery.bindValue(":dialogIdParam3", dialogId);
        if (!internalQuery.exec()) {
            qCCritical(LOG_DATABASE) << "impossible to get date for dialog" << dialogId << internalQuery.lastError()
                                        << internalQuery.lastQuery();
            return false;
        }

        if (!internalQuery.next()) {
            qCCritical(LOG_DATABASE) << "impossible to update date for dialog. Not found dialog related date";
            return false;
        }

        qint32 date = internalQuery.value("date").toInt();

        queries << QString("UPDATE dialogs SET date=%1 WHERE id=%2").arg(QString::number(date), QString::number(dialogId));
    }

    return executeSqlQueries(queries);
}

bool DbManager::upgradeSecretChats() {
    qCDebug(LOG_DATABASE) << "Upgrading database: creating secret chats table";

    purge();
    create();
    return true;
}

bool DbManager::upgradeReadHistory() {
    qCDebug(LOG_DATABASE) << "Upgrading database: invalid history, re-creating db";

    purge();
    create();
    return true;
}

bool DbManager::upgradeEncryptedMediaKeys() {
    qCDebug(LOG_DATABASE) << "Upgrading database: adding key and iv fields to mediaPhotos and mediaVideos tables";

    QStringList queries;
    queries << UpgradeEncryptedMediaKeys::ALTER_MEDIA_VIDEOS_ADD_KEY;
    queries << UpgradeEncryptedMediaKeys::ALTER_MEDIA_VIDEOS_ADD_IV;
    queries << UpgradeEncryptedMediaKeys::ALTER_MEDIA_PHOTOS_ADD_KEY;
    queries << UpgradeEncryptedMediaKeys::ALTER_MEDIA_PHOTOS_ADD_IV;

    return executeSqlQueries(queries);
}

bool DbManager::upgradeEncryptedTtl() {
    qCDebug(LOG_DATABASE) << "Upgrading database: adding ttl to dialogsView";

    QStringList queries;
    queries << UpgradeEncryptedTtl::ALTER_SECRET_CHATS_ADD_TTL;
    queries << UpgradeEncryptedTtl::ALTER_MESSAGES_ADD_READTIME;
    queries << UpgradeEncryptedTtl::ALTER_MESSAGES_ADD_TTL;
    queries << Schema::DROP_DIALOGS_VIEW;
    queries << Schema::CREATE_DIALOGS_VIEW;
    return executeSqlQueries(queries);
}

bool DbManager::upgradeProfilesBigPhotos() {
    qCDebug(LOG_DATABASE) << "Upgrading database: adding bigPhoto to users and chats tables";
    return purgeButSecret();
}

bool DbManager::purgeButSecret() {
    qCDebug(LOG_DATABASE) << "Purging database but secret chats data...";

    QSqlQuery query(mDatabase);
    query.exec("PRAGMA foreign_keys = false");

    QStringList queries;

    queries << PurgeButSecret::DELETE_FILE_LOCATIONS_BUT_SECRET;
    queries << PurgeButSecret::DELETE_PHOTO_SIZES_BUT_SECRET;
    queries << PurgeButSecret::DELETE_MEDIA_PHOTOS_BUT_SECRET;
    queries << PurgeButSecret::DELETE_MEDIA_VIDEOS_BUT_SECRET;
    queries << PurgeButSecret::DELETE_MESSAGE_ACTIONS_BUT_SECRET;
    queries << PurgeButSecret::DELETE_MESSAGES_BUT_SECRET;
    queries << PurgeButSecret::DELETE_DIALOGS_BUT_SECRET;

    queries << UpgradeGroupChat::DROP_PARTICIPANTS;
    queries << Schema::DROP_CHATS;
    queries << Schema::DROP_USERS;
    queries << Schema::CREATE_USERS;
    queries << Schema::CREATE_CHATS;
    queries << UpgradeGroupChat::CREATE_PARTICIPANTS;

    bool result = executeSqlQueries(queries);

    query.exec("PRAGMA foreign_keys = true");

    return result;
}

bool DbManager::beginTransaction() {
    mTransactionsDepth++;
    if (mTransactionsDepth == 1) {
        return mDatabase.transaction();
    }
    return false;
}

bool DbManager::finishTransaction() {
    mTransactionsDepth--;
    if (mTransactionsDepth == 0) {
        return mDatabase.commit();
    }
    return false;
}

bool DbManager::rollbackTransaction() {
    bool result = false;
    if (mTransactionsDepth > 0) {
        result = mDatabase.rollback();
    }
    mTransactionsDepth = 0;
    return result;
}
