/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DBSCHEMA_H
#define DBSCHEMA_H

/*
   Initial db schema is in Schema namespace. Future db schema upgrades
   should have their own namespace.
*/

#include "types/user.h"

namespace Schema {

const QString CREATE_USERS(
"CREATE TABLE users (\n"
"    id INT NOT NULL,\n"
"    firstName VARCHAR(255),\n"
"    lastName VARCHAR(255),\n"
"    phone VARCHAR(15),\n"
"    thumbnail INT DEFAULT 0,\n"
"    bigPhoto INT DEFAULT 0,\n"
"    lastSeenOnline INT,\n"
"    online BOOL,\n"
"    type INT,\n"
"    accessHash LONG,\n"
"    CONSTRAINT pk_users_id PRIMARY KEY (id),\n"
"    CONSTRAINT fk_users_thumbnail FOREIGN KEY (thumbnail) REFERENCES fileLocations(rowid),\n"
"    CONSTRAINT fk_users_bigPhoto FOREIGN KEY (bigPhoto) REFERENCES fileLocations(rowid)\n"
")");

const QString DROP_USERS("DROP TABLE IF EXISTS users");

const QString CREATE_DIALOGS(
"CREATE TABLE dialogs (\n"
"    id int NOT NULL,\n"
"    isChat BOOL,\n"
"    unreadCount INT,\n"
"    muted BOOL DEFAULT 0,\n"
"    date INT,\n"
"    isSecret BOOL DEFAULT 0,\n"
"    CONSTRAINT pk_dialogs_id PRIMARY KEY (id),\n"
"    CONSTRAINT fk_secretChats_id FOREIGN KEY (id) REFERENCES secretChats(id)\n"
")");

const QString DROP_DIALOGS("DROP TABLE IF EXISTS dialogs");

const QString CREATE_CHATS(
"CREATE TABLE chats (\n"
"    id INT NOT NULL,\n"
"    title VARCHAR(255),\n"
"    participantsCount TINYINT,\n"
"    thumbnail INT DEFAULT 0,\n"
"    bigPhoto INT DEFAULT 0,\n"
"    date INT,\n"
"    left BOOL,\n"
"    version INT,\n"
"    CONSTRAINT pk_chats_id PRIMARY KEY (id),\n"
"    CONSTRAINT fk_chat_thumbnail FOREIGN KEY (thumbnail) REFERENCES fileLocations(rowid),\n"
"    CONSTRAINT fk_chat_bigPhoto FOREIGN KEY (bigPhoto) REFERENCES fileLocations(rowid)\n"
")");

const QString DROP_CHATS("DROP TABLE IF EXISTS chats");

const QString CREATE_MESSAGES(
"CREATE TABLE messages(\n"
"    id INT NOT NULL,\n"
"    dialogId INT,\n"
"    randomId LONG DEFAULT 0,\n"
"    fromId INT,\n"
"    toId INT,\n"
"    toIsChat BOOL,\n"
"    out BOOL,\n"
"    sent BOOL,\n"
"    unread BOOL,\n"
"    readTime INT DEFAULT 0,\n"
"    ttl INT DEFAULT 0,\n"
"    date INT,\n"
"    text VARCHAR(1024),\n"
"    mediaType INT,\n"
"    mediaId LONG,\n"
"    CONSTRAINT pk_messages_id PRIMARY KEY (id),\n"
"    CONSTRAINT fk_mediaId_mediaPhotos FOREIGN KEY (mediaId) REFERENCES mediaPhotos(id),\n"
"    CONSTRAINT fk_mediaId_mediaVideos FOREIGN KEY (mediaId) REFERENCES mediaVideos(id),\n"
"    CONSTRAINT fk_id_messageActions FOREIGN KEY (id) REFERENCES messageActions(messageId)\n"
")");

const QString DROP_MESSAGES("DROP TABLE IF EXISTS messages");

const QString CREATE_THUMBNAILS(
"CREATE TABLE fileLocations(\n"
"    localPath VARCHAR(255),\n"
"    dcId INT,\n"
"    localId INT,\n"
"    secret BIGINT,\n"
"    volumeId BIGINT\n"
")");

const QString DROP_THUMBNAILS("DROP TABLE IF EXISTS fileLocations");

const QString CREATE_STATE(
"CREATE TABLE state(\n"
"    name VARCHAR(255),\n"
"    value VARCHAR(255)\n"
")");

const QString DROP_STATE("DROP TABLE IF EXISTS state");

const QString CREATE_MEDIA_PHOTOS(
"CREATE TABLE mediaPhotos(\n"
"    id LONG,\n"
"    accessHash LONG,\n"
"    userId INT,\n"
"    date INT,\n"
"    caption VARCHAR(255),\n"
"    geoLat LONG,\n"
"    geoLong LONG,\n"
"    key TEXT,\n"
"    iv TEXT,\n"
"    CONSTRAINT pk_mediaPhotos_id PRIMARY KEY (id),\n"
"    CONSTRAINT fk_mediaPhotos_id FOREIGN KEY (id) REFERENCES photoSizes(photoId)"
")");

const QString DROP_MEDIA_PHOTOS("DROP TABLE IF EXISTS mediaPhotos");

const QString CREATE_PHOTO_SIZES(
"CREATE TABLE photoSizes(\n"
"    photoId LONG,\n"
"    type VARCHAR(255),\n"
"    width INT,\n"
"    height INT,\n"
"    size INT,\n"
"    fileLocationId INT,\n"
"    CONSTRAINT fk_photoSizes_fileLocationId FOREIGN KEY (fileLocationId) REFERENCES fileLocations(rowid)\n"
")");

const QString DROP_PHOTO_SIZES("DROP TABLE IF EXISTS photoSizes");

const QString CREATE_DIALOGS_VIEW(
"CREATE VIEW IF NOT EXISTS dialogsView AS\n"
"   SELECT dialogs.id AS id,\n"
"       isChat,\n"
"       (SELECT coalesce(users.firstName,'') || ' ' || coalesce(users.lastName,'') FROM users WHERE (users.id=dialogs.id AND isChat=0) OR (users.id IN (SELECT peerId FROM secretChats WHERE secretChats.id=dialogs.id)) UNION SELECT title from chats WHERE chats.id=dialogs.id AND isChat=1) AS name,\n"
"       unreadCount,\n"
"       (SELECT localPath FROM fileLocations JOIN users ON fileLocations.rowid=users.thumbnail WHERE users.id=dialogs.id OR users.id IN (SELECT peerId FROM secretChats WHERE secretChats.id=dialogs.id) UNION SELECT localPath FROM fileLocations JOIN chats ON fileLocations.rowid=chats.thumbnail WHERE chats.id=dialogs.id) AS thumbnail,\n"
"       (SELECT participantsCount FROM chats WHERE chats.id=dialogs.id) AS members,\n"
"       (SELECT online FROM users WHERE users.id=dialogs.id) AS userOnline,\n"
"       (SELECT lastSeenOnline FROM users WHERE users.id=dialogs.id) AS lastSeenOnline,\n"
"       (SELECT peerId FROM secretChats WHERE secretChats.id=dialogs.id) AS peerId,\n"
"       (SELECT type FROM users WHERE users.id=dialogs.id) AS peerType,\n"
"       dialogs.date AS date,\n"
"       isSecret,\n"
"       (SELECT state FROM secretChats WHERE secretChats.id=dialogs.id) AS state,\n"
"       (SELECT ttl FROM secretChats WHERE secretChats.id=dialogs.id) AS ttl,"
"       messages.id AS topMessageId,\n"
"       fromId AS topMessageFromId,\n"
"       (SELECT firstName FROM users WHERE id=fromId) AS topMessageFromFirstName,\n"
"       (SELECT lastName FROM users WHERE id=fromId) AS topMessageFromLastName,\n"
"       (SELECT phone FROM users WHERE id=fromId) AS topMessageFromPhone,\n"
"       toId AS topMessageToId,\n"
"       out AS topMessageOut,\n"
"       unread AS topMessageUnread,\n"
"       sent AS topMessageSent,\n"
"       MAX(messages.date) AS topMessageDate,\n"
"       text AS topMessageText,\n"
"       mediaType AS topMessageMediaType,\n"
"       (SELECT type FROM messageActions WHERE messageId=messages.id) AS topMessageActionType,\n"
"       (SELECT title FROM messageActions WHERE messageId=messages.id) AS topMessageActionTitle,\n"
"       (SELECT coalesce(users.firstName,'') || ' ' || coalesce(users.lastName,'') FROM users WHERE id IN (SELECT userId FROM messageActions WHERE messageId=messages.id)) AS topMessageActionUser\n"
"   FROM dialogs LEFT OUTER JOIN messages\n"
"   ON messages.dialogId=dialogs.id AND toId!=fromId\n"
"   GROUP BY dialogs.id\n"
"   ORDER BY dialogs.date DESC\n");

const QString DROP_DIALOGS_VIEW("DROP VIEW IF EXISTS dialogsView");

const QString CREATE_CONTACTS_VIEW = QString(
"CREATE VIEW IF NOT EXISTS contactsView AS\n"
"   SELECT id,\n"
"       firstName,\n"
"       lastName,\n"
"       phone,\n"
"       (SELECT localPath FROM fileLocations WHERE rowid=thumbnail) AS thumbnail,\n"
"       lastSeenOnline,\n"
"       online,\n"
"       type\n"
"   FROM users\n"
"   WHERE type=%1\n"
"   ORDER BY firstName ASC").arg(User::typeUserContact);

const QString DROP_CONTACTS_VIEW("DROP VIEW IF EXISTS contactsView");

const QString CREATE_MEDIA_VIDEOS(
"CREATE TABLE mediaVideos(\n"
"    id LONG,\n"
"    accessHash LONG,\n"
"    userId INT,\n"
"    date INT,\n"
"    caption VARCHAR(255),\n"
"    duration INT,\n"
"    mimeType VARCHAR(255),\n"
"    size INT,\n"
"    dcId INT,\n"
"    width INT,\n"
"    height INT,\n"
"    localPath VARCHAR(255),\n"
"    key TEXT,\n"
"    iv TEXT,\n"
"    CONSTRAINT pkMediaVideosId PRIMARY KEY (id)\n"
")");

const QString DROP_MEDIA_VIDEOS("DROP TABLE IF EXISTS mediaVideos");

const QString ALTER_MEDIA_VIDEOS("ALTER TABLE mediaVideos ADD COLUMN localPath VARCHAR(255)\n");

const QString CREATE_MESSAGE_ACTIONS(
"CREATE TABLE messageActions(\n"
"   messageId INT,\n"
"   type INT,\n"
"   title VARCHAR(255),\n"
"   userId INT,\n"
"   CONSTRAINT pkMessageActionsId PRIMARY KEY (messageId)\n"
")");

const QString DROP_MESSAGE_ACTIONS("DROP TABLE IF EXISTS messageActions");

}

namespace MessageActionsRemovePhotoId {

const QString CREATE_NEW_MESSAGE_ACTIONS(
"CREATE TABLE messageActions_new(\n"
"   messageId INT,\n"
"   type INT,\n"
"   title VARCHAR(255),\n"
"   userId INT,\n"
"   CONSTRAINT pkMessageActionsId PRIMARY KEY (messageId)\n"
")");

const QString COPY_MESSAGE_ACTIONS_TO_NEW_TABLE("INSERT INTO messageActions_new SELECT messageId, type, title, userId FROM messageActions");

const QString RENAME_NEW_MESSAGE_ACTIONS("ALTER TABLE messageActions_new RENAME TO messageActions");

}

namespace MessagesRandomId {

const QString ALTER_MESSAGES("ALTER TABLE MESSAGES ADD COLUMN randomId LONG DEFAULT 0");

}

namespace UpgradeGroupChat {

const QString ALTER_CHATS_ADD_DATE("ALTER TABLE chats ADD COLUMN date INT");
const QString ALTER_CHATS_ADD_LEFT("ALTER TABLE chats ADD COLUMN left BOOL");
const QString ALTER_CHATS_ADD_VERSION("ALTER TABLE chats ADD COLUMN version INT");

const QString CREATE_PARTICIPANTS(
        "CREATE TABLE participants(\n"
        "    chatId INT,\n"
        "    userId INT,\n"
        "    inviterId INT,\n"
        "    date INT,\n"
        "    isAdmin BOOL,\n"
        "    version INT\n"
        ")");

const QString DROP_PARTICIPANTS("DROP TABLE IF EXISTS participants");

}

namespace UpgradeDialogsMuted {

const QString ALTER_DIALOGS_ADD_MUTED("ALTER TABLE dialogs ADD COLUMN muted BOOL DEFAULT 0");
}

namespace UpgradeDialogsDate {

const QString ALTER_DIALOGS_ADD_DATE("ALTER TABLE dialogs ADD COLUMN date INT");
const QString DELETE_DIALOGS_VIEW("DROP VIEW dialogsView");
}

namespace UpgradeSecretChats {

const QString CREATE_SECRET_CHATS(
        "CREATE TABLE secretChats (\n"
        "    id INT,\n"
        "    peerId INT,\n"
        "    date INT,\n"
        "    state INT,\n"
        "    ttl INT DEFAULT 0,\n"
        "    CONSTRAINT pk_secretchats_id PRIMARY KEY (id)\n"
        ")");
const QString DROP_SECRET_CHATS("DROP TABLE IF EXISTS secretChats");
}

namespace UpgradeEncryptedMediaKeys {

const QString ALTER_MEDIA_VIDEOS_ADD_KEY("ALTER TABLE mediaVideos ADD COLUMN key TEXT");
const QString ALTER_MEDIA_VIDEOS_ADD_IV("ALTER TABLE mediaVideos ADD COLUMN iv TEXT");
const QString ALTER_MEDIA_PHOTOS_ADD_KEY("ALTER TABLE mediaPhotos ADD COLUMN key TEXT");
const QString ALTER_MEDIA_PHOTOS_ADD_IV("ALTER TABLE mediaPhotos ADD COLUMN iv TEXT");
}

namespace UpgradeEncryptedTtl {

const QString ALTER_SECRET_CHATS_ADD_TTL("ALTER TABLE secretChats ADD COLUMN ttl INT DEFAULT 0");
const QString ALTER_MESSAGES_ADD_READTIME("ALTER TABLE messages ADD COLUMN readTime INT DEFAULT 0");
const QString ALTER_MESSAGES_ADD_TTL("ALTER TABLE messages ADD COLUMN ttl INT DEFAULT 0");

}

namespace PurgeButSecret {

const QString DELETE_FILE_LOCATIONS_BUT_SECRET("DELETE FROM fileLocations WHERE rowid IN "
                                               "(SELECT fileLocationId FROM photoSizes WHERE photoId IN "
                                               "(SELECT mediaId FROM messages WHERE dialogId NOT IN "
                                               "(SELECT id FROM secretChats)));");
const QString DELETE_PHOTO_SIZES_BUT_SECRET("DELETE FROM photoSizes WHERE photoId IN "
                                               "(SELECT mediaId FROM messages WHERE dialogId NOT IN "
                                               "(SELECT id FROM secretChats))");
const QString DELETE_MEDIA_PHOTOS_BUT_SECRET("DELETE FROM mediaPhotos WHERE id IN "
                                               "(SELECT mediaId FROM messages WHERE dialogId NOT IN "
                                               "(SELECT id FROM secretChats))");
const QString DELETE_MEDIA_VIDEOS_BUT_SECRET("DELETE FROM mediaVideos WHERE id IN "
                                               "(SELECT mediaId FROM messages WHERE dialogId NOT IN "
                                               "(SELECT id FROM secretChats))");
const QString DELETE_MESSAGE_ACTIONS_BUT_SECRET("DELETE FROM messageActions WHERE messageId IN "
                                                "(SELECT id FROM messages WHERE dialogId NOT IN"
                                                "(SELECT id FROM secretChats))");
const QString DELETE_MESSAGES_BUT_SECRET("DELETE FROM messages WHERE dialogId NOT IN (SELECT id FROM secretChats)");
const QString DELETE_DIALOGS_BUT_SECRET("DELETE FROM dialogs WHERE isSecret=0");

}
#endif // DBSCHEMA_H
