/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "contactsmodel.h"
#include "data.h"
#include "dbmanager.h"
#include "rawapiclient.h"

ContactsModel::ContactsModel(QObject *parent) :
    QAbstractListModel(parent),
    mCanFetchMore(true),
    mTelegramClient(0) {
}

ContactsModel::~ContactsModel() {
}

void ContactsModel::setTelegramClient(TelegramClient *telegramClient) {

    if (!mTelegramClient && telegramClient) {

        mTelegramClient = telegramClient;

        connect(mTelegramClient, SIGNAL(hasConnected()), SLOT(updateQuery()));
        connect(mTelegramClient, SIGNAL(authLogOutAnswer(qint64, bool)), SLOT(resetModel()));

        connect(mTelegramClient, SIGNAL(contactsAdded(QList<ContactItem>)), SLOT(onContactsAdded(QList<ContactItem>)));
        connect(mTelegramClient, SIGNAL(contactThumbnailUpdated(qint32,QString)), SLOT(onContactThumbnailUpdated(qint32,QString)));

        connect(mTelegramClient, SIGNAL(contactsNextPage(QList<ContactItem>)), SLOT(onNextPage(QList<ContactItem>)));
    }
}

QHash<int, QByteArray> ContactsModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[FirstNameRole] = "firstName";
    roles[LastNameRole] = "lastName";
    roles[PhoneRole] = "phone";
    roles[ThumbnailRole] = "thumbnail";
    roles[LastSeenOnlineRole] = "lastSeenOnline";
    roles[OnlineRole] = "online";
    return roles;
}

int ContactsModel::rowCount(const QModelIndex &parent) const {

    if (parent.isValid()) {
        return 0;
    }

    return mContacts.count();
}

QVariant ContactsModel::data(const QModelIndex &index, int role) const {

    if (!index.isValid() || index.row() < 0 || index.row() >= mContacts.count()) {
        return QVariant();
    }

    ContactItem c = mContacts.at(index.row());
    QVariant result;

    switch (role) {
    case IdRole:
        result = c.id();
        break;
    case FirstNameRole:
        result = c.firstName();
        break;
    case LastNameRole:
        result = c.lastName();
        break;
    case PhoneRole:
        result = c.phone();
        break;
    case ThumbnailRole:
        result = c.thumbnail();
        break;
    case LastSeenOnlineRole:
        result = c.lastSeenOnline();
        break;
    case OnlineRole:
        result = c.online();
        break;
    }

    return result;
}

void ContactsModel::updateQuery() {

    if (!mContacts.isEmpty()) {
        mContactIdsMap.clear();
        beginRemoveRows(QModelIndex(), 0, mContacts.count() - 1);
        mContacts.clear();
        endRemoveRows();
    }

    mCanFetchMore = true;

    if (!mTelegramClient) {
        qCWarning(TG_PLUGIN_LOGIC) << "Telegram client is not ready for creating data view";
        return;
    }

    mTelegramClient->initializeContactsView();

    fetchMore();
}


bool ContactsModel::canFetchMore(const QModelIndex &parent) const
{
    if (parent.isValid()) {
        qCDebug(TG_PLUGIN_LOGIC) << "fetch more: false";
        return false;
    }

    qCDebug(TG_PLUGIN_LOGIC) << "fetch more:" << mCanFetchMore;
    return mCanFetchMore;
}

void ContactsModel::fetchMore(const QModelIndex &parent)
{
    if (parent.isValid() || !mTelegramClient) {
        return;
    }
    mTelegramClient->fetchContactsNextPage();
}

void ContactsModel::onNextPage(const QList<ContactItem> &contacts) {
    qCDebug(TG_PLUGIN_LOGIC) << "Got contacts:" << contacts.count();
    if (contacts.isEmpty() && mCanFetchMore) {
        mCanFetchMore = false;
        // view (db) contacts exhausted, query server for more contacts
        if (!mTelegramClient) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get contacts. Telegram client reference is null";
        } else if (!mTelegramClient->authenticated()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get contacts. Telegram client reference is not logged in";
        } else if (!mTelegramClient->connected()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get contacts. Telegram client reference is not connected";
        } else {
            mTelegramClient->contactsGetContacts();
        }
    } else {
        beginInsertRows(QModelIndex(), mContacts.count(), mContacts.count() + contacts.count() - 1);
        Q_FOREACH (const ContactItem newContact, contacts) {
            mContactIdsMap.insert(newContact.id(), mContacts.count());
            mContacts << newContact;
        }
        endInsertRows();
    }
}

void ContactsModel::append(const ContactItem &contact) {
    // if this contact exists, update, otherwise append at the end
    qint32 pos = mContactIdsMap.value(contact.id());
    if (pos) {
        mContacts.replace(pos, contact);
    } else {
        beginInsertRows(QModelIndex(), mContacts.count(), mContacts.count());
        mContactIdsMap.insert(contact.id(), mContacts.count());
        mContacts.append(contact);
        endInsertRows();
    }
}

void ContactsModel::clear() {
    beginResetModel();
    mContactIdsMap.clear();
    mContacts.clear();
    endResetModel();
}

void ContactsModel::resetModel() {
    disconnect(mTelegramClient);
    mTelegramClient = 0;
    mCanFetchMore = true;
    clear();
}

ContactItem &ContactsModel::contact(qint32 id) {
    Q_ASSERT(id);
    Q_ASSERT(mContacts.count() > 0);
    qint32 pos = mContactIdsMap.value(id);
    return mContacts[pos];
}

void ContactsModel::updateThumbnail(qint32 id, const QString &thumbnailPath) {
    qint32 pos = mContactIdsMap.value(id);
    if ((pos >= 0) && (pos < mContacts.count())) {
        ContactItem &c = mContacts[pos];
        c.setThumbnail(thumbnailPath);
        const QVector<qint32> changedRoles(QVector<qint32>() << ThumbnailRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

// SLOTS
void ContactsModel::onContactsAdded(const QList<ContactItem> &contacts) {

    // fill a list with the new contacts not existing yet in model
    QList<ContactItem> contactsToAppend;

    // replace contact or add to contacts to append list
    Q_FOREACH (ContactItem c, contacts) {
        qint32 pos = mContactIdsMap.value(c.id(), -1);
        if (pos >= 0) {
            mContacts.replace(pos, c);
        } else {
            contactsToAppend.append(c);
        }
    }

    beginInsertRows(QModelIndex(), mContacts.count(), mContacts.count() + contactsToAppend.count() - 1);
    Q_FOREACH (const ContactItem newContact, contactsToAppend) {
        mContactIdsMap.insert(newContact.id(), mContacts.count());
        mContacts << newContact;
    }
    endInsertRows();
}

void ContactsModel::onContactThumbnailUpdated(qint32 id, const QString &thumbnail) {
    qint32 pos = mContactIdsMap.value(id, -1);
    if (pos >= 0) {
        mContacts[pos].setThumbnail(thumbnail);
        const QVector<qint32> changedRoles(QVector<qint32>() << ThumbnailRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}
