/*
 * Copyright 2015 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "dialogmodel.h"
#include "../types/tluser.h"
#include "../tools.h"

#define TYPING_INTERVAL 5000

DialogModel::DialogModel(QObject *parent) : QObject(parent),
        mTelegramClient(NULL),
        mChatId(0),
        mPeerId(0),
        mPeerType((qint32) InputPeer::typeInputPeerContact),
        mWhoIsTyping(""),
        mIsTyping(false),
        mIsSecret(false),
        mState(0),
        mTitle(""),
        mPhone(""),
        mTtl(0),
        mTypingTimerId(0) {
    qRegisterMetaType<InputPeer>("InputPeer");
}

DialogModel::~DialogModel() {
    mTelegramClient = 0;
}

void DialogModel::resetDialogDetails() {
    setChatId(0);
    setPeerId(0);
    setPeerType((qint32) InputPeer::typeInputPeerContact);
    setIsChat(false);
    setIsTyping(false);
    setIsSecret(false);
    setState(0);
    setTitle(QString());
    setPhone(QString());
    setTtl(0);
}

/**
 * @brief DialogModel::setup
 *        We assume the entry for this chat id exists in the database.
 *        In fact, this is currently known as dialogId (todo: refactor name).
 * @param telegramClient
 * @param chatId
 */
void DialogModel::setup(TelegramClient *telegramClient, qint32 chatId) {
    Q_ASSERT(telegramClient != NULL);

    if (mTelegramClient == NULL) {
        mTelegramClient = telegramClient;

        connect(mTelegramClient, SIGNAL(secretChatUpdated(qint32)),
                           this,   SLOT(refreshDialogDetails(qint32)));
        connect(mTelegramClient, SIGNAL(userTyping(qint32,QString)),
                           this,   SLOT(userTyping(qint32,QString)));
        connect(mTelegramClient, SIGNAL(messageAdded(MessageItem)),
                            this, SLOT(onMessageReceived(MessageItem)));

        // TODO karni: hook up live state
    }

    if (mTelegramClient != NULL) {
        setChatId(chatId);
    }
}

void DialogModel::onMessageReceived(const MessageItem &message) {
    if (!message.out() && message.dialogId() == mChatId) {
        setIsTyping(false);
    }
}

qint32 DialogModel::getChatId() const {
    return mChatId;
}

void DialogModel::setChatId(qint32 chatId) {
    if (mChatId != chatId) {
        mChatId = chatId;
        if (chatId != 0) {
            refreshDialogDetails(chatId);
        }
        Q_EMIT chatIdChanged(chatId);
    }
}

qint32 DialogModel::getPeerId() const {
    return mPeerId;
}

void DialogModel::setPeerId(qint32 peerId) {
    if (mPeerId != peerId) {
        mPeerId = peerId;
        Q_EMIT peerIdChanged(peerId);
    }
}

qint32 DialogModel::getPeerType() const {
    return mPeerType;
}

void DialogModel::setPeerType(qint32 peerType) {
    if (mPeerType != peerType) {
        mPeerType = peerType;
        Q_EMIT peerTypeChanged(peerType);
    }
}

bool DialogModel::getIsChat() const {
    return mIsChat;
}

void DialogModel::setIsChat(bool isChat) {
    if (mIsChat != isChat) {
        mIsChat = isChat;
        Q_EMIT isChatChanged(isChat);
    }
}

QString DialogModel::getWhoIsTyping() const {
    return mWhoIsTyping;
}

void DialogModel::setWhoIsTyping(QString whoIsTyping) {
    if (mWhoIsTyping != whoIsTyping) {
        mWhoIsTyping = whoIsTyping;
        Q_EMIT whoIsTypingChanged(whoIsTyping);
    }
}

bool DialogModel::getIsTyping() const {
    return mIsTyping;
}

void DialogModel::setIsTyping(bool isTyping) {
    // Intentionally not comparing to previous value.
    mIsTyping = isTyping;
    Q_EMIT isTypingChanged(isTyping);
}

bool DialogModel::getIsSecret() const {
    return mIsSecret;
}

void DialogModel::setIsSecret(bool isSecret) {
    if (mIsSecret != isSecret) {
        mIsSecret = isSecret;
        Q_EMIT isSecretChanged(isSecret);
    }
}

qint32 DialogModel::getState() const {
    return mState;
}

void DialogModel::setState(qint32 state) {
    if (mState != state) {
        mState = state;
        Q_EMIT stateChanged(state);
    }
}

QString DialogModel::getTitle() const {
    return mTitle;
}

void DialogModel::setTitle(const QString &title) {
    if (mTitle != title) {
        mTitle = title;
        Q_EMIT titleChanged(title);
    }
}

QString DialogModel::getPhone() const {
    return mPhone;
}

void DialogModel::setPhone(QString phone) {
    if (mPhone != phone) {
        mPhone = phone;
        Q_EMIT phoneChanged(phone);
    }
}

qint32 DialogModel::getTtl() const {
    return mTtl;
}

void DialogModel::setTtl(qint32 ttl) {
    if (mTtl != ttl) {
        mTtl = ttl;
        Q_EMIT ttlChanged(ttl);
    }
}

void DialogModel::refreshDialogDetails(qint32 chatId) {
    if (mChatId != chatId) return;
    qCDebug(TG_PLUGIN_LOGIC) << "Querying dialog details" << chatId;

    QSqlQuery query;
    query.prepare(QString("SELECT id, isChat, name, thumbnail, members, userOnline, "
                  "lastSeenOnline, peerId, peerType, date, isSecret, state, ttl "
                  "FROM dialogsView WHERE id = %1").arg(QString::number(chatId)));

    // query.bindValue(":chatId", chatId); // no worky?

    // FUTURE: make use of thumbnail, members, userOnline and
    // lastSeenOnline in the chat page header.

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Failed to query dialog details for "
                << chatId << query.lastError() << query.lastQuery();
        resetDialogDetails();

    } else if (!query.next()) {
        qCDebug(TG_PLUGIN_LOGIC) << "No dialog details found for " << chatId;
        qCWarning(TG_PLUGIN_LOGIC) << "Opening chat with contact.";

        query.prepare(QString("SELECT firstName, lastName, type, accessHash, phone "
                              "FROM users WHERE id = %1")
                                    .arg(QString::number(chatId)));
        if (!query.exec() || !query.next()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Failed to get user for DialogModel!";
            resetDialogDetails();
            return;
        }

        setIsChat(false);

        setIsSecret(false);

        qint32 userType = query.value("type").toInt();
        setPeerType(Tools::toInputPeerType((User::UserType) userType));

        setPeerId(chatId);

        setState(0);

        QString firstName;
        QString lastName;
        if (query.exec() && query.next()) {
            firstName = query.value("firstName").toString();
            lastName = query.value("lastName").toString();
        } else {
            firstName = "Unknown";
        }
        if (lastName.length() > 0) {
            firstName.append(" ");
            firstName.append(lastName);
        }
        setTitle(firstName);

        QString phone = query.value("phone").toString();
        setPhone(phone);

        Q_EMIT dialogDetailsRefreshed();
    } else {
        qCDebug(TG_PLUGIN_LOGIC) << "Got dialog details";

        bool isChat = query.value("isChat").toBool();
        setIsChat(isChat);

        bool isSecret = query.value("isSecret").toBool();
        setIsSecret(isSecret);

        if (isSecret) {
            qint32 ttl = query.value("ttl").toInt();
            setTtl(ttl);
        }

        qint32 userType = query.value("peerType").toInt();
        setPeerType(Tools::toInputPeerType((User::UserType) userType));

        qint32 peerId = query.value("peerId").toInt();
        setPeerId(isSecret ? peerId : chatId); // *sigh*

        qint32 state = query.value("state").toInt();
        setState(state);

        QString title = query.value("name").toString();
        setTitle(title);

        // Ulgy it's here, I know. No time now.
        QSqlQuery phoneQuery;
        phoneQuery.prepare("SELECT phone FROM users WHERE id = :id");
        phoneQuery.bindValue(":id", isSecret ? peerId : chatId);
        if (phoneQuery.exec() && phoneQuery.next()) {
            QString phone = phoneQuery.value("phone").toString();
            setPhone(phone);
        }

        Q_EMIT dialogDetailsRefreshed();
    }
}

void DialogModel::userTyping(qint32 chatId, QString who) {
    if (mChatId != chatId) return;

    if (mTypingTimerId != 0) {
        killTimer(mTypingTimerId);
        mTypingTimerId = 0;
    }

    mTypingTimerId = startTimer(TYPING_INTERVAL);

    mWhoIsTyping = who;
    Q_EMIT whoIsTypingChanged(who);

    mIsTyping = true;
    Q_EMIT isTypingChanged(true);
}

void DialogModel::timerEvent(QTimerEvent *timerEvent) {
    if (mTypingTimerId != timerEvent->timerId()) return;

    killTimer(mTypingTimerId);
    mTypingTimerId = 0;
    mIsTyping = false;
    setWhoIsTyping("");
    setIsTyping(false);
}
