/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "dialogsmodel.h"

#include "telegramservice.h"
#include "types/tlmessagemedia.h"
#include "data.h"
#include <QFile>
#include <QDir>
#include "dbmanager.h"

DialogsModel::DialogsModel(QObject *parent) :
    QAbstractListModel(parent),
    mCanFetchMore(true),
    mState(Created),
    mSecretDialogsCount(0),
    mTelegramClient(0) {
}

DialogsModel::~DialogsModel() {
}

void DialogsModel::setTelegramClient(TelegramClient *telegramClient) {

    if (!mTelegramClient && telegramClient) {

        mTelegramClient = telegramClient;

        connect(mTelegramClient, SIGNAL(databaseInitialized()), SLOT(initialize()));
        connect(mTelegramClient, SIGNAL(authLoggedIn()), SLOT(initialize()));
        connect(mTelegramClient, SIGNAL(hasConnected()), SLOT(initialize()));
        connect(mTelegramClient, SIGNAL(authLogOutAnswer(qint64, bool)), SLOT(resetModel()));

        connect(mTelegramClient, SIGNAL(dialogsAdded(QList<DialogItem>,qint32)), SLOT(onDialogsAdded(QList<DialogItem>,qint32)));
        connect(mTelegramClient, SIGNAL(dialogThumbnailUpdated(qint32,QString)), SLOT(onThumbnailUpdated(qint32,QString)));
        connect(mTelegramClient, SIGNAL(dialogTitleUpdated(qint32,QString)), SLOT(onTitleUpdated(qint32,QString)));

        connect(mTelegramClient, SIGNAL(messageAdded(MessageItem)), SLOT(onMessageAdded(MessageItem)));
        connect(mTelegramClient, SIGNAL(messagesAdded(QList<MessageItem>,qint32)), SLOT(onMessagesAdded(QList<MessageItem>,qint32)));
        connect(mTelegramClient, SIGNAL(outgoingMessageAdded(qint32,MessageItem)), SLOT(onOutgoingMessageAdded(qint32,MessageItem)));
        connect(mTelegramClient, SIGNAL(messageUpdated(qint32,MessageItem)), SLOT(onMessageUpdated(qint32,MessageItem)));
        connect(mTelegramClient, SIGNAL(messagesMarkedAsRead(QList<qint32>)), SLOT(onMessagesMarkedAsRead(QList<qint32>)));
        connect(mTelegramClient, SIGNAL(messagesDeleted(QList<qint32>)), SLOT(onMessagesDeleted(QList<qint32>)));
        connect(mTelegramClient, SIGNAL(chatMarkedAsRead(qint32)), SLOT(onChatMarkedAsRead(qint32)));
        connect(mTelegramClient, SIGNAL(chatHistoryDeleted(qint32)), SLOT(onChatHistoryDeleted(qint32)));
        connect(mTelegramClient, SIGNAL(chatDeleted(qint32)), SLOT(onChatDeleted(qint32)));

        connect(mTelegramClient, SIGNAL(dialogsNextPage(QList<DialogItem>)), SLOT(onNextPage(QList<DialogItem>)));

        connect(mTelegramClient, SIGNAL(contactsAdded(QList<ContactItem>)), SLOT(onContactsAdded(QList<ContactItem>)));

        connect(mTelegramClient, SIGNAL(userTyping(qint32,QString)), SLOT(onUserTyping(qint32,QString)));
    }
}

QHash<int, QByteArray> DialogsModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[IsChatRole] = "isChat";
    roles[PeerIdRole] = "peerId";
    roles[PeerTypeRole] = "peerType";
    roles[PeerAccessHashRole] = "peerAccessHash";
    roles[NameRole] = "name";
    roles[UnreadCountRole] = "unreadCount";
    roles[ThumbnailRole] = "thumbnail";
    roles[TypingRole] = "typing";
    roles[WhoIsTypingRole] = "whoIsTyping";
    roles[MembersRole] = "members";
    roles[MembersOnlineRole] = "membersOnline";
    roles[UserOnlineRole] = "userOnline";
    roles[LastSeenOnlineRole] = "lastSeenOnline";
    roles[DateRole] = "date";
    roles[IsSecretRole]  = "isSecret";
    roles[StateRole] = "state";
    roles[TopMessageIdRole] = "topMessageId";
    roles[TopMessageFromIdRole] = "topMessageFromId";
    roles[TopMessageFromFirstNameRole] = "topMessageFromFirstName";
    roles[TopMessageFromLastNameRole] = "topMessageFromLastName";
    roles[TopMessageFromPhoneRole] = "topMessageFromPhone";
    roles[TopMessageToIdRole] = "topMessageToId";
    roles[TopMessageOutRole] = "topMessageOut";
    roles[TopMessageUnreadRole] = "topMessageUnread";
    roles[TopMessageSentRole] = "topMessageSent";
    roles[TopMessageDateRole] = "topMessageDate";
    roles[TopMessageTextRole] = "topMessageText";
    roles[TopMessageMediaTypeRole] = "topMessageMediaType";
    roles[TopMessageIsActionRole] = "topMessageIsAction";
    roles[TopMessageActionTypeRole] = "topMessageActionType";
    roles[TopMessageActionTitleRole] = "topMessageActionTitle";
    roles[TopMessageActionUserRole] = "topMessageActionUser";
    return roles;
}

qint32 DialogsModel::rowCount(const QModelIndex &parent) const {

    if (parent.isValid()) {
        return 0;
    }

    return mDialogs.size();
}

QVariant DialogsModel::get(int row) {
    if ((row < 0) || (row >= mDialogs.size())) {
        return QVariant();
    }
    return QVariant::fromValue(&mDialogs[row]);
}

QVariant DialogsModel::getDialogFromDialogId(int dialogId) {
    for (int i = 0; i < mDialogs.count(); i++) {
        const DialogItem &d = mDialogs.at(i);

        if (dialogId == d.id()) {
            return get(i);
        }
    }
    qCWarning(TG_PLUGIN_LOGIC) << "getFromId did not find a dialog!";
    return QVariant();
}

QVariant DialogsModel::data(const QModelIndex &index, int role) const {

    if (!index.isValid() || index.row() < 0 || index.row() >= mDialogs.count()) {
        return QVariant();
    }

    DialogItem dialogItem = mDialogs[index.row()];

    QVariant result;

    switch (role) {
    case IdRole:
        result = dialogItem.id();
        break;
    case IsChatRole:
        result = dialogItem.isChat();
        break;
    case PeerIdRole:
        result = dialogItem.peerId();
        break;
    case PeerTypeRole:
        result = dialogItem.peerType();
        break;
    case NameRole:
        result = dialogItem.name();
        break;
    case UnreadCountRole:
        result = dialogItem.unreadCount();
        break;
    case ThumbnailRole:
        result = dialogItem.thumbnail();
        break;
    case TypingRole:
        result = dialogItem.typing();
        break;
    case WhoIsTypingRole:
        result = dialogItem.whoisTyping();
        break;
    case MembersRole:
        result = dialogItem.members();
        break;
    case MembersOnlineRole:
        result = dialogItem.membersOnline();
        break;
    case UserOnlineRole:
        result = dialogItem.userOnline();
        break;
    case LastSeenOnlineRole:
        result = dialogItem.lastSeenOnline();
        break;
    case DateRole:
        result = dialogItem.date();
        break;
    case IsSecretRole:
        result = dialogItem.isSecret();
        break;
    case StateRole:
        result = dialogItem.state();
        break;
    case TopMessageIdRole:
        result = dialogItem.topMessage().id();
        break;
    case TopMessageFromIdRole:
        result = dialogItem.topMessage().fromId();
        break;
    case TopMessageFromFirstNameRole:
        result = dialogItem.topMessage().fromFirstName();
        break;
    case TopMessageFromLastNameRole:
        result = dialogItem.topMessage().fromLastName();
        break;
    case TopMessageFromPhoneRole:
        result = dialogItem.topMessage().fromPhone();
        break;
    case TopMessageToIdRole:
        result = dialogItem.topMessage().toId();
        break;
    case TopMessageOutRole:
        result = dialogItem.topMessage().out();
        break;
    case TopMessageUnreadRole:
        result = dialogItem.topMessage().unread();
        break;
    case TopMessageSentRole:
        result = dialogItem.topMessage().sent();
        break;
    case TopMessageDateRole:
        result = dialogItem.topMessage().date();
        break;
    case TopMessageTextRole:
        result = dialogItem.topMessage().text();
        break;
    case TopMessageMediaTypeRole:
        result = static_cast<qint32>(dialogItem.topMessage().mediaType());
        break;
    case TopMessageIsActionRole:
        result = dialogItem.topMessage().isAction();
        break;
    case TopMessageActionTypeRole:
        result = dialogItem.topMessage().actionType();
        break;
    case TopMessageActionTitleRole:
        result = dialogItem.topMessage().actionTitle();
        break;
    case TopMessageActionUserRole:
        result = dialogItem.topMessage().actionUser();
        break;
    }
    return result;
}

void DialogsModel::updateQuery() {

    mState = Initializing;

    if (!mDialogs.isEmpty()) {
        mDlgIdsMap.clear();
        mTopMessageIdsMap.clear();
        beginRemoveRows(QModelIndex(), 0, mDialogs.count() - 1);
        mDialogs.clear();
        endRemoveRows();
        mSecretDialogsCount = 0;
    }

    mCanFetchMore = true;

    if (!mTelegramClient) {
        qCWarning(TG_PLUGIN_LOGIC) << "Telegram client is not ready for creating data view";
        return;
    }

    mTelegramClient->initializeDialogsView();

    fetchMore();
}

void DialogsModel::initialize() {

    switch (mState) {
    case Created:
        updateQuery();
        break;
    case Initializing:
        break;
    case Ready:
        if (regularDialogsCount() == 0) {
            fetchMore();
        }
        break;
    }
}

bool DialogsModel::canFetchMore(const QModelIndex &parent) const
{
    if (parent.isValid()) {
        qCDebug(TG_PLUGIN_LOGIC) << "fetch more: false";
        return false;
    }

    qCDebug(TG_PLUGIN_LOGIC) << "fetch more:" << mCanFetchMore;

    return mCanFetchMore;
}

void DialogsModel::fetchMore(const QModelIndex &parent)
{
    if (parent.isValid() || !mTelegramClient) {
        return;
    }

    mTelegramClient->fetchDialogsNextPage();
}

void DialogsModel::onNextPage(const QList<DialogItem> &dialogs) {

    qCDebug(TG_PLUGIN_LOGIC) << "Got dialogs:" << dialogs.count();

    if (dialogs.isEmpty() && mCanFetchMore) {
        // view (db) dialogs exhausted, query server for more dialogs
        if (!mTelegramClient) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get dialogs. Telegram client reference is null";
        } else if (!mTelegramClient->authenticated()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get dialogs. Telegram client reference is not logged in";
        } else if (!mTelegramClient->connected()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get dialogs. Telegram client reference is not connected";
        } else {
            mTelegramClient->messagesGetDialogs(regularDialogsCount(), 0, 20);
        }
    } else {
        insertOrUpdateDialogs(dialogs);
    }

    mState = Ready;
}

void DialogsModel::append(DialogItem &dialog) {
    beginInsertRows(QModelIndex(), mDialogs.count(), mDialogs.count());
    mDlgIdsMap.insert(dialog.id(), mDialogs.count());
    mTopMessageIdsMap.insert(dialog.topMessage().id(), dialog.id());
    mDialogs.append(dialog);
    endInsertRows();

    if (dialog.isSecret()) {
        mSecretDialogsCount++;
    }
}

void DialogsModel::clear() {
    beginResetModel();
    mDlgIdsMap.clear();
    mTopMessageIdsMap.clear();
    mDialogs.clear();
    mSecretDialogsCount = 0;
    endResetModel();
}

void DialogsModel::resetModel() {
    disconnect(mTelegramClient);
    mTelegramClient = 0;
    mState = Created;
    mCanFetchMore = true;
    clear();
}

void DialogsModel::insertOrUpdateDialogs(const QList<DialogItem> &dialogs) {

    QList<DialogItem> dialogsToAppend;
    Q_FOREACH (DialogItem d, dialogs) {
        qint32 pos = mDlgIdsMap.value(d.id(), -1);
        if (pos >= 0) {
            mDialogs.replace(pos, d);
            Q_EMIT dataChanged(index(pos), index(pos));
            break;
        } else {
            dialogsToAppend.append(d);
        }
    }

    beginInsertRows(QModelIndex(), mDialogs.count(), mDialogs.count() + dialogsToAppend.count() - 1);
    Q_FOREACH (DialogItem newDialog, dialogsToAppend) {
        mDlgIdsMap.insert(newDialog.id(), mDialogs.count());
        mTopMessageIdsMap.insert(newDialog.topMessage().id(), newDialog.id());
        mDialogs << newDialog;

        if (newDialog.isSecret()) {
            mSecretDialogsCount++;
        }
    }
    endInsertRows();
}

// SLOTS
void DialogsModel::onDialogsAdded(const QList<DialogItem> &dialogs, qint32 totalCount) {

    if (!dialogs.count()) {
        mCanFetchMore = false;
        return;
    }

    insertOrUpdateDialogs(dialogs);

    mCanFetchMore = totalCount > 0 && totalCount > mDialogs.count();
}

qint32 DialogsModel::regularDialogsCount() {
    return mDialogs.count() - mSecretDialogsCount;
}

void DialogsModel::onThumbnailUpdated(qint32 id, const QString &thumbnail) {
    qint32 pos = mDlgIdsMap.value(id, -1);
    if (pos >= 0) {
        mDialogs[pos].setThumbnail(thumbnail);
        const QVector<qint32> changedRoles(QVector<qint32>() << ThumbnailRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onTitleUpdated(qint32 id, const QString &title) {
    qint32 pos = mDlgIdsMap.value(id, -1);
    if (pos >= 0) {
        mDialogs[pos].setName(title);
        const QVector<qint32> changedRoles(QVector<qint32>() << NameRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onMessageAdded(const MessageItem &message) {

    qint32 dialogId = message.dialogId();
    qint32 pos = mDlgIdsMap.value(dialogId, -1);

    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dlg = mDialogs[pos];

        if (!message.out()) {
            dlg.setTyping(false);
        }

        //Remove previous top message id from mapping
        MessageItem oldTopMessage = dlg.topMessage();
        if (!oldTopMessage.isEmpty()) {
            mTopMessageIdsMap.remove(oldTopMessage.id());
        }

        dlg.setTopMessage(message);
        dlg.setDate(message.date());
        mTopMessageIdsMap.insert(message.id(), dlg.id());

        if (!message.out() && message.unread()) {
            dlg.setUnreadCount(dlg.unreadCount() + 1);
        }

        const QVector<qint32> changedRoles(QVector<qint32>() << UnreadCountRole << DateRole << TopMessageIdRole << TopMessageFromIdRole <<
                                           TopMessageFromFirstNameRole << TopMessageFromLastNameRole << TopMessageFromPhoneRole <<
                                           TopMessageToIdRole << TopMessageOutRole << TopMessageUnreadRole << TopMessageSentRole <<
                                           TopMessageDateRole << TopMessageTextRole << TopMessageMediaTypeRole <<
                                           TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole <<
                                           TypingRole << WhoIsTypingRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "onMessageAdded: dialog not found for new top message";
    }
}

void DialogsModel::onMessagesAdded(const QList<MessageItem> &messages, qint32 totalCount) {

    if (!messages.count()) {
        return;
    }

    //Get message related model. Check only first one, because all them belongs to same dialog
    const MessageItem &message = messages.at(0);
    qint32 dialogId = message.dialogId();
    qint32 pos = mDlgIdsMap.value(dialogId, -1);

    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dlg = mDialogs[pos];

        Q_FOREACH (MessageItem msg, messages) {

            if (msg.date() >= dlg.date()) {
                //Remove previous top message id from mapping
                MessageItem oldTopMessage = dlg.topMessage();
                if (!oldTopMessage.isEmpty()) {
                    mTopMessageIdsMap.remove(oldTopMessage.id());
                }

                dlg.setTopMessage(msg);
                dlg.setDate(msg.date());
                mTopMessageIdsMap.insert(msg.id(), dlg.id());

                const QVector<qint32> changedRoles(QVector<qint32>() << DateRole << TopMessageIdRole << TopMessageFromIdRole <<
                                                   TopMessageFromFirstNameRole << TopMessageFromLastNameRole <<
                                                   TopMessageFromPhoneRole << TopMessageToIdRole << TopMessageOutRole <<
                                                   TopMessageUnreadRole << TopMessageSentRole << TopMessageDateRole <<
                                                   TopMessageTextRole << TopMessageMediaTypeRole <<
                                                   TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole);
                Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
            }
        }
    }
}

void DialogsModel::onMessageUpdated(qint32 msgId, const MessageItem &message) {

    qint32 dialogId = mTopMessageIdsMap.take(msgId);
    if (dialogId) {
        qint32 pos = mDlgIdsMap.value(dialogId, -1);

        if (0 <= pos && pos < mDialogs.size()) {

            DialogItem &dlg = mDialogs[pos];
            dlg.setTopMessage(message);
            dlg.setDate(message.date());
            mTopMessageIdsMap.insert(message.id(), dlg.id());

            const QVector<qint32> changedRoles(QVector<qint32>() << UnreadCountRole << DateRole << TopMessageIdRole << TopMessageFromIdRole <<
                                               TopMessageFromFirstNameRole << TopMessageFromLastNameRole << TopMessageFromPhoneRole <<
                                               TopMessageToIdRole << TopMessageOutRole << TopMessageUnreadRole << TopMessageSentRole <<
                                               TopMessageDateRole << TopMessageTextRole << TopMessageMediaTypeRole <<
                                               TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole);
            Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
        }
    }
}

void DialogsModel::updateThumbnail(qint32 dialogId, QString filePath) {
    // update thumbnail path for affected dialogsItem
    qint32 pos = mDlgIdsMap.value(dialogId, -1);
    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &d = mDialogs[pos];
        d.setThumbnail(filePath);
        const QVector<qint32> changedRoles(QVector<qint32>() << ThumbnailRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onOutgoingMessageAdded(qint32 recipientId, const MessageItem &message) {
    qint32 pos = mDlgIdsMap.value(recipientId, -1);
    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dialog = mDialogs[pos];
        MessageItem messageItem(message);
        dialog.setTopMessage(messageItem);
        dialog.setDate(messageItem.date());

        mTopMessageIdsMap.insert(messageItem.id(), dialog.id());

        const QVector<qint32> changedRoles(QVector<qint32>() << DateRole << TopMessageIdRole << TopMessageFromIdRole <<
                                           TopMessageFromFirstNameRole << TopMessageFromLastNameRole << TopMessageFromPhoneRole <<
                                           TopMessageToIdRole << TopMessageOutRole << TopMessageUnreadRole << TopMessageDateRole <<
                                           TopMessageSentRole << TopMessageTextRole << TopMessageMediaTypeRole <<
                                           TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onChatHistoryDeleted(qint32 dialogId) {

    qint32 pos = mDlgIdsMap.value(dialogId, -1);
    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dialog = mDialogs[pos];
        qint32 oldTopMessageId = dialog.topMessage().id();

        MessageItem emptyMessage;
        dialog.setTopMessage(emptyMessage);

        mTopMessageIdsMap.remove(oldTopMessageId);

        const QVector<qint32> changedRoles(QVector<qint32>() << DateRole << TopMessageIdRole << TopMessageFromIdRole <<
                                           TopMessageFromFirstNameRole << TopMessageFromLastNameRole << TopMessageFromPhoneRole <<
                                           TopMessageToIdRole << TopMessageOutRole << TopMessageUnreadRole << TopMessageDateRole <<
                                           TopMessageSentRole << TopMessageTextRole << TopMessageMediaTypeRole <<
                                           TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onChatDeleted(qint32 dialogId) {
    deleteDialog(dialogId);
}

void DialogsModel::onChatMarkedAsRead(qint32 dialogId) {
    qint32 pos = mDlgIdsMap.value(dialogId, -1);
    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dialog = mDialogs[pos];

        QVector<qint32> changedRoles(QVector<qint32>() << UnreadCountRole);

        if (!dialog.topMessage().out() && dialog.topMessage().unread()) {
            dialog.topMessage().setUnread(false);
            changedRoles.append(TopMessageUnreadRole);
        }
        dialog.setUnreadCount(0);

        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onMessagesMarkedAsRead(const QList<qint32> &msgIds) {

    if (!msgIds.count()) {
        return;
    }

    // let's suppose all received messages belongs to same dialog

    //FIXME is supposed that only querying first messages is enough to know the dialogId,
    //but while not implemented offline updates, if a message is received when the app is
    //offline, that message is not inserted in db, but could be included in this list, with
    //the result of having a dialog with id == -1. This workaroud searches for first msgId
    //belonging to a valid dialog
    qint32 dialogId = -1;
    int i = 0;
    int count = msgIds.count();
    while (dialogId == -1 && i < count) {
        dialogId = mTelegramClient->getDialogIdFromMessageId(msgIds.at(i));
        i++;
    }

    if (dialogId != 0) {
        qint32 pos = mDlgIdsMap.value(dialogId, -1);
        if (0 <= pos && pos < mDialogs.size()) {
            DialogItem &dialog = mDialogs[pos];

            QVector<qint32> changedRoles(QVector<qint32>() << UnreadCountRole);
            if (dialog.topMessage().unread()) {
                dialog.topMessage().setUnread(false);
                changedRoles.append(TopMessageUnreadRole);
            }

            // let's suppose all not previously read messages in the dialog has been set to read at the same time
            dialog.setUnreadCount(0);

            Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
        }
    }
}

void DialogsModel::onMessagesDeleted(const QList<qint32> msgIds) {

    if (!msgIds.count()) {
        return;
    }

    Q_FOREACH (qint32 msgId, msgIds) {
        qint32 dialogId = mTopMessageIdsMap.value(msgId, 0);

        if (dialogId != 0) {
            // remove msgId as top message cos we're going to substitute it by the current one
            mTopMessageIdsMap.remove(msgId);
            // once deleted top message, we must ask Data() for a new top message (it will be an empty object if not found any)
            MessageItem newTopMessage = mTelegramClient->topMessage(dialogId);

            qint32 pos = mDlgIdsMap.value(dialogId);

            mDialogs[pos].setTopMessage(newTopMessage);
            if (!newTopMessage.isEmpty()) {
                mTopMessageIdsMap.insert(newTopMessage.id(), dialogId);
            }

            const QVector<qint32> changedRoles(QVector<qint32>() << TopMessageIdRole << TopMessageFromIdRole <<
                                               TopMessageFromFirstNameRole << TopMessageFromLastNameRole << TopMessageFromPhoneRole <<
                                               TopMessageToIdRole << TopMessageOutRole << TopMessageUnreadRole << TopMessageSentRole <<
                                               TopMessageDateRole << TopMessageTextRole << TopMessageMediaTypeRole <<
                                               TopMessageActionTitleRole << TopMessageActionTypeRole << TopMessageActionUserRole);
            Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
        }
    }

    // regardless of previous top messages updates, all the unreadCount for all dialogs must be updated, because
    // a dialog non top message can be deleted and if it was unread, this counter must be decreased. This data must
    // be provided by data instance
    for (int i = 0; i < mDialogs.count(); i++) {
        DialogItem &dialog = mDialogs[i];
        qint32 unreadCount = mTelegramClient->getUnreadCount(dialog.id());
        dialog.setUnreadCount(unreadCount);

        QVector<qint32> changedRoles(QVector<qint32>() << UnreadCountRole);
        Q_EMIT dataChanged(index(i), index(i), changedRoles);
    }
}

void DialogsModel::onContactsAdded(const QList<ContactItem> &contacts) {
    // When added a contact, if there is any dialog related, change its peer type
    Q_FOREACH (const ContactItem &contact, contacts) {
        qint32 pos = mDlgIdsMap.value(contact.id(), -1);
        if (0 <= pos && pos < mDialogs.size()) {
            DialogItem &dialog = mDialogs[pos];
            dialog.setPeerType(contact.classType());
        }
    }
}

void DialogsModel::onUserTyping(qint32 dialogId, const QString &whoIsTyping) {
    qint32 pos = mDlgIdsMap.value(dialogId, -1);
    if (0 <= pos && pos < mDialogs.size()) {
        DialogItem &dialog = mDialogs[pos];

        connect(&dialog, SIGNAL(typingFinished()), this, SLOT(onUserTypingFinished()), Qt::UniqueConnection);
        dialog.setWhoIsTyping(whoIsTyping);
        dialog.setTyping(true);

        const QVector<qint32> changedRoles(QVector<qint32>() << TypingRole << WhoIsTypingRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::onUserTypingFinished() {
    DialogItem &dialogItem = *qobject_cast<DialogItem *>(sender());
    qint32 pos = mDlgIdsMap.value(dialogItem.id(), -1);
    if (0 <= pos && pos < mDialogs.size()) {
        const QVector<qint32> changedRoles(QVector<qint32>() << TypingRole << WhoIsTypingRole);
        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void DialogsModel::deleteDialog(qint32 dialogId) {

    qint32 pos = mDlgIdsMap.value(dialogId, -1);
    if (0 <= pos && pos < mDialogs.size()) {

        if (mDialogs.at(pos).isSecret()) {
            mSecretDialogsCount--;
        }

        mDlgIdsMap.clear();
        beginRemoveRows(QModelIndex(), pos, pos);
        mDialogs.removeAt(pos);
        endRemoveRows();

        // rebuild dialog->id() --> index map
        for (int i = 0; i < mDialogs.count(); i++) {
            const DialogItem &dlg = mDialogs.at(i);
            mDlgIdsMap.insert(dlg.id(), i);
        }
    }
}
