/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DIALOGSMODEL_H
#define DIALOGSMODEL_H

#include <QAbstractListModel>
#include <QVector>
#include "dialogitem.h"
#include "dialogsview.h"

#include "telegramclient.h"

class DialogsModel : public QAbstractListModel
{
    Q_OBJECT
    Q_ENUMS(DialogRoles)
public:

    enum ModelState {
        Created,
        Initializing,
        Ready
    };

    enum DialogRoles {
        IdRole = Qt::UserRole,
        IsChatRole,
        PeerIdRole,
        PeerTypeRole,
        PeerAccessHashRole,
        NameRole,
        UnreadCountRole,
        ThumbnailRole,
        TypingRole,
        WhoIsTypingRole,
        MembersRole,
        MembersOnlineRole,
        UserOnlineRole,
        LastSeenOnlineRole,
        DateRole,
        IsSecretRole,
        StateRole,
        TopMessageIdRole,
        TopMessageFromIdRole,
        TopMessageFromFirstNameRole,
        TopMessageFromLastNameRole,
        TopMessageFromPhoneRole,
        TopMessageToIdRole,
        TopMessageOutRole,
        TopMessageUnreadRole,
        TopMessageSentRole,
        TopMessageDateRole,
        TopMessageTextRole,
        TopMessageMediaTypeRole,
        TopMessageIsActionRole,
        TopMessageActionTypeRole,
        TopMessageActionTitleRole,
        TopMessageActionUserRole
    };

    explicit DialogsModel(QObject *parent = 0);
    ~DialogsModel();

    // dialog items
    void append(DialogItem &dialog);
    void clear();

    // updates
    void updateThumbnail(qint32 dialogId, QString filePath);

    // Overriding
    QHash<qint32, QByteArray> roleNames() const;
    qint32 rowCount(const QModelIndex &parent = QModelIndex()) const;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;

    Q_INVOKABLE bool canFetchMore(const QModelIndex &parent = QModelIndex()) const;
    Q_INVOKABLE void fetchMore(const QModelIndex &parent = QModelIndex());

    Q_INVOKABLE QVariant get(int row);
    Q_INVOKABLE QVariant getDialogFromDialogId(int dialogId);

    Q_INVOKABLE void setTelegramClient(TelegramClient *telegramClient);

private:
    QList<DialogItem> mDialogs;
    // correlates dialog->id() --> index in m_dialogs
    QMap<qint32, qint32> mDlgIdsMap;
    // correlates dialog->topMessage()->id() --> dialog->id()
    QMap<qint32, qint32> mTopMessageIdsMap;

    bool mCanFetchMore;
    ModelState mState;
    qint32 mSecretDialogsCount;

    // reference to telegram client
    TelegramClient *mTelegramClient;

    void deleteDialog(qint32 dialogId);
    void insertOrUpdateDialogs(const QList<DialogItem> &dialogs);
    qint32 regularDialogsCount();

private Q_SLOTS:
    void onNextPage(const QList<DialogItem> &dialogs);
    void updateQuery();
    // this slot is called when logged in and dialogs model hasn't any data. It is needed, then call fetch more to obtain
    // data from server
    void initialize();
    void resetModel();


    void onDialogsAdded(const QList<DialogItem> &dialogs, qint32 totalCount);
    void onThumbnailUpdated(qint32 id, const QString &thumbnail);
    void onTitleUpdated(qint32 id, const QString &title);

    void onMessageAdded(const MessageItem &message);
    void onMessagesAdded(const QList<MessageItem> &messages, qint32 totalCount);
    void onOutgoingMessageAdded(qint32 recipientId, const MessageItem &message);
    void onMessageUpdated(qint32 msgId, const MessageItem &message);
    void onMessagesDeleted(const QList<qint32> msgIds);
    void onMessagesMarkedAsRead(const QList<qint32> &msgIds);
    void onChatMarkedAsRead(qint32 dialogId);
    void onChatHistoryDeleted(qint32 dialogId);
    void onChatDeleted(qint32 dialogId);

    void onContactsAdded(const QList<ContactItem> &contacts);

    void onUserTyping(qint32 dialogId, const QString &whoIsTyping);
    void onUserTypingFinished();
};

#endif // DIALOGSMODEL_H
