/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "models/groupmembersmodel.h"
#include "types/tluser.h"
#include "types/tlinputpeer.h"

GroupMembersModel::GroupMembersModel(QObject *parent) :
    Model(parent), mTelegramClient(NULL), mChatId(0) {
}

GroupMembersModel::~GroupMembersModel() {
    mTelegramClient = NULL;
}

void GroupMembersModel::setup(TelegramClient *telegramClient, qint32 chatId) {
    Q_ASSERT(telegramClient != NULL);

    if (mTelegramClient == NULL) {
        mTelegramClient = telegramClient;
        connect(mTelegramClient, SIGNAL(chatUpdated(qint32)),
                           this, SLOT(refresh(qint32)));
    }
    mChatId = chatId;
    refresh(mChatId);
}

void GroupMembersModel::refresh(qint32 chatId) {
    if (mChatId != chatId || mTelegramClient == NULL) return;
    /*
     * Member is removalbe from chat if one of the conditions is met:
     *  - we are the chat admin
     *  - we have invited that member
     */
    mQueryString = QString("SELECT id, "
                           "    (firstName || \" \" || lastName) AS name, "
                           "    (SELECT localPath FROM fileLocations WHERE rowid=thumbnail) AS photo, "
                           "    lastSeenOnline, online, isAdmin, "
          /* admin */      "    (SELECT CASE WHEN EXISTS"
                           "        (SELECT isAdmin FROM participants "
                           "         WHERE participants.chatId = %1 "
                           "           AND participants.userId = %2 "
                           "           AND participants.isAdmin = 1) "
          /* or inviter */ "     THEN 1 ELSE 0 END) OR inviterId = %2 "
          /* or me */      "    OR userId = %2 AS isRemovable, "
                           "    (SELECT type FROM users WHERE users.id=participants.userId) AS peerType "
                           "FROM users, participants "
                           "WHERE participants.chatId = %1 AND users.id=participants.userId "
                           "ORDER BY online DESC, lastSeenOnline DESC")
            .arg(chatId)
            .arg(mTelegramClient->ourId());
    Model::refresh();
}

qint32 GroupMembersModel::getInputPeerType(qint32 peerType) {
    switch (peerType) {
    case User::typeUserContact:
        return InputPeer::typeInputPeerContact;
    case User::typeUserRequest:
        // FALL-THROUGH
    case User::typeUserForeign:
        return InputPeer::typeInputPeerForeign;
        break;
    case User::typeUserDeleted:
        // FALL-THROUGH
    case User::typeUserEmpty:
        return InputPeer::typeInputPeerEmpty;
    case User::typeUserSelf:
        return InputPeer::typeInputPeerSelf;
    }
    return InputPeer::typeInputPeerEmpty;
}

QHash<int, QByteArray> GroupMembersModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[NameRole] = "name";
    roles[PhotoRole] = "photo";
    roles[LastSeenOnlineRole] = "lastSeenOnline";
    roles[OnlineRole] = "online";
    roles[IsAdminRole] = "isAdmin";
    roles[IsRemovableRole] = "isRemovable";
    roles[PeerTypeRole] = "peerType";
    return roles;
}
