/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "groupmodel.h"

GroupModel::GroupModel(QObject *parent) : QObject(parent),
    mTelegramClient(NULL), mChatId(0), mTitle(""), mPhoto(""), mCount(0), mMuted(false) {
}

GroupModel::~GroupModel() {
    mTelegramClient = 0;
}

void GroupModel::setup(TelegramClient *telegramClient, qint32 chatId) {
    Q_ASSERT(telegramClient != NULL);

    if (mTelegramClient == NULL) {
        mTelegramClient = telegramClient;
        // TODO karni: later -- receive title changed update, avoid refreshing participants
        connect(mTelegramClient, SIGNAL(chatUpdated(qint32)),
                this, SLOT(refreshGroupDetails(qint32)));
        connect(mTelegramClient, SIGNAL(dialogThumbnailUpdated(qint32, QString)),
                this, SLOT(chatThumbnailUpdated(qint32, QString)));
        connect(mTelegramClient, SIGNAL(dialogMuted(qint32,bool)),
                this, SLOT(chatMuted(qint32,bool)));
        connect(mTelegramClient, SIGNAL(dialogsMuted(NotifyPeer::NotifyPeerType,bool)),
                this, SLOT(chatsMuted(NotifyPeer::NotifyPeerType,bool)));
    }

    if (mTelegramClient != NULL) {
        setChatId(chatId);
    }
}

qint32 GroupModel::getChatId() const {
    return mChatId;
}

void GroupModel::setChatId(qint32 chatId) {
    if (mChatId != chatId) {
        mChatId = chatId;
        refreshGroupDetails(chatId);
        Q_EMIT chatIdChanged(chatId);
        mTelegramClient->messagesGetFullChat(chatId);
    }
}

QString GroupModel::getTitle() const {
    return mTitle;
}

void GroupModel::setTitle(const QString &title) {
    if (mTitle != title) {
        mTitle = title;
        Q_EMIT titleChanged(title);
    }
}

QString GroupModel::getThumbnail() const {
    return mThumbnail;
}

void GroupModel::setThumbnail(const QString &thumbnail) {
    if (mThumbnail != thumbnail) {
        mThumbnail = thumbnail;
        Q_EMIT thumbnailChanged(thumbnail);
    }
}

QString GroupModel::getPhoto() const {
    return mPhoto;
}

void GroupModel::setPhoto(const QString &photo) {
    if (mPhoto != photo) {
        mPhoto = photo;
        Q_EMIT photoChanged(photo);
    }
}

qint32 GroupModel::getCount() const {
    return mCount;
}

void GroupModel::setCount(const qint32 count) {
    if (mCount != count) {
        mCount = count;
        Q_EMIT countChanged(count);
    }
}

bool GroupModel::getMuted() const {
    return mMuted;
}

void GroupModel::setMuted(bool muted) {
    if (mMuted != muted) {
        mMuted = muted;
        Q_EMIT mutedChanged(muted);
    }
}

void GroupModel::resetGroupDetails() {
    mChatId = 0;
    setTitle(QString());
    setThumbnail("");
    setPhoto("");
    setCount(0);
    setMuted(false);
}

void GroupModel::refreshGroupDetails(qint32 chatId) {
    if (mChatId != chatId) return;
    qCDebug(TG_PLUGIN_LOGIC) << "Refreshing group chat profile model for chat:" << chatId;

    QSqlQuery query;
    query.prepare(QString("SELECT title, "
                          "(SELECT localPath FROM fileLocations WHERE fileLocations.rowid=thumbnail) AS thumbnail, "
                          "(SELECT localPath FROM fileLocations WHERE fileLocations.rowid=bigPhoto) AS photo, "
                          "participantsCount AS count, muted "
                          "FROM chats JOIN dialogs ON chats.id = dialogs.id "
                          "WHERE dialogs.id = :id"));
    query.bindValue(":id", chatId);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Failed to query group details for chat id"
                << chatId << query.lastError() << query.lastQuery();
        resetGroupDetails();
    } else if (query.next()) {
        QString title = query.value("title").toString();
        setTitle(title);
        QString thumbnail = query.value("thumbnail").toString();
        setThumbnail(thumbnail);
        QString photo = query.value("photo").toString();
        setPhoto(photo);
        qint32 count = query.value("count").toInt();
        setCount(count);
        bool muted = query.value("muted").toBool();
        setMuted(muted);

        Q_EMIT groupDetailsRefreshed();
    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "No group chat details for chat:" << chatId;
        resetGroupDetails();
        return;
    }
}

void GroupModel::chatThumbnailUpdated(qint32 chatId, QString thumbnail) {
    if (mChatId != chatId) return;
    setThumbnail(thumbnail);
}

void GroupModel::chatMuted(qint32 chatId, bool muted) {
    if (mChatId != chatId) return;
    setMuted(muted);
}

void GroupModel::chatsMuted(NotifyPeer::NotifyPeerType notifyType, bool muted) {
    if (notifyType == NotifyPeer::typeNotifyAll || notifyType == NotifyPeer::typeNotifyChats) {
        setMuted(muted);
    }
}
