/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "messageitem.h"
#include "data.h"

MessageItem::MessageItem(QObject *parent) :
    TLMessage(parent),
    mDownloading(false),
    mDownloadedPercentage(0),
    mSent(true) {
}

MessageItem::MessageItem(const MessageItem &other) {
    *this = other;
}

MessageItem &MessageItem::operator=(const MessageItem &messageItem) {

    const TLMessage *tlMessage = static_cast<const TLMessage*>(&messageItem);
    setId(tlMessage->id());
    setDialogId(messageItem.dialogId());
    setFromId(tlMessage->fromId());
    setToId(*tlMessage->toId());
    setOut(tlMessage->out());
    setUnread(tlMessage->unread());
    setTtl(messageItem.ttl());
    setDate(tlMessage->date());
    setMessage(tlMessage->message());
    setMedia(*tlMessage->media());
    setThumbnail(messageItem.thumbnail());
    setPhoto(messageItem.photo());
    setVideo(messageItem.video());
    setDownloading(messageItem.downloading());
    setDownloadedPercentage(messageItem.downloadedPercentage());
    setFromFirstName(messageItem.fromFirstName());
    setFromLastName(messageItem.fromLastName());
    setFromPhone(messageItem.fromPhone());
    setFromThumbnail(messageItem.fromThumbnail());
    setFromUserType(messageItem.fromUserType());
    setSent(messageItem.sent());
    setAction(*tlMessage->action());
    setActionUser(messageItem.actionUser());
    return *this;
}

bool MessageItem::operator==(const MessageItem &other) const {
    return this->id() == other.id();
}

MessageItem::MessageItem(const Message &msg, const ContactItem &sender, bool sent) {

    qint32 dialogId = msg.fromId();
    bool isChat = msg.toId().classType() == TLPeer::typePeerChat;
    if (isChat) {
        dialogId = msg.toId().chatId();
    } else if (msg.out()) {
        dialogId = msg.toId().userId();
    }

    setId(msg.id());
    setDialogId(dialogId);
    setFromId(msg.fromId());
    setToId(msg.toId());
    setOut(msg.out());
    setUnread(msg.unread());
    setTtl(msg.ttl());
    setDate(msg.date());
    setMessage(msg.message());
    setMedia(msg.media());
    setFromFirstName(sender.firstName());
    setFromLastName(sender.lastName());
    setFromPhone(sender.phone());
    setFromThumbnail(sender.thumbnail());
    setFromUserType(sender.classType());
    setSent(sent);
    setAction(msg.action());
}

MessageItem::~MessageItem() {
}

qint32 MessageItem::dialogId() const {
    return mDialogId;
}

QString MessageItem::fromFirstName() const {
    return mFromFirstName;
}
QString MessageItem::fromLastName() const {
    return mFromLastName;
}
QString MessageItem::fromPhone() const {
    return mFromPhone;
}

QString MessageItem::fromThumbnail() const {
    return mFromThumbnail;
}
qint32 MessageItem::fromUserType() const {
    return mFromUserType;
}

qint32 MessageItem::toId() const {
    if (toIsChat()) {
        return TLMessage::toId()->chatId();
    }
    return TLMessage::toId()->userId();
}

bool MessageItem::toIsChat() const {
    return TLMessage::toId()->classType() == TLPeer::typePeerChat;
}

QString MessageItem::text() const {
    return this->message();
}

TLMessageMedia::MessageMediaType MessageItem::mediaType() const {
    return TLMessage::media()->classType();
}

QString MessageItem::thumbnail() const {
    if (!mThumbnail.length()) {
        return mPhoto;
    }
    return mThumbnail;
}

QString MessageItem::photo() const {
    if (!mPhoto.length()) {
        return mThumbnail;
    }
    return mPhoto;
}

QString MessageItem::photoSizeType() const {
    return mPhotoSizeType;
}

QString MessageItem::video() const {
    return mVideo;
}

bool MessageItem::downloading() const {
    return mDownloading;
}

qint8 MessageItem::downloadedPercentage() const {
    return mDownloadedPercentage;
}

bool MessageItem::isAction() const {
    return action() && action()->classType() != MessageAction::typeMessageActionEmpty;
}

bool MessageItem::sent() const {
    return mSent;
}

qint32 MessageItem::actionType() const {
    return action() ? action()->classType() : MessageAction::typeMessageActionEmpty;
}

QString MessageItem::actionTitle() const {
    return action() ? action()->title() : "";
}

QString MessageItem::actionUser() const {
    return mActionUser;
}

void MessageItem::setDialogId(qint32 dialogId) {
    mDialogId = dialogId;
}

void MessageItem::setThumbnail(const QString &thumbnail) {
    mThumbnail = thumbnail;
}

void MessageItem::setPhoto(const QString &photo) {
    mPhoto = photo;
}

void MessageItem::setPhotoSizeType(const QString &photoSizeType) {
    mPhotoSizeType = photoSizeType;
}

void MessageItem::setVideo(const QString &video) {
    mVideo = video;
}

void MessageItem::setDownloading(bool downloading) {
    mDownloading = downloading;
}

void MessageItem::setDownloadedPercentage(qint8 downloadedPercentage) {
    mDownloadedPercentage = downloadedPercentage;
}

void MessageItem::setFromFirstName(const QString &firstName) {
    mFromFirstName = firstName;
}

void MessageItem::setFromLastName(const QString &lastName) {
    mFromLastName = lastName;
}

void MessageItem::setFromPhone(const QString &phone) {
    mFromPhone = phone;
}

void MessageItem::setFromThumbnail(const QString &fromThumb) {
    mFromThumbnail = fromThumb;
}

void MessageItem::setFromUserType(qint32 fromUserType) {
    mFromUserType = fromUserType;
}

void MessageItem::setSender(const ContactItem &sender) {
    setFromFirstName(sender.firstName());
    setFromLastName(sender.lastName());
    setFromPhone(sender.phone());
    setFromThumbnail(sender.thumbnail());
}

void MessageItem::setSent(bool sent) {
    mSent = sent;
}

void MessageItem::setActionUser(const QString &actionUser) {
    mActionUser = actionUser;
}

bool MessageItem::isEmpty() const {
    return id() <= 0;
}

void MessageItem::dump() {
    qCDebug(TG_PLUGIN_LOGIC) << "ID" << id();
    qCDebug(TG_PLUGIN_LOGIC) << "DIALOGID" << dialogId();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMID" << fromId();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMFIRSTNAME" << fromFirstName();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMLASTNAME" << fromLastName();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMPHONE" << fromPhone();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMTHUMBNAIL" << fromThumbnail();
    qCDebug(TG_PLUGIN_LOGIC) << "FROMUSERTYPE" << fromUserType();
    qCDebug(TG_PLUGIN_LOGIC) << "TOID" << toId();
    qCDebug(TG_PLUGIN_LOGIC) << "TOISCHAT" << toIsChat();
    qCDebug(TG_PLUGIN_LOGIC) << "OUT" << out();
    qCDebug(TG_PLUGIN_LOGIC) << "UNREAD" << unread();
    qCDebug(TG_PLUGIN_LOGIC) << "DATE" << date();
    qCDebug(TG_PLUGIN_LOGIC) << "TTL" << ttl();
    qCDebug(TG_PLUGIN_LOGIC) << "TEXT" << text();
    qCDebug(TG_PLUGIN_LOGIC) << "MEDIA" << media()->classType();
    qCDebug(TG_PLUGIN_LOGIC) << "THUMBNAIL" << mThumbnail;
    qCDebug(TG_PLUGIN_LOGIC) << "PHOTO" << mPhoto;
    qCDebug(TG_PLUGIN_LOGIC) << "VIDEO" << mVideo;
    qCDebug(TG_PLUGIN_LOGIC) << "DOWNLOADING" << mDownloading;
    qCDebug(TG_PLUGIN_LOGIC) << "DOWNLOADED PERCENTAGE" << mDownloadedPercentage;
    qCDebug(TG_PLUGIN_LOGIC) << "ISACTION" << isAction();
    qCDebug(TG_PLUGIN_LOGIC) << "SENT" << sent();
    qCDebug(TG_PLUGIN_LOGIC) << "ACTIONTYPE" << actionType();
    qCDebug(TG_PLUGIN_LOGIC) << "ACTIONTITLE" << actionTitle();
    qCDebug(TG_PLUGIN_LOGIC) << "ACTIONUSER" << actionUser();
}
