/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QDateTime>
#include "messagesmodel.h"
#include "data.h"
#include "dbmanager.h"
#include "rawapiclient.h"
#include "tools.h"

MessagesModel::MessagesModel(QObject *parent) :
    QAbstractListModel(parent),
    DOWNLOADED_PERCENTAGE_CHANGED_ROLES(QVector<qint32>() << MessagesModel::DownloadedPercentageRole << MessagesModel::DownloadingRole),
    mDialogId(0),
    mPeerType(0),
    mAccessHash(0),
    mIsSecret(false),
    mCanFetchMore(true),
    mTelegramClient(0) {

    connect(this, SIGNAL(rowsInserted(QModelIndex,int,int)), this, SIGNAL(countIncreased()));
}

MessagesModel::~MessagesModel() {
}

qint32 MessagesModel::dialogId() const {
    return mDialogId;
}

qint32 MessagesModel::peerType() const {
    return mPeerType;
}

bool MessagesModel::isSecret() const {
    return mIsSecret;
}

qint32 MessagesModel::getCount() const {
    return mMessages.count();
}

void MessagesModel::setDialogId(qint32 dialogId) {
    mDialogId = dialogId;
    Q_EMIT dialogIdChanged(dialogId);
}

void MessagesModel::setPeerType(qint32 peerType) {
    mPeerType = peerType;
    Q_EMIT peerTypeChanged(peerType);
}

void MessagesModel::setIsSecret(bool isSecret) {
    mIsSecret = isSecret;
    Q_EMIT isSecretChanged(isSecret);
}

void MessagesModel::setTelegramClient(TelegramClient *telegramClient) {

    if (!mTelegramClient && telegramClient) {

        mTelegramClient = telegramClient;

        connect(mTelegramClient, SIGNAL(authLogOutAnswer(qint64, bool)), SLOT(resetModel()));

        connect(mTelegramClient, SIGNAL(messageAdded(MessageItem)), SLOT(onMessageAdded(MessageItem)));
        connect(mTelegramClient, SIGNAL(messagesAdded(QList<MessageItem>,qint32)), SLOT(onMessagesAdded(QList<MessageItem>,qint32)));
        connect(mTelegramClient, SIGNAL(outgoingMessageAdded(qint32,MessageItem)), SLOT(onOutgoingMessageAdded(qint32,MessageItem)));
        connect(mTelegramClient, SIGNAL(messageUpdated(qint32,MessageItem)), SLOT(onMessageUpdated(qint32,MessageItem)));
        connect(mTelegramClient, SIGNAL(messagesMarkedAsRead(QList<qint32>)), SLOT(onMessagesMarkedAsRead(QList<qint32>)));
        connect(mTelegramClient, SIGNAL(messagesDeleted(QList<qint32>)), SLOT(onMessagesDeleted(QList<qint32>)));
        connect(mTelegramClient, SIGNAL(chatHistoryDeleted(qint32)), SLOT(onChatHistoryDeleted(qint32)));

        connect(mTelegramClient, SIGNAL(messagePhotoUpdated(qint32,QString,bool)), SLOT(onMessagePhotoUpdated(qint32,QString,bool)));
        connect(mTelegramClient, SIGNAL(messageVideoUpdated(qint32,QString)), SLOT(onMessageVideoUpdated(qint32,QString)));
        connect(mTelegramClient, SIGNAL(downloadingUpdated(qint32,bool)), SLOT(onMessageDownloadingUpdated(qint32,bool)));
        connect(mTelegramClient, SIGNAL(downloadedPercentageUpdated(qint32,qint8)), SLOT(onMessageDownloadedPercentageUpdated(qint32,qint8)));

        connect(mTelegramClient, SIGNAL(messagesNextPage(QList<MessageItem>)), SLOT(onNextPage(QList<MessageItem>)));

        connect(mTelegramClient, SIGNAL(contactThumbnailUpdated(qint32,QString)), SLOT(onContactThumbnailUpdated(qint32,QString)));
    }
}

QVariant MessagesModel::get(int rowIndex) const {
    QVariantMap rowData;
    QModelIndex sourceIndex = index(rowIndex, 0);
    if (sourceIndex.isValid()) {
        QHash<int, QByteArray> roles = roleNames();
        Q_FOREACH(int role, roles.keys()) {
            rowData.insert(roles[role], data(sourceIndex, role));
        }
    }
    return rowData;
}

void MessagesModel::updateQuery() {

    if (!mDialogId) {
        return;
    }

    if (!mMessages.isEmpty()) {
        mKeysMap.clear();
        beginRemoveRows(QModelIndex(), 0, mMessages.count() - 1);
        mMessages.clear();
        endRemoveRows();
        Q_EMIT countChanged(0);
    }

    mCanFetchMore = true;

    if (!mTelegramClient) {
        qCWarning(TG_PLUGIN_LOGIC) << "Telegram client is not ready for creating data view";
        return;
    }

    mTelegramClient->initializeMessagesView(mDialogId, mPeerType);
}

void MessagesModel::refresh() {
    updateQuery();
    fetchMore();
}

bool MessagesModel::canFetchMore(const QModelIndex &parent) const {

    if (parent.isValid()) {
        qCDebug(TG_PLUGIN_LOGIC) << "fetch more: false";
        return false;
    }

    qCDebug(TG_PLUGIN_LOGIC) << "fetch more:" << mCanFetchMore;
    return mCanFetchMore;
}

void MessagesModel::fetchMore(const QModelIndex &parent) {

    if (parent.isValid() || !mTelegramClient) {
        return;
    }

    mTelegramClient->fetchMessagesNextPage();
}

void MessagesModel::clear() {
    beginResetModel();
    mKeysMap.clear();
    mMessages.clear();
    endResetModel();
    Q_EMIT countChanged(0);
}

void MessagesModel::resetModel() {
    disconnect(mTelegramClient);
    mTelegramClient = 0;
    mDialogId = 0;
    mPeerType = 0;
    mAccessHash = 0;
    mCanFetchMore = true;
    clear();
}

void MessagesModel::onNextPage(const QList<MessageItem> &messages) {

    qCDebug(TG_PLUGIN_LOGIC) << "Got messages:" << messages.count();

    if (messages.isEmpty() && mCanFetchMore && !mIsSecret) {
        // view (db) messages exhausted, query server for more messages
        if (!mTelegramClient) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get history messages. Telegram client reference is null";
        } else if (!mTelegramClient->authenticated()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get history messages. Telegram client reference is not logged in";
        } else if (!mTelegramClient->connected()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not get history messages. Telegram client reference is not connected";
        } else {

            if (!mPeerType || (mPeerType == TLInputPeer::typeInputPeerForeign && !mAccessHash)) {
                connect(mTelegramClient, SIGNAL(getUserAnswer(QVariant)), SLOT(onGetUserAnswer(QVariant)));
                mTelegramClient->getUser(mDialogId);
            } else {
                getHistoryFromServer();
            }
        }
    } else {
        insertNewMessages(messages);
    }
}

void MessagesModel::insertNewMessages(const QList<MessageItem> &messages) {

    if (!messages.count()) {
        mCanFetchMore = false;
        return;
    }

    QList<MessageItem> filteredMessages;
    Q_FOREACH (const MessageItem &message, messages) {
        if (!exists(message)) {
            filteredMessages << message;
        }
    }

    beginInsertRows(QModelIndex(), mMessages.count(), mMessages.count() + filteredMessages.count() - 1);
    Q_FOREACH (const MessageItem newMessage, filteredMessages) {
        mMessages << newMessage;
        mKeysMap.insert(newMessage.id(), &mMessages[mMessages.count() -1]);
    }
    endInsertRows();
    Q_EMIT countChanged(mMessages.count());
}

void MessagesModel::onGetUserAnswer(QVariant qvariantUser) {

    ContactItem *user = qvariant_cast<ContactItem*>(qvariantUser);

    if (!mPeerType) {
        // calculate inputpeer type from user type
        if (user->isEmpty()) {
            mPeerType = InputPeer::typeInputPeerChat;
        } else {

            if (user->id() != mDialogId) {
                return;
            }

            mPeerType = Tools::toInputPeerType(user->classType());
        }
    }

    if (!mAccessHash && user->accessHash()) {
        mAccessHash = user->accessHash();
    }

    // once the received data is used, we can disconnect this signal-slot for not
    // querying messages history if used getUser for another case when
    // having this model in screen (for instance, when adding a contact)
    disconnect(mTelegramClient, SIGNAL(getUserAnswer(QVariant)), this, SLOT(onGetUserAnswer(QVariant)));

    // once got all needed user data, get history from server
    getHistoryFromServer();
}

void MessagesModel::getHistoryFromServer() {

    TLInputPeer peer;
    peer.setClassType(mPeerType);
    if (mPeerType == TLInputPeer::typeInputPeerChat) {
        peer.setChatId(mDialogId);
    } else {
        peer.setUserId(mDialogId);
    }

    if (mPeerType == TLInputPeer::typeInputPeerForeign) {
        peer.setAccessHash(mAccessHash);
    }

    // retrieve the 20 messages inmediatly before received than the older one
    qint32 olderId = 0;
    if (mMessages.count() > 0) {
        olderId = mMessages.at(mMessages.count() - 1).id();
    }

    if (mTelegramClient) {
        mTelegramClient->messagesGetHistory(&peer, 0, olderId, 20);
    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "Impossible to get history. Telegram client reference is null";
    }

}

QHash<int, QByteArray> MessagesModel::roleNames() const {
    QHash<int, QByteArray> roles;
    roles[IdRole] = "id";
    roles[DialogIdRole] = "dialogId";
    roles[FromIdRole] = "fromId";
    roles[FromFirstNameRole] = "fromFirstName";
    roles[FromLastNameRole] = "fromLastName";
    roles[FromPhoneRole] = "fromPhone";
    roles[FromThumbnailRole] = "fromThumbnail";
    roles[FromUserTypeRole] = "fromUserType";
    roles[ToIdRole] = "toId";
    roles[OutRole] = "out";
    roles[SentRole] = "sent";
    roles[UnreadRole] = "unread";
    roles[SectionDateRole] = "sectionDate";
    roles[DateRole] = "date";
    roles[TtlRole] = "ttl";
    roles[TextRole] = "text";
    roles[MediaTypeRole] = "mediaType";
    roles[PhotoRole] = "photo";
    roles[ThumbnailRole] = "thumbnail";
    roles[VideoRole] = "video";
    roles[DownloadingRole] = "downloading";
    roles[DownloadedPercentageRole] = "downloadedPercentage";
    roles[IsActionRole] = "isAction";
    roles[ActionTypeRole] = "actionType";
    roles[ActionTitleRole] = "actionTitle";
    roles[ActionUserRole] = "actionUser";
    return roles;
}

qint32 MessagesModel::rowCount(const QModelIndex &parent) const {

    if (parent.isValid()) {
        return 0;
    }

    return mMessages.size(); // why is this not .count() ?
}

QVariant MessagesModel::data(const QModelIndex &index, int role) const {

    if (!index.isValid() || index.row() < 0 || index.row() >= mMessages.count()) {
        return QVariant();
    }

    MessageItem message = mMessages[index.row()];

    QVariant result;
    // depending on role value, return one or another property as QVariant
    switch (role) {
    case IdRole:
        result = message.id();
        break;
    case DialogIdRole:
        result = message.dialogId();
        break;
    case FromIdRole:
        result = message.fromId();
        break;
    case FromFirstNameRole:
        result = message.fromFirstName();
        break;
    case FromLastNameRole:
        result = message.fromLastName();
        break;
    case FromPhoneRole:
        result = message.fromPhone();
        break;
    case FromThumbnailRole:
        result = message.fromThumbnail();
        break;
    case FromUserTypeRole:
        result = message.fromUserType();
        break;
    case ToIdRole:
        result = message.toId();
        break;
    case OutRole:
        result = message.out();
        break;
    case UnreadRole:
        return message.unread();
        break;
    case SentRole:
        return message.sent();
        break;
    // TODO karni: We should probably refactor and follow the messaging app http://paste.ubuntu.com/7981163/
    case SectionDateRole: {
            qint64 msec = (qint64) message.date() * 1000;
            QDateTime date = QDateTime::fromMSecsSinceEpoch(msec);
            return date.toString("MMMM d");
        }
    case DateRole:
        return message.date();
        break;
    case TtlRole:
        return message.ttl();
        break;
    case TextRole:
        result = message.text();
        break;
    case MediaTypeRole:
        result = static_cast<qint32>(message.mediaType());
        break;
    case PhotoRole:
        result = message.photo();
        break;
    case ThumbnailRole:
        result = message.thumbnail();
        break;
    case VideoRole:
        result = message.video();
        break;
    case DownloadingRole:
        result = message.downloading();
        break;
    case DownloadedPercentageRole:
        result = message.downloadedPercentage();
        break;
    case IsActionRole:
        result = message.isAction();
        break;
    case ActionTypeRole:
        result = message.actionType();
        break;
    case ActionTitleRole:
        result = message.actionTitle();
        break;
    case ActionUserRole:
        result = message.actionUser();
        break;
    }

    return result;
}

bool MessagesModel::acceptMessage(const MessageItem &message) {

    return message.dialogId() == mDialogId;
}

void MessagesModel::onMessageAdded(const MessageItem &message) {

    if (acceptMessage(message)) {

        if (!exists(message)) {
            beginInsertRows(QModelIndex(), 0, 0);
            mMessages.prepend(message);
            endInsertRows();
            mKeysMap.insert(message.id(), &mMessages[0]);
            Q_EMIT countChanged(mMessages.count());
        }
    }
}

void MessagesModel::onMessagesAdded(const QList<MessageItem> &messages, qint32 totalCount) {

    insertNewMessages(messages);
    mCanFetchMore = totalCount > 0 && totalCount > mMessages.count();
}

void MessagesModel::onOutgoingMessageAdded(qint32, const MessageItem &message) {
    onMessageAdded(message);
}

void MessagesModel::onMessageUpdated(qint32 msgId, const MessageItem &message) {

    MessageItem *messageItem = mKeysMap.value(msgId);

    if (messageItem) {
        if (msgId != message.id()) {
            mKeysMap.remove(msgId);
            mKeysMap.insert(message.id(), messageItem);
        }

        *messageItem = message;

        int pos = mMessages.indexOf(*messageItem);

        Q_EMIT dataChanged(index(pos), index(pos));
    }
}

void MessagesModel::onMessagesMarkedAsRead(const QList<qint32> &msgIds) {
// FIXME This is O(n^2), we can improve this to O(nlogn) or even O(n) if msgIds is sorted.
    for (int i=0; i< mMessages.count(); i++) {
        Q_FOREACH (qint32 msgId, msgIds) {
            if (mMessages.at(i).id() == msgId) {
                mMessages[i].setUnread(false);
                const QVector<qint32> changedRoles(QVector<qint32>() << UnreadRole);
                Q_EMIT dataChanged(index(i), index(i), changedRoles);
                break;
            }
        }
    }
}

void MessagesModel::onMessagePhotoUpdated(qint32 msgId, const QString &localPath, bool isThumbnail) {

    MessageItem *messageItem = mKeysMap.value(msgId);

    if (messageItem) {
        qint32 pos = mMessages.indexOf(*messageItem);

        const QVector<qint32> changedRoles(QVector<qint32>() << ThumbnailRole << PhotoRole);

        isThumbnail ? messageItem->setThumbnail(localPath) : messageItem->setPhoto(localPath);

        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void MessagesModel::onMessageVideoUpdated(qint32 msgId, const QString &videoLocalPath) {

    MessageItem *messageItem = mKeysMap.value(msgId);
    if (messageItem) {

        qint32 pos = mMessages.indexOf(*messageItem);

        const QVector<qint32> changedRoles(QVector<qint32>() << VideoRole);

        messageItem->setVideo(videoLocalPath);

        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void MessagesModel::onMessageDownloadingUpdated(qint32 msgId, bool downloading) {

    MessageItem *messageItem = mKeysMap.value(msgId);
    if (messageItem) {

        qint32 pos = mMessages.indexOf(*messageItem);

        const QVector<qint32> changedRoles(QVector<qint32>() << DownloadingRole);

        messageItem->setDownloading(downloading);

        Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
    }
}

void MessagesModel::onMessageDownloadedPercentageUpdated(qint32 msgId, qint8 downloadedPercentage) {

    MessageItem *messageItem = mKeysMap.value(msgId);
    if (messageItem) {
        qint32 pos = mMessages.indexOf(*messageItem);
        bool downloading = downloadedPercentage > 0 && downloadedPercentage < 100;
        messageItem->setDownloading(downloading);
        messageItem->setDownloadedPercentage(downloadedPercentage);
        QModelIndex i = index(pos);
        Q_EMIT dataChanged(i, i, DOWNLOADED_PERCENTAGE_CHANGED_ROLES);
    }
}

void MessagesModel::onMessagesDeleted(const QList<qint32> &msgIds) {

    for (int i = 0; i < mMessages.count(); i++) {
        Q_FOREACH (qint32 msgId, msgIds) {
            if (mMessages.at(i).id() == msgId) {
                beginRemoveRows(QModelIndex(), i, i);
                mMessages.removeAt(i);
                endRemoveRows();
                mKeysMap.remove(msgId);
            }
        }
    }
    Q_EMIT countChanged(mMessages.count());
}

void MessagesModel::onChatHistoryDeleted(qint32 chatId) {
    if (mDialogId != chatId) return;
    clear();
}

void MessagesModel::onContactThumbnailUpdated(qint32 userId, const QString &localPath) {
    for (qint32 pos = 0; pos < mMessages.count(); pos++) {
        MessageItem &message = mMessages[pos];
        if (message.fromId() == userId) {
            message.setFromThumbnail(localPath);
            const QVector<qint32> changedRoles(QVector<qint32>() << FromThumbnailRole);
            Q_EMIT dataChanged(index(pos), index(pos), changedRoles);
        }
    }
}

bool MessagesModel::exists(const MessageItem &message) {
    return mKeysMap.value(message.id(), 0) != 0;
}
