﻿/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MESSAGESMODEL_H
#define MESSAGESMODEL_H

#include <QAbstractListModel>
#include "messageitem.h"
#include "messagesview.h"
#include "telegramclient.h"

class MessagesModel : public QAbstractListModel
{
    Q_OBJECT
    Q_ENUMS(MessageRoles)
    Q_PROPERTY(qint32 dialogId READ dialogId WRITE setDialogId NOTIFY dialogIdChanged)
    Q_PROPERTY(qint32 peerType READ peerType WRITE setPeerType NOTIFY peerTypeChanged)
    Q_PROPERTY(bool isSecret READ isSecret WRITE setIsSecret NOTIFY isSecretChanged)
    Q_PROPERTY(qint32 count READ getCount NOTIFY countChanged)
public:

    enum MessageRoles {
        IdRole = Qt::UserRole,
        DialogIdRole,
        FromIdRole,
        FromFirstNameRole,
        FromLastNameRole,
        FromPhoneRole,
        FromThumbnailRole,
        FromUserTypeRole,
        ToIdRole,
        OutRole,
        UnreadRole,
        SentRole,
        SectionDateRole,
        DateRole,
        TtlRole,
        TextRole,
        MediaTypeRole,
        PhotoRole,
        ThumbnailRole,
        VideoRole,
        DownloadingRole,
        DownloadedPercentageRole,
        IsActionRole,
        ActionTypeRole,
        ActionTitleRole,
        ActionUserRole
    };
    explicit MessagesModel(QObject *parent = 0);
    ~MessagesModel();

    // abstract model overriding
    QHash<qint32, QByteArray> roleNames() const;
    qint32 rowCount(const QModelIndex &parent = QModelIndex()) const;
    QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;

    Q_INVOKABLE bool canFetchMore(const QModelIndex &parent = QModelIndex()) const;
    Q_INVOKABLE void fetchMore(const QModelIndex &parent = QModelIndex());

    Q_INVOKABLE void setTelegramClient(TelegramClient *telegramClient);
    Q_INVOKABLE QVariant get(int rowIndex) const;

    qint32 dialogId() const;
    qint32 peerType() const;
    bool isSecret() const;
    qint32 getCount() const;
    void setDialogId(qint32 dialogId);
    void setPeerType(qint32 peerType);
    void setIsSecret(bool isSecret);

Q_SIGNALS:
    void dialogIdChanged(qint32);
    void peerTypeChanged(qint32);
    void isSecretChanged(bool);
    void countIncreased();
    void countChanged(qint32);

public Q_SLOTS:
    Q_INVOKABLE void updateQuery();
    Q_INVOKABLE void refresh();

protected Q_SLOTS:
    void clear();
    void resetModel();

    void onNextPage(const QList<MessageItem> &messages);
    void onGetUserAnswer(QVariant user);

    void onMessagesAdded(const QList<MessageItem> &messages, qint32 totalCount);
    void onMessageAdded(const MessageItem &message);
    void onOutgoingMessageAdded(qint32 recipientId, const MessageItem &message);
    void onMessageUpdated(qint32 msgId, const MessageItem &message);
    void onMessagesDeleted(const QList<qint32> &msgIds);
    void onChatHistoryDeleted(qint32 chatId);

    void onMessagePhotoUpdated(qint32 msgId, const QString &localPath, bool isThumbnail);
    void onMessageVideoUpdated(qint32 msgId, const QString &videoLocalPath);

    void onMessagesMarkedAsRead(const QList<qint32> &msgIds);

    void onMessageDownloadingUpdated(qint32 msgId, bool downloading);
    void onMessageDownloadedPercentageUpdated(qint32 msgId, qint8 downloadedPercentage);

    void onContactThumbnailUpdated(qint32 userId, const QString &localPath);

private:
    const QVector<qint32> DOWNLOADED_PERCENTAGE_CHANGED_ROLES;

    qint32 mDialogId;
    qint32 mPeerType;
    qint64 mAccessHash;
    bool mIsSecret;

    // list of the messages in model
    QList<MessageItem> mMessages;
    // correlates message->id() --> pointer to message in mMessages
    QMap<qint32, MessageItem *> mKeysMap;

    bool mCanFetchMore;

    // reference to telegram client
    TelegramClient *mTelegramClient;

    void getHistoryFromServer();
    bool acceptMessage(const MessageItem &message);
    bool exists(const MessageItem &message);
    void insertNewMessages(const QList<MessageItem> &messages);
};

#endif // MESSAGESMODEL_H
