/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "messagesview.h"

#include <QDateTime>
#include "data.h"
#include "dbmanager.h"

MessagesView::MessagesView(qint32 dialogId, qint32 peerType, const QSqlDatabase &database, QObject *parent) :
    View(database, parent),
    mDialogId(dialogId),
    mIsChat(peerType == InputPeer::typeInputPeerChat) {

    createView();
}

void MessagesView::createView() {
    if (mDialogId) {
        QString queryText = "CREATE VIEW IF NOT EXISTS messagesView_%1 AS\n"
                "   SELECT\n"
                "       id,\n"
                "       fromId,\n"
                "       toId,\n"
                "       toIsChat,\n"
                "       out,\n"
                "       sent,\n"
                "       unread,\n"
                "       ttl,\n"
                "       date,\n"
                "       text,\n"
                "       mediaType,\n"
                "       (SELECT firstName FROM users WHERE id=fromId) AS fromFirstName,\n"
                "       (SELECT lastName FROM users WHERE id=fromId) AS fromLastName,\n"
                "       (SELECT phone FROM users WHERE id=fromId) AS fromPhone,\n"
                "       (SELECT localPath FROM fileLocations WHERE rowid IN (SELECT thumbnail FROM users WHERE id=fromId)) AS fromThumbnail,\n"
                "       (SELECT type FROM users WHERE id=fromId) AS fromUserType,\n"
                "       (SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                            "photoId=mediaId AND (type='s' OR type='m') ORDER BY size DESC LIMIT 1) AS photoSmall,\n"
                "       (SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                            "photoId=mediaId AND type='local' ORDER BY size DESC LIMIT 1) AS photoBigLocal,\n"
                "       (SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                            "photoId=mediaId ORDER BY size DESC LIMIT 1) AS photoBigRemote,\n"
                "       (SELECT localPath FROM mediaVideos WHERE id=mediaId) AS video,\n"
                "       (SELECT type FROM messageActions WHERE messageId=id) AS actionType,\n"
                "       (SELECT title FROM messageActions WHERE messageId=id) AS actionTitle,\n"
                "       (SELECT coalesce(users.firstName,'') || ' ' || coalesce(users.lastName,'') FROM users WHERE id IN "
                            "(SELECT userId FROM messageActions WHERE messageId=messages.id)) AS actionUser\n"
                "   FROM messages\n"
                "   WHERE dialogId=%2\n"
                "   ORDER BY date DESC";

        // It is needed to cast dialogId to unsigned for avoid having negative values in view name that would crash the view creation
        // but use dialogId with its original value to query for its messages
        queryText = queryText.arg(QString::number((quint32)mDialogId), QString::number(mDialogId));

        mQuery.prepare(queryText);
        if (!mQuery.exec()) {
            Q_EMIT invalidated();
            qCCritical(TG_PLUGIN_LOGIC) << "Error:" << mQuery.lastError() << mQuery.lastQuery();
            return;
        }
    }
}

void MessagesView::getNextPage() {

    QList<MessageItem> messages;

    mQuery.prepare(QString("SELECT id, fromId, toId, toIsChat, out, unread, sent, date, ttl, text, mediaType, fromFirstName, "
                           "fromLastName, fromPhone, fromThumbnail, fromUserType, photoSmall, photoBigLocal, photoBigRemote, "
                           "video, actionType, actionTitle, actionUser "
                           "FROM messagesView_%1 LIMIT %2 OFFSET %3")
                   .arg(QString::number((quint32)mDialogId), QString::number(mPageSize), QString::number(mOffset)));
    if (!mQuery.exec()) {
        Q_EMIT invalidated();
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << mQuery.lastError() << mQuery.lastQuery();
    }

    while (mQuery.next()) {
        MessageItem message;

        qint32 id = mQuery.value("id").toInt();
        qint32 fromId = mQuery.value("fromId").toInt();
        QString fromFirstName = mQuery.value("fromFirstName").toString();
        QString fromLastName = mQuery.value("fromLastName").toString();
        QString fromPhone = mQuery.value(("fromPhone")).toString();
        QString fromThumbnail = mQuery.value("fromThumbnail").toString();
        qint32 fromUserType = mQuery.value("fromUserType").toInt();
        qint32 toId = mQuery.value("toId").toInt();
        bool toIsChat = mQuery.value("toIsChat").toBool();
        bool out = mQuery.value("out").toBool();
        bool unread = mQuery.value("unread").toBool();
        bool sent = mQuery.value("sent").toBool();
        qint32 date = mQuery.value("date").toInt();
        qint32 ttl = mQuery.value("ttl").toInt();
        QString text = mQuery.value("text").toString();
        qint32 mediaType = mQuery.value("mediaType").toInt();
        QString photoSmall = mQuery.value("photoSmall").toString();
        QString photoBigLocal = mQuery.value("photoBigLocal").toString();
        QString photoBigRemote = mQuery.value("photoBigRemote").toString();
        QString video = mQuery.value("video").toString();
        qint32 actionType = mQuery.value("actionType").toInt();
        QString actionTitle = mQuery.value("actionTitle").toString();
        QString actionUser = mQuery.value("actionUser").toString();

        QString photoBig = photoBigLocal.length() > 0 ? photoBigLocal : photoBigRemote;

        message.setDialogId(mDialogId);
        message.setId(id);
        message.setFromId(fromId);
        message.setFromFirstName(fromFirstName);
        message.setFromLastName(fromLastName);
        message.setFromPhone(fromPhone);
        message.setFromThumbnail(fromThumbnail);
        message.setFromUserType(fromUserType);
        Peer peer(Peer::typePeerUser);
        toIsChat ? peer.setChatId(toId) : peer.setUserId(toId);
        message.setToId(peer);
        message.setOut(out);
        message.setUnread(unread);
        message.setSent(sent);
        message.setDate(date);
        message.setTtl(ttl);
        message.setMessage(text);
        MessageMedia media((MessageMedia::MessageMediaType)mediaType);
        message.setMedia(media);
        message.setThumbnail(photoSmall);
        message.setPhoto(photoBig);
        message.setVideo(video);

        //Evaluate if downloaded
        bool downloaded = false;
        if (out) {
            downloaded = sent;
        } else {
            switch (mediaType){
            case MessageMedia::typeMessageMediaPhoto:
                downloaded = !photoBig.isEmpty();
                break;
            case MessageMedia::typeMessageMediaVideo:
                downloaded = !video.isEmpty();
                break;
            }
        }
        downloaded ? message.setDownloadedPercentage(100) : message.setDownloadedPercentage(0);
        message.setDownloading(false);

        if (actionType) {
            MessageAction messageAction((MessageAction::MessageActionType)actionType);
            messageAction.setTitle(actionTitle);
            message.setAction(messageAction);
            message.setActionUser(actionUser);
        }

        messages << message;
    }

    mOffset += mPageSize;
    mQuery.clear();

    Q_EMIT nextPage(messages);
}
