/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *      Roberto Mier
 *      Tiago Herrmann
 */
#ifndef TELEGRAMCLIENT_H
#define TELEGRAMCLIENT_H

#include <QObject>
#include "rawapiclient.h"

class TelegramClient : public RawApiClient
{
    Q_OBJECT
    Q_PROPERTY(qint32 ourId READ ourId NOTIFY ourIdChanged)
    Q_PROPERTY(QString databasePath READ databasePath WRITE setDatabasePath NOTIFY databasePathChanged)
    Q_PROPERTY(QString downloadsPath READ downloadsPath WRITE setDownloadsPath NOTIFY downloadsPathChanged)

    Q_PROPERTY(qint32 chatToOpen READ getChatToOpen WRITE setChatToOpen)

public:
    explicit TelegramClient(QObject *parent = 0);
    ~TelegramClient();

    qint32 ourId();
    QString databasePath() const;
    void setDatabasePath(const QString &databasePath);
    QString downloadsPath() const;
    void setDownloadsPath(const QString &downloadsPath);
    qint32 getChatToOpen() const;
    void setChatToOpen(qint32 chatToOpen);

    Q_INVOKABLE void init();
    Q_INVOKABLE void shutdown();
    Q_INVOKABLE void sleep();
    Q_INVOKABLE void wake();

    // Models
    Q_INVOKABLE void initializeDialogsView();
    Q_INVOKABLE void initializeContactsView();
    Q_INVOKABLE void initializeMessagesView(qint32 dialogId, qint32 peerType);
    Q_INVOKABLE void fetchDialogsNextPage();
    Q_INVOKABLE void fetchContactsNextPage();
    Q_INVOKABLE void fetchMessagesNextPage();

    Q_INVOKABLE void addContact(const QString &firstName, const QString &lastName, const QString &phone);
    Q_INVOKABLE void getUser(qint32 userId);
    Q_INVOKABLE void downloadPhoto(qint32 messageId);
    Q_INVOKABLE void downloadVideo(qint32 messageId);
    Q_INVOKABLE void downloadUserBigPhoto(qint32 userId);
    Q_INVOKABLE void downloadChatBigPhoto(qint32 chatId);
    Q_INVOKABLE void cancelDownload(qint32 id);
    Q_INVOKABLE void mute(qint32 peerId);
    Q_INVOKABLE void unmute(qint32 peerId);

    Q_INVOKABLE void createSecretChat(qint32 userId);
    Q_INVOKABLE void acceptSecretChat(qint32 chatId);
    Q_INVOKABLE void discardSecretChat(qint32 chatId);
    Q_INVOKABLE void sendSecretChatMessage(qint32 chatId, const QString &message, qint32 ttl = 0);
    Q_INVOKABLE void sendSecretChatPhoto(qint32 chatId, const QString &filePath, qint32 ttl = 0);
    Q_INVOKABLE void sendSecretChatVideo(qint32 chatId, const QString &filePath, qint32 ttl = 0);
    Q_INVOKABLE void readSecretChatHistory(qint32 chatId);
    Q_INVOKABLE void deleteSecretChatHistory(qint32 dialogId, bool deleteChat = false);
    Q_INVOKABLE void setSecretChatTTL(qint32 dialogId, qint32 ttl);

    // Data
    qint32 getDialogIdFromMessageId(qint32 msgId);
    MessageItem topMessage(qint32 dialogId);
    qint32 getUnreadCount(qint32 dialogId);

    //Auth
    Q_INVOKABLE void authSendInvite(const QString &phoneNumber, const QString &message);

    //helpers
    Q_INVOKABLE QString fileMimeType(const QString &fileName);

Q_SIGNALS:
    void databasePathChanged(QString);
    void downloadsPathChanged(QString);
    void ourIdChanged(qint32);

    void openChat(qint32 id);

    void databaseInitialized();

    void error(qint64 id, qint32 errorCode, QString errorText);
    void authSignInError(qint64 id, qint32 errorCode, QString errorText);
    void authSignUpError(qint64 id, qint32 errorCode, QString errorText);

    void authNeeded();
    void authLoggedIn();
    void authCheckPhoneAnswer(qint64 id, bool phoneRegistered, bool phoneInvited);
    void authSendCodeAnswer(qint64 id, bool phoneRegistered, qint32 sendCallTimeout);
    void authSendCallAnswer(qint64 id, bool ok);

    void woken();

    void messagesCreateChatAnswer(QVariant chat);
    void addContactsAnswer(const QList<qint32> &contactIds);
    void getUserAnswer(QVariant user);

    // models signals
    void contactsAdded(const QList<ContactItem> &contacts);
    void dialogsAdded(const QList<DialogItem> &dialogs, qint32 totalCount);
    void contactThumbnailUpdated(qint32 userId, const QString &thumbnail);
    void contactBigPhotoUpdated(qint32 userId, const QString &bigPhoto);
    void chatBigPhotoUpdated(qint32 chatId, const QString &bigPhoto);
    void dialogThumbnailUpdated(qint32 dialogId, const QString &thumbnail);
    void dialogTitleUpdated(qint32 dialogId, const QString &title);
    void dialogMuted(qint32 dialogId, bool muted);
    void dialogsMuted(NotifyPeer::NotifyPeerType notifyType, bool muted);
    void messagesAdded(const QList<MessageItem> &messages, qint32 totalCount);
    void messageAdded(const MessageItem &message);
    void outgoingMessageAdded(qint32 recipientId, const MessageItem &message);
    void messageUpdated(qint32 msgId, const MessageItem &message);
    void chatMarkedAsRead(qint32);
    void chatHistoryDeleted(qint32);
    void chatDeleted(qint32);
    void chatUpdated(qint32 chatId);
    void messagesMarkedAsRead(const QList<qint32> &msgIds);
    void messagesDeleted(const QList<qint32> &msgIds);
    void messagePhotoUpdated(qint32 msgId, const QString &localPath, bool isThumbnail);
    void messageVideoUpdated(qint32 msgId, const QString &videoLocalPath);
    void downloadingUpdated(qint32 id, bool downloading);
    void downloadedPercentageUpdated(qint32 id, qint8 downloadedPercentage);

    void secretChatRequested(qint32 chatId, qint32 date, qint32 creatorId, qint64 creatorAccessHash);
    void secretChatCreated(qint32 chatId);
    void secretChatDiscarded(qint32 chatId);
    void secretChatUpdated(qint32 chatId);

    void dialogsNextPage(const QList<DialogItem> &dialogs);
    void messagesNextPage(const QList<MessageItem> &messages);
    void contactsNextPage(const QList<ContactItem> &contacts);

    void unreadCountChanged(qint32 unreadCount);

    void userTyping(qint32 dialogId, const QString &whoIsTyping = QString());

public Q_SLOTS:
    void onMessagesCreateChatAnswer(qint64 id, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq);
    void onPushNotifications(const QStringList &notifications);
    void onPushError(const QString &error);
    void initTelegramLibrary();

private Q_SLOTS:
    void onGetUserAnswer(ContactItem user);
    void openChat();

private:
    qint32 mChatToOpen;

};

#endif // TELEGRAMCLIENT_H
