/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "telegramservice.h"
#include <QDir>
#include <QRegExp>
#include <QDateTime>
#include <secret/secretchatmessage.h>
#include "models/dialogitem.h"
#include "models/contactsmodel.h"
#include "data.h"
#include "dbmanager.h"
#include "tools.h"

#include <thumbnailer.h>

Q_LOGGING_CATEGORY(TG_PLUGIN_LOGIC, "tg.plugin.logic")

TelegramService::TelegramService() :
    RawApiService(),
    mData(this),
    mDialogsView(0),
    mMessagesView(0),
    mContactsView(0) {

    //Data signals linked to 'this' signals
    connect(&mData, SIGNAL(contactsAdded(QList<ContactItem>)), SIGNAL(contactsAdded(QList<ContactItem>)));
    connect(&mData, SIGNAL(contactThumbnailUpdated(qint32,QString)), SIGNAL(contactThumbnailUpdated(qint32,QString)));
    connect(&mData, SIGNAL(contactBigPhotoUpdated(qint32,QString)), SIGNAL(contactBigPhotoUpdated(qint32,QString)));
    connect(&mData, SIGNAL(chatBigPhotoUpdated(qint32,QString)), SIGNAL(chatBigPhotoUpdated(qint32,QString)));
    connect(&mData, SIGNAL(dialogThumbnailUpdated(qint32,QString)), SIGNAL(dialogThumbnailUpdated(qint32,QString)));
    connect(&mData, SIGNAL(dialogTitleUpdated(qint32,QString)), SIGNAL(dialogTitleUpdated(qint32,QString)));
    connect(&mData, SIGNAL(dialogMuted(qint32,bool)), SIGNAL(dialogMuted(qint32,bool)));
    connect(&mData, SIGNAL(dialogsMuted(NotifyPeer::NotifyPeerType,bool)), SIGNAL(dialogsMuted(NotifyPeer::NotifyPeerType,bool)));
    connect(&mData, SIGNAL(messageAdded(MessageItem)), SIGNAL(messageAdded(MessageItem)));
    connect(&mData, SIGNAL(messagesAdded(QList<MessageItem>,qint32)), SIGNAL(messagesAdded(QList<MessageItem>,qint32)));
    connect(&mData, SIGNAL(outgoingMessageAdded(qint32,MessageItem)), SIGNAL(outgoingMessageAdded(qint32,MessageItem)));
    connect(&mData, SIGNAL(messageUpdated(qint32,MessageItem)), SIGNAL(messageUpdated(qint32,MessageItem)));
    connect(&mData, SIGNAL(chatMarkedAsRead(qint32)), SIGNAL(chatMarkedAsRead(qint32)));
    connect(&mData, SIGNAL(chatUpdated(qint32)), SIGNAL(chatUpdated(qint32)));
    connect(&mData, SIGNAL(chatDeleted(qint32)), SIGNAL(chatDeleted(qint32)));
    connect(&mData, SIGNAL(messagesMarkedAsRead(QList<qint32>)), SIGNAL(messagesMarkedAsRead(QList<qint32>)));
    connect(&mData, SIGNAL(messagesDeleted(QList<qint32>)), SIGNAL(messagesDeleted(QList<qint32>)));
    connect(&mData, SIGNAL(messagePhotoUpdated(qint32,QString,bool)), SIGNAL(messagePhotoUpdated(qint32,QString,bool)));
    connect(&mData, SIGNAL(messageVideoUpdated(qint32,QString)), SIGNAL(messageVideoUpdated(qint32,QString)));
    connect(&mData, SIGNAL(downloadingUpdated(qint32,bool)), SIGNAL(downloadingUpdated(qint32,bool)));
    connect(&mData, SIGNAL(downloadedPercentageUpdated(qint32,qint8)), SIGNAL(downloadedPercentageUpdated(qint32,qint8)));
    connect(&mData, SIGNAL(dialogsAdded(QList<DialogItem>,qint32)), SIGNAL(dialogsAdded(QList<DialogItem>,qint32)));
    connect(&mData, SIGNAL(unreadCountChanged(qint32)), SIGNAL(unreadCountChanged(qint32)));
    connect(&mData, SIGNAL(secretChatRequested(qint32,qint32,qint32,qint64)), SIGNAL(secretChatRequested(qint32,qint32,qint32,qint64)));
    connect(&mData, SIGNAL(secretChatCreated(qint32)), SIGNAL(secretChatCreated(qint32)));
    connect(&mData, SIGNAL(secretChatDiscarded(qint32)), SIGNAL(secretChatDiscarded(qint32)));

    connect(&mData, SIGNAL(userTyping(qint32,QString)), SIGNAL(userTyping(qint32,QString)));
    connect(&mData, SIGNAL(secretChatUpdated(qint32)), SIGNAL(secretChatUpdated(qint32)));

    mSelfDestructTimer.setSingleShot(true);
    connect(&mSelfDestructTimer, SIGNAL(timeout()), this, SLOT(processSelfDestructMessages()));
    connect(&mData, SIGNAL(messageAdded(MessageItem)), &mSelfDestructTimer, SIGNAL(timeout()));
    connect(&mData, SIGNAL(messagesMarkedAsRead(QList<qint32>)),&mSelfDestructTimer, SIGNAL(timeout()));
    connect(&mData, SIGNAL(outgoingMessageAdded(qint32,MessageItem)), &mSelfDestructTimer, SIGNAL(timeout()));
    connect(this, SIGNAL(messagesReadEncryptedHistoryAnswer(qint64,bool)), &mSelfDestructTimer, SIGNAL(timeout()));
}

TelegramService::~TelegramService() {
    if (mDialogsView) {
        delete mDialogsView;
        mDialogsView = 0;
    }
    if (mContactsView) {
        delete mContactsView;
        mContactsView = 0;
    }
    if (mMessagesView) {
        delete mMessagesView;
        mMessagesView = 0;
    }
}

void TelegramService::init() {

    qCDebug(TG_PLUGIN_LOGIC) << "Initializing telegram service...";

    QString path = mConfig.databasePath();
    DbManager &dbManager = mData.dbManager();
    dbManager.initialize(path);

    bool createLib = !mTelegramLib;

    RawApiService::initTelegramLibrary();

    if (createLib) {
        connect(mTelegramLib, SIGNAL(fatalError()), SLOT(createCrashFile()));

        connect(mTelegramLib, SIGNAL(authLoggedIn()), SLOT(onAuthLoggedIn()));
        connect(mTelegramLib, SIGNAL(authLogOutAnswer(qint64,bool)), SLOT(onAuthLoggedOut()));

        connect(mTelegramLib, SIGNAL(accountRegisterDeviceAnswer(qint64,bool)),
                        this, SLOT(onAccountRegisterDeviceAnswer(qint64,bool)));
        connect(mTelegramLib, SIGNAL(messagesAddChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
                        this, SLOT(onMessagesAddChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
        connect(mTelegramLib, SIGNAL(messagesGetFullChatAnswer(qint64, ChatFull, QList<Chat>, QList<User>)),
                        this, SIGNAL(messagesGetFullChatAnswer(qint64, ChatFull, QList<Chat>, QList<User>)));
        connect(mTelegramLib, SIGNAL(contactsImportContactsAnswer(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)),
                        this, SLOT(onContactsImportContactsAnswer(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)));
    }

    mTelegramLib->init();
    mSelfDestructTimer.start();
}

void TelegramService::shutdown() {

    qCDebug(TG_PLUGIN_LOGIC) << "Shutdown telegram client";

    mSelfDestructTimer.stop();

    mTelegramLib->disconnect(this);
    mTelegramLib->deleteLater();
    mTelegramLib = 0;

    mData.dbManager().close();

    Q_EMIT finished();
}

void TelegramService::sleep() {
    if (mTelegramLib->sleep()) {
        mData.dbManager().close();
    }
    mSelfDestructTimer.stop();
}

void TelegramService::wake() {
    connect(mTelegramLib, SIGNAL(woken()), SLOT(onWoken()), Qt::UniqueConnection);
    if (mTelegramLib->wake()) {
        mData.dbManager().reopen();
    }
}

void TelegramService::onWoken() {
    mTelegramLib->updatesGetState();
    mSelfDestructTimer.start();
    updateUserProfiles();
    Q_EMIT woken();
}

void TelegramService::onAuthLoggedIn() {

    qCDebug(TG_PLUGIN_LOGIC) << "Logged in";

    deleteCrashFile();

    // Connect lib object server received rpc messages with data instance.
    // Data instance will store in local db and communicate events to models
    connect(mTelegramLib,
            SIGNAL(updatesGetStateAnswer(qint64,qint32,qint32,qint32,qint32,qint32)),
            &mData,
            SLOT(onUpdatesGetStateAnswer(qint64,qint32,qint32,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updatesGetStateAnswer(qint64,qint32,qint32,qint32,qint32,qint32)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(updatesGetDifferenceAnswer(qint64,qint32,qint32)),
            &mData,
            SLOT(onUpdatesGetDifferenceAnswer(qint64,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updatesGetDifferenceAnswer(qint64,qint32,qint32)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(updatesGetDifferenceAnswer(qint64,QList<Message>,QList<SecretChatMessage>,QList<Update>,QList<Chat>,QList<User>,UpdatesState,bool)),
            &mData,
            SLOT(onUpdatesGetDifferenceAnswer(qint64,QList<Message>,QList<SecretChatMessage>,QList<Update>,QList<Chat>,QList<User>,UpdatesState,bool)));
    connect(mTelegramLib,
            SIGNAL(updatesGetDifferenceAnswer(qint64,QList<Message>,QList<SecretChatMessage>,QList<Update>,QList<Chat>,QList<User>,UpdatesState,bool)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(usersGetUsersAnswer(qint64,QList<User>)),
            &mData,
            SLOT(onUsersGetUsersAnswer(qint64,QList<User>)));
    connect(mTelegramLib,
            SIGNAL(messagesGetDialogsAnswer(qint64,qint32,QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)),
            &mData,
            SLOT(onMessagesGetDialogsAnswer(qint64,qint32,QList<Dialog>,QList<Message>,QList<Chat>,QList<User>)));
    connect(mTelegramLib,
            SIGNAL(messagesDeleteMessagesAnswer(qint64,QList<qint32>)),
            &mData,
            SLOT(onMessagesDeleteMessagesAnswer(qint64,QList<qint32>)));
    connect(mTelegramLib,
            SIGNAL(uploadGetFileAnswer(qint64,StorageFileType,qint32,QByteArray,qint32,qint32,qint32)),
            &mData,
            SLOT(onUploadGetFileAnswer(qint64,StorageFileType,qint32,QByteArray,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(uploadCancelFileAnswer(qint64,bool)),
            &mData,
            SLOT(onUploadCancelFileAnswer(qint64,bool)));
    connect(mTelegramLib,
            SIGNAL(uploadSendFileAnswer(qint64,qint32,qint32,qint32)),
            &mData,
            SLOT(onUploadSendFileAnswer(qint64,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(contactsGetContactsAnswer(qint64,bool,QList<Contact>,QList<User>)),
            &mData,
            SLOT(onContactsGetContactsAnswer(qint64,bool,QList<Contact>,QList<User>)));
    connect(mTelegramLib,
            SIGNAL(messagesGetHistoryAnswer(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)),
            &mData,
            SLOT(onMessagesGetHistoryAnswer(qint64,qint32,QList<Message>,QList<Chat>,QList<User>)));
    connect(mTelegramLib,
            SIGNAL(messagesSendMessageAnswer(qint64,qint32,qint32,qint32,qint32,QList<ContactsLink>)),
            &mData,
            SLOT(onMessagesSendMessageAnswer(qint64,qint32,qint32,qint32,qint32,QList<ContactsLink>)));
    connect(mTelegramLib,
            SIGNAL(messagesSendPhotoAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
            &mData,
            SLOT(onMessagesSendPhotoAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesSendVideoAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
            &mData,
            SLOT(onMessagesSendVideoAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesReadHistoryAnswer(qint64,qint32,qint32,qint32)),
            &mData,
            SLOT(onMessagesReadHistoryAnswer(qint64,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesDeleteHistoryAnswer(qint64,qint32,qint32,qint32)),
            &mData,
            SLOT(onMessagesDeleteHistoryAnswer(qint64,qint32,qint32,qint32)));
    connect(&mData,
            SIGNAL(chatHistoryDeleted(qint32)),
            SIGNAL(chatHistoryDeleted(qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesCreateChatAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
            &mData,
            SLOT(onMessagesCreateChatAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesAddChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
            &mData,
            SLOT(onMessagesAddChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesDeleteChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)),
            &mData,
            SLOT(onMessagesDeleteChatUserAnswer(qint64,Message,QList<Chat>,QList<User>,QList<ContactsLink>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesGetFullChatAnswer(qint64,ChatFull,QList<Chat>,QList<User>)),
            &mData,
            SLOT(onMessagesGetFullChatAnswer(qint64,ChatFull,QList<Chat>, QList<User>)));
    connect(mTelegramLib,
            SIGNAL(contactsImportContactsAnswer(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)),
            &mData,
            SLOT(onContactsImportContactsAnswer(qint64,QList<ImportedContact>,QList<qint64>,QList<User>)));
    connect(mTelegramLib,
            SIGNAL(accountUpdateNotifySettingsAnswer(qint64,bool)),
            &mData,
            SLOT(onAccountUpdateNotifySettingsAnswer(qint64,bool)));

    connect(mTelegramLib,
            SIGNAL(messagesCreateEncryptedChatAnswer(qint32,qint32,qint32,qint64)),
            &mData,
            SLOT(onMessagesCreateEncryptedChatAnswer(qint32,qint32,qint32,qint64)));
    connect(mTelegramLib,
            SIGNAL(messagesEncryptedChatRequested(qint32,qint32,qint32,qint64)),
            &mData,
            SLOT(onMessagesEncryptedChatRequested(qint32,qint32,qint32,qint64)));
    connect(mTelegramLib,
            SIGNAL(messagesEncryptedChatCreated(qint32,qint32,qint32,qint64)),
            &mData,
            SLOT(onMessagesEncryptedChatCreated(qint32,qint32,qint32,qint64)));
    connect(mTelegramLib,
            SIGNAL(messagesEncryptedChatDiscarded(qint32)),
            &mData,
            SLOT(onMessagesEncryptedChatDiscarded(qint32)));
    connect(mTelegramLib,
            SIGNAL(messagesSendEncryptedAnswer(qint64,qint32,EncryptedFile)),
            &mData,
            SLOT(onMessagesSendEncryptedAnswer(qint64,qint32,EncryptedFile)));
    connect(mTelegramLib,
            SIGNAL(messagesSendEncryptedFileAnswer(qint64, qint32, EncryptedFile)),
            &mData,
            SLOT(onMessagesSendEncryptedFileAnswer(qint64, qint32, EncryptedFile)));

    // Connect server received notifications to data instance also
    connect(mTelegramLib,
            SIGNAL(updateShortMessage(qint32,qint32,QString,qint32,qint32,qint32)),
            &mData,
            SLOT(onUpdateShortMessage(qint32,qint32,QString,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updateShortMessage(qint32,qint32,QString,qint32,qint32,qint32)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(updateShortChatMessage(qint32,qint32,qint32,QString,qint32,qint32,qint32)),
            &mData,
            SLOT(onUpdateShortChatMessage(qint32,qint32,qint32,QString,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updateShortChatMessage(qint32,qint32,qint32,QString,qint32,qint32,qint32)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(updateShort(Update,qint32)),
            &mData,
            SLOT(onUpdateShort(Update,qint32)));
    connect(mTelegramLib,
            SIGNAL(updateShort(Update,qint32)),
            this,
            SIGNAL(openChat()));

    connect(mTelegramLib,
            SIGNAL(updatesCombined(QList<Update>,QList<User>,QList<Chat>,qint32,qint32,qint32)),
            &mData,
            SLOT(onUpdatesCombined(QList<Update>,QList<User>,QList<Chat>,qint32,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updates(QList<Update>,QList<User>,QList<Chat>,qint32,qint32)),
            &mData,
            SLOT(onUpdates(QList<Update>,QList<User>,QList<Chat>,qint32,qint32)));
    connect(mTelegramLib,
            SIGNAL(updatesTooLong()),
            &mData,
            SLOT(onUpdatesTooLong()));

    connect(mTelegramLib,
            SIGNAL(updateSecretChatMessage(SecretChatMessage,qint32)),
            &mData,
            SLOT(onUpdateSecretChatMessage(SecretChatMessage,qint32)));

    mTelegramLib->updatesGetState();
    updateUserProfiles();
}

void TelegramService::onAuthLoggedOut() {
    mTelegramLib->disconnect(&mData);
}

void TelegramService::getUser(qint32 userId) {
    ContactItem contactItem = mData.getUser(userId);
    Q_EMIT getUserAnswer(contactItem);
}

qint64 TelegramService::addContact(const QString &firstName, const QString &lastName, const QString &phoneNumber) {
    InputContact inputContact;
    inputContact.setFirstName(firstName);
    inputContact.setLastName(lastName);
    inputContact.setPhone(phoneNumber);
    QList<InputContact> contacts;
    contacts << inputContact;
    return RawApiService::contactsImportContacts(contacts);
}

qint64 TelegramService::messagesReadHistory(InputPeer inputPeer, qint32 maxId, qint32 offset, bool readContents) {

    // if peer is foreign, get access hash from db
    if (inputPeer.classType() == InputPeer::typeInputPeerForeign) {
        const ContactItem &user = mData.getUser(inputPeer.userId());
        inputPeer.setAccessHash(user.accessHash());
    }

    qint64 requestId = -1;
    bool isChat = inputPeer.classType() == TLInputPeer::typeInputPeerChat;
    qint32 chatId = isChat ? inputPeer.chatId() : inputPeer.userId();
    bool markedAsRead = mData.markChatAsRead(chatId);
    if (markedAsRead) {
        requestId = RawApiService::messagesReadHistory(inputPeer, maxId, offset, readContents);
    }
    return requestId;
}

qint64 TelegramService::deleteChatHistory(InputPeer inputPeer, qint32 offset, bool deleteChat) {
    if (inputPeer.classType() == InputPeer::typeInputPeerForeign) {
        const ContactItem &user = mData.getUser(inputPeer.userId());
        inputPeer.setAccessHash(user.accessHash());
    }

    bool isChat = inputPeer.classType() == TLInputPeer::typeInputPeerChat;
    qint32 chatId = isChat ? inputPeer.chatId() : inputPeer.userId();
    mData.markChatAsRead(chatId);

    qint64 requestId = RawApiService::messagesDeleteHistory(inputPeer, offset);
    // Potential race condition here? *Answer may have already been called.
    mData.mChatsToClearHistory.insert(requestId, chatId);
    if (deleteChat) {
        mData.mChatsToDelete.insert(requestId, chatId);
    }
    return requestId;
}

qint64 TelegramService::messagesDeleteMessages(const QList<qint32> &msgIds, bool isSecret) {
    if (isSecret) {
        mData.deleteMessages(msgIds);
        return 0;
    } else {
        return RawApiService::messagesDeleteMessages(msgIds);
    }
}

qint64 TelegramService::messagesSendMessage(InputPeer inputPeer, const QString &text) {

    // if peer is foreign, get access hash from db
    if (inputPeer.classType() == InputPeer::typeInputPeerForeign) {
        const ContactItem &user = mData.getUser(inputPeer.userId());
        inputPeer.setAccessHash(user.accessHash());
    }

    //Unique client message ID required to prevent message resending
    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSendMessage(inputPeer, randomId, text);
    qint32 recipientId = inputPeer.userId();
    bool isChat = false;
    if (inputPeer.classType() == TLInputPeer::typeInputPeerChat) {
       recipientId = inputPeer.chatId();
       isChat = true;
    }

    Message message(Message::typeMessage);
    // Use requestId as provisional message identifier generator until having one definitive.
    qint32 msgId = Tools::requestIdToMsgId(requestId);
    message.setId(msgId);
    message.setMessage(text);
    message.setFromId(Settings::getInstance()->ourId());
    Peer peer(Peer::typePeerUser);
    if (isChat) {
       peer.setClassType(Peer::typePeerChat);
       peer.setChatId(recipientId);
    } else {
       peer.setUserId(recipientId);
    }
    message.setToId(peer);
    message.setOut(true);
    message.setUnread(true);
    message.setDate(QDateTime::currentDateTime().toTime_t());

    mData.appendOutgoingMessage(recipientId, randomId, message);

    return requestId;
}

qint64 TelegramService::messagesSendPhoto(InputPeer inputPeer, const QString &photoPath) {

    // if peer is foreign, get access hash from db
    if (inputPeer.classType() == InputPeer::typeInputPeerForeign) {
        const ContactItem &user = mData.getUser(inputPeer.userId());
        inputPeer.setAccessHash(user.accessHash());
    }

    QString photoLocalPath = copyToDownloadsPath(photoPath);
    if (photoLocalPath.isEmpty()) {
        return -1;
    }

    //Unique client message ID required to prevent message resending
    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSendPhoto(inputPeer, randomId, photoLocalPath);
    qint32 recipientId = inputPeer.userId();
    bool isChat = false;
    if (inputPeer.classType() == TLInputPeer::typeInputPeerChat) {
       recipientId = inputPeer.chatId();
       isChat = true;
    }

    Message message(Message::typeMessage);
    // Use requestId as provisional message identifier generator until having one definitive.
    qint32 msgId = Tools::requestIdToMsgId(requestId);
    message.setId(msgId);
    message.setFromId(Settings::getInstance()->ourId());
    Peer peer(Peer::typePeerUser);
    if (isChat) {
       peer.setClassType(Peer::typePeerChat);
       peer.setChatId(recipientId);
    } else {
       peer.setUserId(recipientId);
    }
    message.setToId(peer);
    message.setOut(true);
    message.setUnread(true);
    message.setDate(QDateTime::currentDateTime().toTime_t());

    // Compose media
    MessageMedia media(MessageMedia::typeMessageMediaPhoto);
    Photo photo(Photo::typePhoto);
    // set provisional id for media the temp generated msgId. When photo is sent, update this values with real one.
    photo.setId(msgId);
    media.setPhoto(photo);
    message.setMedia(media);
    mData.appendOutgoingMessage(recipientId, randomId, message, photoLocalPath, requestId);

    return requestId;
}

qint64 TelegramService::messagesSetTyping(const InputPeer peer) {
    SendMessageAction action;
    action.setClassType(SendMessageAction::typeSendMessageTypingAction);
    RawApiService::messagesSetTyping(peer, action);
}

bool TelegramService::downloadPhoto(qint32 messageId) {
    return mData.downloadPhoto(messageId);
}

bool TelegramService::downloadUserBigPhoto(qint32 userId) {
    return mData.downloadUserBigPhoto(userId);
}

bool TelegramService::downloadChatBigPhoto(qint32 chatId) {
    return mData.downloadChatBigPhoto(chatId);
}

qint64 TelegramService::sendVideo(qint32 peerId, const QString &videoPath, qint32 duration, qint32 width, qint32 height, const QString &thumbnailPath) {

    QString videoLocalPath = copyToDownloadsPath(videoPath);
    QString thumbnailLocalPath;
    if (thumbnailPath.isEmpty()) {
        QString generatedThumbnail = Tools::generateThumbnail(videoPath);
        if (!generatedThumbnail.isEmpty()) {
            thumbnailLocalPath = copyToDownloadsPath(generatedThumbnail);
            if (thumbnailLocalPath.isEmpty()) {
                return -1;
            }
            qCDebug(TG_PLUGIN_LOGIC) << "generated thumbnail at path:" << thumbnailLocalPath;
        }
    } else {
        thumbnailLocalPath = copyToDownloadsPath(thumbnailPath);
    }

    // Thumbnail is generated in .png. Transform to jpeg, since Telegram accepts only jpeg format
    thumbnailLocalPath = Tools::pngToJpeg(thumbnailLocalPath);

    ContactItem user = mData.getUser(peerId);
    InputPeer inputPeer = Tools::toInputPeer(peerId, user);

    //TODO Roberto: HD resolution hardcoded workaround for having video working on IOs.
    if (width == 0 || height == 0) {
        width = 1920;
        height = 1080;
    }

    //Unique client message ID required to prevent message resending
    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = RawApiService::messagesSendVideo(inputPeer, randomId, videoLocalPath, duration, width, height, thumbnailLocalPath);

    Message message(Message::typeMessage);
    // Use requestId as provisional message identifier generator until having one definitive.
    qint32 msgId = Tools::requestIdToMsgId(requestId);
    message.setId(msgId);
    message.setFromId(Settings::getInstance()->ourId());
    Peer peer(Peer::typePeerUser);
    if (user.isEmpty()) {
       peer.setClassType(Peer::typePeerChat);
       peer.setChatId(peerId);
    } else {
       peer.setUserId(peerId);
    }
    message.setToId(peer);
    message.setOut(true);
    message.setUnread(true);
    message.setDate(QDateTime::currentDateTime().toTime_t());

    // Compose media
    MessageMedia media(MessageMedia::typeMessageMediaVideo);
    Video video(Video::typeVideo);
    // set provisional id for media the temp generated msgId. When photo is sent, update this values with real one.
    video.setId(msgId);
    video.setDuration(duration);
    video.setW(width);
    video.setH(height);
    media.setVideo(video);
    message.setMedia(media);
    mData.appendOutgoingMessage(peerId, randomId, message, videoLocalPath, requestId, thumbnailLocalPath);

    return requestId;
}

bool TelegramService::downloadVideo(qint32 messageId) {
    return  mData.downloadVideo(messageId);
}

qint64 TelegramService::cancelDownload(qint32 id) {
    qint64 fileId = mData.getDownloadingFileId(id);
    Q_EMIT downloadedPercentageUpdated(id, 0);
    Q_EMIT downloadingUpdated(id, false);
    return mTelegramLib->uploadCancelFile(fileId);
}

qint64 TelegramService::addChatUser(qint32 chatId, qint32 userId, qint32 fwdLimit) {
    ContactItem user = mData.getUser(userId);
    InputUser inputUser = Tools::toInputUser(user);
    RawApiService::messagesAddChatUser(chatId, inputUser, fwdLimit);
}

qint64 TelegramService::deleteChatUser(qint32 chatId, qint32 userId) {
    ContactItem user = mData.getUser(userId);
    InputUser inputUser = Tools::toInputUser(user);
    RawApiService::messagesDeleteChatUser(chatId, inputUser);
}

qint64 TelegramService::mute(qint32 peerId) {
    // not possible to mute forever. Leave this muted for a year
    qint32 oneYearLater = QDateTime::currentDateTime().addYears(1).toTime_t();
    qint64 requestId = updateMuteUntilSetting(peerId, oneYearLater);
    mData.setMutingInProgress(requestId, peerId, true);
    return requestId;
}

qint64 TelegramService::unmute(qint32 peerId) {
    qint32 currentDate = QDateTime::currentDateTime().toTime_t();
    qint64 requestId = updateMuteUntilSetting(peerId, currentDate);
    mData.setMutingInProgress(requestId, peerId, false);
    return requestId;
}

qint64 TelegramService::updateMuteUntilSetting(qint32 peerId, qint32 muteUntil) {
    ContactItem user = mData.getUser(peerId);
    InputPeer inputPeer = Tools::toInputPeer(peerId, user);
    InputNotifyPeer inputNotifyPeer(InputNotifyPeer::typeInputNotifyPeer);
    inputNotifyPeer.setPeer(inputPeer);

    InputPeerNotifySettings settings;
    settings.setMuteUntil(muteUntil);
    return mTelegramLib->accountUpdateNotifySettings(inputNotifyPeer, settings);
}

void TelegramService::updateUserProfiles() {
    // Get all the users info to update db when started up
    QList<ContactItem> contacts = mData.getUsers();
    QList<InputUser> inputUsers;
    Q_FOREACH (const ContactItem &contact, contacts) {
        inputUsers << Tools::toInputUser(contact);
    }
    mTelegramLib->usersGetUsers(inputUsers);
}

qint64 TelegramService::createSecretChat(qint32 userId) {
    ContactItem user = mData.getUser(userId);
    InputUser inputUser = Tools::toInputUser(user);
    return mTelegramLib->messagesCreateEncryptedChat(inputUser);
}

qint64 TelegramService::acceptSecretChat(qint32 chatId) {
    return mTelegramLib->messagesAcceptEncryptedChat(chatId);
}

qint64 TelegramService::discardSecretChat(qint32 chatId) {
    return mTelegramLib->messagesDiscardEncryptedChat(chatId);
}

qint64 TelegramService::setSecretChatTTL(qint32 chatId, qint32 ttl) {
    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSetEncryptedTTL(randomId, chatId, ttl);

    DecryptedMessage message(DecryptedMessage::typeDecryptedMessageService);
    message.setRandomId(randomId);
    message.setTtl(ttl);

    mData.setSecretChatMessageTTL(chatId, ttl);

    return requestId;
}

qint64 TelegramService::sendSecretChatMessage(qint32 chatId, const QString &text, qint32 ttl) {
    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSendEncrypted(chatId, randomId, ttl, text);

    DecryptedMessage message(DecryptedMessage::typeDecryptedMessage);
    message.setRandomId(randomId);
    message.setTtl(ttl);
    message.setMessage(text);

    mData.appendOutgoingDecryptedMessage(requestId, chatId, message);

    return requestId;
}

qint64 TelegramService::sendSecretChatPhoto(qint32 chatId, const QString &photoPath, qint32 ttl) {
    QString photoLocalPath = copyToDownloadsPath(photoPath);
    if (photoLocalPath.isEmpty()) {
        return -1;
    }

    // Android secret chat only previews photos with max size 4096x4096
    Tools::resizePhotoIfNeeded(photoLocalPath);

    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSendEncryptedPhoto(chatId, randomId, ttl, photoLocalPath);
    DecryptedMessage message;
    DecryptedMessageMedia media(DecryptedMessageMedia::typeDecryptedMessageMediaPhoto);
    message.setRandomId(randomId);
    message.setTtl(ttl);
    message.setMedia(media);
    mData.appendOutgoingDecryptedMessage(requestId, chatId, message, photoLocalPath);
    return requestId;
}

qint64 TelegramService::sendSecretChatVideo(qint32 chatId, const QString &videoPath, qint32 ttl) {
    QString videoLocalPath = copyToDownloadsPath(videoPath);
    QByteArray thumbnailBytes;
    //As said in telegram doc, thumbnail must be JPEGfile, quality 55, set in a square 90x90
    //https://core.telegram.org/constructor/decryptedMessageMediaPhoto
    //Our thumbnailer library used to create them is not able to have that precission
    //I've tested that it accepts even jpeg files of quality 90 not beeing 90x90, but no more quality or different format
    QString thumbnailLocalPath = Tools::generateThumbnail(videoPath);
    // Thumbnail is generated in .png. Transform to jpeg
    thumbnailLocalPath = Tools::pngToJpeg(thumbnailLocalPath, 55);

    if (!thumbnailLocalPath.isEmpty()) {
        QFile file(thumbnailLocalPath);
        file.open(QIODevice::ReadOnly);
        thumbnailBytes = file.readAll();
        file.close();
        qCDebug(TG_PLUGIN_LOGIC) << "Created thumbnail of" << thumbnailBytes.length() << "bytes for video";
    }

    //TODO Roberto: HD resolution hardcoded workaround for having video working on IOs.
    qint32 duration = 0;
    qint32 width = 1920;
    qint32 height = 1080;

    qint64 randomId = Tools::generateRandomId();
    qint64 requestId = mTelegramLib->messagesSendEncryptedVideo(chatId, randomId, ttl, videoLocalPath, duration, width, height, thumbnailBytes);
    DecryptedMessage message;
    DecryptedMessageMedia media(DecryptedMessageMedia::typeDecryptedMessageMediaVideo_layer8);
    message.setRandomId(randomId);
    message.setTtl(ttl);
    message.setMedia(media);
    message.setRandomId(randomId);
    mData.appendOutgoingDecryptedMessage(requestId, chatId, message, videoLocalPath, thumbnailLocalPath);
    return requestId;
}

qint64 TelegramService::readSecretChatHistory(qint32 chatId) {

    qint64 requestId = -1;
    if (mData.markSecretChatAsRead(chatId)) {
        qint32 maxDate = mData.getDecryptedMessagesMaxDate(chatId);
        if (maxDate > 0) {
            requestId = RawApiService::messagesReadEncryptedHistory(chatId, maxDate);
        }
    }
    return requestId;
}

void TelegramService::deleteSecretChatHistory(qint32 dialogId, bool deleteChat) {
    mData.markChatAsRead(dialogId);
    mData.deleteSecretChatHistory(dialogId, deleteChat);
}

void TelegramService::onMessagesAddChatUserAnswer(qint64, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq) {
    if (chats.length() > 0) {
        qint32 chatId = chats.at(0).id();
        qint32 userId = message.action().userId();
        Q_EMIT addChatUserAnswer(chatId, userId);
    }
}

void TelegramService::onContactsImportContactsAnswer(qint64, QList<ImportedContact> importedContacts, QList<qint64>, QList<User>) {
    QList<qint32> imported;
    Q_FOREACH (const ImportedContact importedContact, importedContacts) {
        imported << importedContact.userId();
    }
    Q_EMIT addContactsAnswer(imported);
}

void TelegramService::onAccountRegisterDeviceAnswer(qint64 id, bool ok) {
    qCDebug(TG_PLUGIN_LOGIC) << "accountRegisterDeviceAnswer:" << ok;
}

void TelegramService::processSelfDestructMessages() {
    QSqlQuery query;

    // Purge expired messages.
    query.prepare("SELECT id FROM messages WHERE unread = 0 AND 0 < readTime AND 0 < ttl AND readTime + ttl <= :now");
    query.bindValue(":now", QDateTime::currentDateTime().toTime_t());
    if (query.exec()) {
        QList<qint32> messageIdsToDelete;
        while (query.next()) {
            qint32 id = query.value("id").toInt();
            messageIdsToDelete.append(id);
        }
        if (!messageIdsToDelete.empty()) {
            messagesDeleteMessages(messageIdsToDelete, true);
        }
    }

    // Track only the earliest message to delete.
    query.prepare("SELECT MIN(ttl+readTime) AS expires FROM messages WHERE 0 < readTime AND 0 < ttl");
    if (query.exec() && query.next()) {
        if (query.value("expires").isNull()) {
            qCDebug(TG_PLUGIN_LOGIC) << "TTL: all expired messages deleted";
            return;
        }

        qint32 nextInSeconds = query.value("expires").toInt() - QDateTime::currentDateTime().toTime_t();
        if (nextInSeconds > 0) {
            qCDebug(TG_PLUGIN_LOGIC) << "TTL: next message in" << nextInSeconds << "s";
            if (mSelfDestructTimer.isActive()) {
                mSelfDestructTimer.stop();
            }
            mSelfDestructTimer.start(1000*nextInSeconds);
        } else {
            // We already have an expired message.
            mSelfDestructTimer.start(500);
        }
    }
}

void TelegramService::createCrashFile() {
    QFile crashFile(mConfig.configPath() + QDir::separator() + ".crash");
    QFileInfo fileInfo(crashFile);
    crashFile.open(QIODevice::WriteOnly);
    crashFile.close();
    qCDebug(TG_PLUGIN_LOGIC) << "created crash file at " << fileInfo.absoluteFilePath();
}

void TelegramService::deleteCrashFile() {
    QFile crashFile(mConfig.configPath() + QDir::separator() + ".crash");
    if (crashFile.exists()) {
        crashFile.remove();
    }
}


QString TelegramService::copyToDownloadsPath(const QString &file) {
    QString downloadsPath = mConfig.downloadsPath() + QDir::separator() + OUTGOING_MEDIA_FOLDER;
    QFileInfo dir = QFileInfo(downloadsPath);
    QString dirPath = dir.absoluteFilePath();
    if (!dir.exists()) {
        if (QDir::root().mkpath(dirPath)) {
            qCDebug(TG_PLUGIN_LOGIC) << dirPath << "directory created";
        } else {
            qCCritical(TG_PLUGIN_LOGIC) << "error creating directory: " << dirPath;
            return QString();
        }
    }

    QFile qfile(file);
    QFileInfo fileInfo(file);
    QString fileName = fileInfo.fileName();
    QString newPath = downloadsPath + QDir::separator() + fileName;

    // don't copy if origin and destination are the same
    if (file == newPath) {
        return newPath;
    }

    if (QFile::exists(newPath)) {
        QFile::remove(newPath);
    }

    if (!qfile.copy(newPath)) {
        qCCritical(TG_PLUGIN_LOGIC()) << "Could not copy" << file << "to" << newPath;
    }

    return newPath;
}

qint32 TelegramService::ourId() const {
    return mTelegramLib->ourId();
}

QString TelegramService::databasePath() {
    return mConfig.databasePath();
}
void TelegramService::setDatabasePath(const QString &path) {
    mConfig.setDatabasePath(path);
}

QString TelegramService::downloadsPath() const {
    return mConfig.downloadsPath();
}

void TelegramService::setDownloadsPath(const QString &downloadsPath) {
    mConfig.setDownloadsPath(downloadsPath);
}

Data &TelegramService::data() {
    return mData;
}

void TelegramService::initializeDialogsView() {
    if (mDialogsView) {
        mDialogsView->deleteLater();
        mDialogsView = 0;
    }
    const QSqlDatabase &database = mData.dbManager().database();
    mDialogsView = new DialogsView(database);
    connect(mDialogsView, SIGNAL(nextPage(QList<DialogItem>)), SIGNAL(dialogsNextPage(QList<DialogItem>)), Qt::UniqueConnection);
}

void TelegramService::initializeMessagesView(qint32 dialogId, qint32 peerType) {
    if (mMessagesView) {
        mMessagesView->deleteLater();
        mMessagesView = 0;
    }
    const QSqlDatabase &database = mData.dbManager().database();
    mMessagesView = new MessagesView(dialogId, peerType, database);
    connect(mMessagesView, SIGNAL(nextPage(QList<MessageItem>)), SIGNAL(messagesNextPage(QList<MessageItem>)), Qt::UniqueConnection);
}

void TelegramService::initializeContactsView() {
    if (mContactsView) {
        mContactsView->deleteLater();
        mContactsView = 0;
    }
    const QSqlDatabase &database = mData.dbManager().database();
    mContactsView = new ContactsView(database);
    connect(mContactsView, SIGNAL(nextPage(QList<ContactItem>)), SIGNAL(contactsNextPage(QList<ContactItem>)), Qt::UniqueConnection);
}

void TelegramService::fetchDialogsNextPage() {
    if (!mDialogsView) {
        return;
    }
    mDialogsView->getNextPage();
}

void TelegramService::fetchMessagesNextPage() {
    if (!mMessagesView) {
        return;
    }
    mMessagesView->getNextPage();
}

void TelegramService::fetchContactsNextPage() {
    if (!mContactsView) {
        return;
    }
    mContactsView->getNextPage();
}

qint32 TelegramService::getDialogIdFromMessageId(qint32 msgId) {
    return mData.getDialogIdFromMessageId(msgId);
}

MessageItem TelegramService::topMessage(qint32 dialogId) {
    return mData.topMessage(dialogId);
}

qint32 TelegramService::getUnreadCount(qint32 dialogId) {
    return mData.getUnreadCount(dialogId);
}

qint64 TelegramService::authSendInvite(const QString &phoneNumber, const QString &message) {
    return authSendInvites(QStringList(phoneNumber),message);
}
