/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TELEGRAMSERVICE_H
#define TELEGRAMSERVICE_H

#include "config.h"
#include "data.h"
#include "rawapiservice.h"
#include "models/dialogsproxy.h"
#include "models/dialogitem.h"
#include "models/contactsproxy.h"

#include <QStringList>
#include <QSortFilterProxyModel>
#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(TG_PLUGIN_LOGIC)

class TelegramService : public RawApiService
{
    Q_OBJECT
public:
    explicit TelegramService();
    ~TelegramService();

    Data &data();

public Q_SLOTS:

    qint32 ourId() const;
    QString databasePath();
    void setDatabasePath(const QString &path);
    QString downloadsPath() const;
    void setDownloadsPath(const QString &downloadsPath);

    void init();
    void shutdown();
    void sleep();
    void wake();

    void getUser(qint32 userId);
    qint64 addContact(const QString &firstName, const QString &lastName, const QString &phoneNumber);
    qint64 messagesReadHistory(InputPeer peer, qint32 maxId = 0, qint32 offset = 0, bool readContents = true);
    qint64 deleteChatHistory(InputPeer inputPeer, qint32 offset = 0, bool deleteChat = false);
    qint64 messagesSendMessage(InputPeer peer, const QString &message);
    qint64 messagesSendPhoto(InputPeer peer, const QString &filePath);
    qint64 messagesSetTyping(const InputPeer peer);
    qint64 messagesDeleteMessages(const QList<qint32> &msgIds, bool isSecret = false);
    bool downloadPhoto(qint32 messageId);
    bool downloadUserBigPhoto(qint32 userId);
    bool downloadChatBigPhoto(qint32 chatId);
    qint64 sendVideo(qint32 peerId, const QString &videoFilePath, qint32 duration = 0, qint32 width = 0, qint32 height = 0, const QString &thumbnailFilePath = "");
    bool downloadVideo(qint32 messageId);
    qint64 cancelDownload(qint32 id);
    qint64 addChatUser(qint32 chatId, qint32 userId, qint32 fwdLimit = 0);
    qint64 deleteChatUser(qint32 chatId, qint32 userId);
    qint64 mute(qint32 peerId);
    qint64 unmute(qint32 peerId);
    qint64 createSecretChat(qint32 userId);
    qint64 acceptSecretChat(qint32 chatId);
    qint64 discardSecretChat(qint32 chatId);
    qint64 setSecretChatTTL(qint32 chatId, qint32 ttl);
    qint64 sendSecretChatMessage(qint32 chatId, const QString &message, qint32 ttl = 0);
    qint64 sendSecretChatPhoto(qint32 chatId, const QString &filePath, qint32 ttl = 0);
    qint64 sendSecretChatVideo(qint32 chatId, const QString &filePath, qint32 ttl = 0);
    qint64 readSecretChatHistory(qint32 chatId);
    void deleteSecretChatHistory(qint32 dialogId, bool deleteChat = false);

    void initializeDialogsView();
    void initializeMessagesView(qint32 dialogId, qint32 peerType);
    void initializeContactsView();
    void fetchDialogsNextPage();
    void fetchMessagesNextPage();
    void fetchContactsNextPage();

    qint32 getDialogIdFromMessageId(qint32 msgId);
    MessageItem topMessage(qint32 dialogId);
    qint32 getUnreadCount(qint32 dialogId);

    qint64 authSendInvite(const QString &phoneNumber, const QString &message);

Q_SIGNALS:
    void finished();

    void databasePathChanged(QString);
    void ourIdChanged(qint32);
    void woken();
    void openChat();

    void getUserAnswer(const ContactItem &user);
    void addContactsAnswer(const QList<qint32> &importedContacts);
    void addChatUserAnswer(qint32 chatId, qint32 userId);
    void muteAnswer(qint32 peerId, bool ok);
    void unmuteAnswer(qint32 peerId, bool ok);

    void dialogsNextPage(const QList<DialogItem> &dialogs);
    void messagesNextPage(const QList<MessageItem> &messages);
    void contactsNextPage(const QList<ContactItem> &contacts);

    // models signals
    void contactsAdded(const QList<ContactItem> &contacts);
    void contactThumbnailUpdated(qint32 userId, const QString &thumbnail);
    void contactBigPhotoUpdated(qint32 userId, const QString &bigPhoto);
    void chatBigPhotoUpdated(qint32 chatId, const QString &bigPhoto);
    void dialogThumbnailUpdated(qint32 dialogId, const QString &thumbnail);
    void dialogTitleUpdated(qint32 dialogId, const QString &title);
    void dialogMuted(qint32 dialogId, bool muted);
    void dialogsMuted(NotifyPeer::NotifyPeerType notifyType, bool muted);
    void messagesAdded(const QList<MessageItem> &messages, qint32 totalCount);
    void messageAdded(const MessageItem &message);
    void outgoingMessageAdded(qint32 recipientId, const MessageItem &message);
    void messageUpdated(qint32 msgId, const MessageItem &message);
    void messagesGetFullChatAnswer(qint64 id, ChatFull chatFull, QList<Chat> chats, QList<User> users);
    void chatMarkedAsRead(qint32);
    void chatHistoryDeleted(qint32);
    void chatDeleted(qint32);
    void chatUpdated(qint32);
    void messagesMarkedAsRead(const QList<qint32> &msgIds);
    void messagesDeleted(const QList<qint32> &msgIds);
    void messagePhotoUpdated(qint32 msgId, const QString &localPath, bool isThumbnail);
    void messageVideoUpdated(qint32 msgId, const QString &videoLocalPath);
    void downloadingUpdated(qint32 id, bool downloading);
    void downloadedPercentageUpdated(qint32 id, qint8 downloadedPercentage);
    void dialogsAdded(const QList<DialogItem> &dialogs, qint32 totalCount);

    //secret chats
    void secretChatRequested(qint32 chatId, qint32 date, qint32 creatorId, qint64 accessHash);
    void secretChatCreated(qint32 chatId);
    void secretChatDiscarded(qint32 chatId);
    void secretChatUpdated(qint32 chatId);

    void unreadCountChanged(qint32 unreadCount);

    void userTyping(qint32 dialogId, const QString &whoIsTyping = QString());

private:
    // data object
    Data mData;

    // models views
    DialogsView *mDialogsView;
    MessagesView *mMessagesView;
    ContactsView *mContactsView;

    QTimer mSelfDestructTimer;

    QString copyToDownloadsPath(const QString &file);
    qint64 updateMuteUntilSetting(qint32 peerId, qint32 muteUntil);
    void updateUserProfiles();

private Q_SLOTS:
    void onMessagesAddChatUserAnswer(qint64 id, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq);
    void onContactsImportContactsAnswer(qint64 id, QList<ImportedContact> importedContacts, QList<qint64> retryContacts, QList<User> users);
    void onAuthLoggedIn();
    void onAuthLoggedOut();
    void onWoken();
    void onAccountRegisterDeviceAnswer(qint64 id, bool ok);
    void processSelfDestructMessages();
    void createCrashFile();
    void deleteCrashFile();
};

#endif // TELEGRAMCLIENT_H
