/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *      Roberto Mier
 *      Tiago Herrmann
 */
#include "tools.h"
#include "core/utils.h"
#include "config.h"
#include <iostream>
#include <stdexcept>
#include <QMimeDatabase>
#include <QMimeType>
#include <QDir>
#include <QFileInfo>
#include <QImage>

InputUser Tools::toInputUser(const ContactItem &user) {

    InputUser::InputUserType inputUserType = Tools::toInputUserType(user.classType());
    InputUser inputUser(inputUserType);

    switch (inputUserType) {
    case InputUser::typeInputUserForeign:
        inputUser.setAccessHash(user.accessHash());
    case InputUser::typeInputUserContact:
        inputUser.setUserId(user.id());
        break;
    }

    return inputUser;
}

InputUser::InputUserType Tools::toInputUserType(User::UserType userType) {

    InputUser::InputUserType inputUserType;

    switch (userType) {
    case User::typeUserContact:
        inputUserType = InputUser::typeInputUserContact;
        break;
    case User::typeUserRequest:
    case User::typeUserForeign:
        inputUserType = InputUser::typeInputUserForeign;
        break;
    case User::typeUserDeleted:
    case User::typeUserEmpty:
        inputUserType = InputUser::typeInputUserEmpty;
        break;
    case User::typeUserSelf:
        inputUserType = InputUser::typeInputUserSelf;
        break;
    default:
        inputUserType = InputUser::typeInputUserEmpty;
        break;
    }

    return inputUserType;
}

InputPeer Tools::toInputPeer(qint32 peerId, const ContactItem &user) {

    InputPeer::InputPeerType inputPeerType;
    InputPeer inputPeer;

    if (user.isEmpty()){
        inputPeerType = InputPeer::typeInputPeerChat;
        inputPeer.setChatId(peerId);
    } else {
        inputPeerType = toInputPeerType(user.classType());
        inputPeer.setUserId(user.id());

        if (inputPeerType == InputPeer::typeInputPeerForeign) {
            inputPeer.setAccessHash(user.accessHash());
        }
    }

    inputPeer.setClassType(inputPeerType);
    return inputPeer;
}

InputPeer::InputPeerType Tools::toInputPeerType(User::UserType userType) {

    InputPeer::InputPeerType inputPeerType;

    switch (userType) {
    case User::typeUserContact:
        inputPeerType = InputPeer::typeInputPeerContact;
        break;
    case User::typeUserRequest:
    case User::typeUserForeign:
        inputPeerType = InputPeer::typeInputPeerForeign;
        break;
    case User::typeUserDeleted:
    case User::typeUserEmpty:
        inputPeerType = InputPeer::typeInputPeerEmpty;
        break;
    case User::typeUserSelf:
        inputPeerType = InputPeer::typeInputPeerSelf;
        break;
    default:
        // if not found a case, take it as chat
        inputPeerType = InputPeer::typeInputPeerChat;
        break;
    }

    return inputPeerType;
}

QString Tools::generateThumbnail(const QString &videoPath) {    
    QString returnValue;

    try {
        Thumbnailer thumbnailer;
        std::string thumbnail = thumbnailer.get_thumbnail(videoPath.toStdString(), TN_SIZE_SMALL);
        returnValue = QString::fromStdString(thumbnail);
    } catch (std::runtime_error &error) {
        qCWarning(TG_PLUGIN_LOGIC) << "Error building thumbnail:" << error.what();
    }

    return returnValue;
}

QString Tools::getExtensionFromMimeType(const QString &mimeType) {
    QString extension;
    QMimeType mt = QMimeDatabase().mimeTypeForName(mimeType);
    QStringList suffixes = mt.suffixes();
    if (suffixes.length() > 0) {
        extension = QString(".%1").arg(suffixes.at(0));
    }
    return extension;
}

qint64 Tools::generateRandomId() {
    qint64 randomId;
    Utils::randomBytes(&randomId, 8);
    return randomId;
}

qint32 Tools::requestIdToMsgId(qint64 requestId) {
    qint32 lessSignificative = requestId;
    qint32 mostSignificative = (qint32) (requestId >> 32);
    return mostSignificative + lessSignificative;
}

MessageMedia::MessageMediaType Tools::toMessageMediaType(DecryptedMessageMedia::DecryptedMessageMediaType decryptedMessageMediaType) {

    MessageMedia::MessageMediaType messageMediaType;

    switch (decryptedMessageMediaType) {
    case DecryptedMessageMedia::typeDecryptedMessageMediaPhoto:
        messageMediaType = MessageMedia::typeMessageMediaPhoto;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaVideo_layer8:
    case DecryptedMessageMedia::typeDecryptedMessageMediaVideo:
        messageMediaType = MessageMedia::typeMessageMediaVideo;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaGeoPoint:
        messageMediaType = MessageMedia::typeMessageMediaGeo;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaContact:
        messageMediaType = MessageMedia::typeMessageMediaContact;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaDocument:
        messageMediaType = MessageMedia::typeMessageMediaDocument;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaAudio_layer8:
    case DecryptedMessageMedia::typeDecryptedMessageMediaAudio:
        messageMediaType = MessageMedia::typeMessageMediaAudio;
        break;
    case DecryptedMessageMedia::typeDecryptedMessageMediaEmpty:
        messageMediaType = MessageMedia::typeMessageMediaEmpty;
        break;
    default:
        messageMediaType = MessageMedia::typeMessageMediaEmpty;
        break;
    }

    return messageMediaType;
}

qint32 Tools::getBelongingDialogId(const Message &message) {
    qint32 dialogId = message.fromId();
    bool isChat = message.toId().classType() == Peer::typePeerChat;
    if (isChat || message.out()) {
        dialogId = isChat ? message.toId().chatId() : message.toId().userId();
    }
    return dialogId;
}

QString Tools::pngToJpeg(const QString &pngFile, qint32 quality) {
    QImage image;
    image.load(pngFile);
    QFileInfo fileInfo(pngFile);
    QString jpegFile = fileInfo.path() + QDir::separator() + fileInfo.baseName() + ".jpg";
    image.save(jpegFile, "JPEG");
    return jpegFile;
}

void Tools::resizePhotoIfNeeded(const QString &photoFilePath) {
    QImage image;
    image.load(photoFilePath);
    qint32 originalWidth = image.width();
    qint32 originalHeight = image.height();

    qint32 max = originalWidth > originalHeight ? originalWidth : originalHeight;

    if (max > PHOTO_MAX_DIMENSION) {
        QImage small;
        if (originalWidth == max) {
            small = image.scaled(PHOTO_MAX_DIMENSION, originalHeight, Qt::KeepAspectRatio);
        } else if (originalHeight == max) {
            small = image.scaled(originalWidth, PHOTO_MAX_DIMENSION, Qt::KeepAspectRatio);
        }
        qCDebug(TG_PLUGIN_LOGIC) << "original size " << originalWidth << "x" << originalHeight << "reduced to"
                                    << small.width() << "x" << small.height();
        small.save(photoFilePath);
    }
}
