/*
 * Copyright 2014 Canonical Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "data.h"
#include "dbmanager.h"
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QVariant>
#include <QDir>
#include <QTime>
#include "rawapiclient.h"
#include "telegramservice.h"
#include "tools.h"

Q_LOGGING_CATEGORY(TG_PLUGIN_PROFILING, "tg.plugin.profiling")

Data::Data(QObject *parent) :
    QObject(parent),
    mServiceRef(parent),
    mFilesAvgTime(0),
    mFilesCount(0),
    mFilesTotalTime(0),
    mLastUnreadCount(0) {

    // Priorities for thumbnail sizes are m, b, s, a, x, c, y, d, w
    mPhotoSmallPriorities.insert("m", 1);
    mPhotoSmallPriorities.insert("b", 2);
    mPhotoSmallPriorities.insert("s", 3);
    mPhotoSmallPriorities.insert("a", 4);
    mPhotoSmallPriorities.insert("x", 5);
    mPhotoSmallPriorities.insert("c", 6);
    mPhotoSmallPriorities.insert("y", 7);
    mPhotoSmallPriorities.insert("d", 8);
    mPhotoSmallPriorities.insert("w", 9);

    // Priorities for big photo sizes are w, d, y, x, c, m, b, a, s
    mPhotoBigPriorities.insert("w", 1);
    mPhotoBigPriorities.insert("d", 2);
    mPhotoBigPriorities.insert("y", 3);
    mPhotoBigPriorities.insert("x", 4);
    mPhotoBigPriorities.insert("c", 5);
    mPhotoBigPriorities.insert("m", 6);
    mPhotoBigPriorities.insert("b", 7);
    mPhotoBigPriorities.insert("a", 8);
    mPhotoBigPriorities.insert("s", 9);

    // See https://core.telegram.org/constructor/photoSize#thumbnail-type-and-its-sizes to know details about every size
}

Data::~Data() {
    Q_FOREACH (IncomingFile *file, mFiles) {
        if (file) {
            delete file;
            file = 0;
        }
    }
    Q_FOREACH (QTime *time, mFileTimes) {
        if (time) {
            delete time;
            time = 0;
        }
    }
    Q_FOREACH (qint64 fileId, mActiveDownloads.values()) {
        onUploadCancelFileAnswer(fileId, true);
    }
}

DbManager &Data::dbManager() {
    return mDbManager;
}

// ### Client state
void Data::onUpdatesGetStateAnswer(qint64, qint32 pts, qint32 qts, qint32 date, qint32 seq, qint32 unreadCount) {

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdatesGetStateAnswer";

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);

    UpdatesState clientState = getState();

    if ( (clientState.seq() > 0 && clientState.seq() < seq) || (seq == 1 && clientState.pts() < pts) ){
        service->updatesGetDifference(clientState.pts(), clientState.date(), clientState.qts());
    }

    clientState.setPts(pts);
    clientState.setQts(qts);
    clientState.setDate(date);
    clientState.setSeq(seq);
    clientState.setUnreadCount(unreadCount);

    if (!insertOrUpdateState(clientState)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating client state in local database";
        return;
    }

    surfaceUnreadCount();

    qCDebug(TG_PLUGIN_PROFILING) << "onUpdatesGetStateAnswer - end:" << time.elapsed() << "ms";
}

void Data::onUpdatesGetDifferenceAnswer(qint64, qint32 date, qint32 seq) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdatesGetDifferenceAnswer (state only)";

    UpdatesState clientState = getState();

    clientState.setDate(date);
    clientState.setSeq(seq);

    if (!insertOrUpdateState(clientState)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating client state in local database";
        return;
    }

    qCDebug(TG_PLUGIN_PROFILING) << "onUpdatesGetDifferenceAnswer - end:" << time.elapsed() << "ms";
}

void Data::onUpdatesGetDifferenceAnswer(qint64, const QList<Message> &messages, const QList<SecretChatMessage> &secretChatMessages, const QList<Update> &otherUpdates, const QList<Chat> &chats, const QList<User> &users, const UpdatesState &state, bool isIntermediateState) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdatesGetDifferenceAnswer";

    mDbManager.beginTransaction();

    // Process first the updates before messages, because they could include the new chat creation, for instance
    // and the messages received could be part of that new chat
    Q_FOREACH (Update update, otherUpdates) {
        if (!processUpdate(update, users, chats)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error processing updates differences";
            return;
        }
    }

    Q_FOREACH (const Message &message, messages) {

        qint32 dialogId = Tools::getBelongingDialogId(message);

        // search for message related dialog, but use it after inserting message in db
        bool newDialog = !existsDialog(dialogId);

        if (!insertOrUpdateMessage(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating message in local database";
            return;
        }

        if (!executeAction(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not execute message related action";
            return;
        }

        MessageItem messageItem = getMessage(message.id());
        Q_EMIT messageAdded(messageItem);

        if (newDialog) {
             // Ask telegram servers for info about this just created dialog where it will be included the phone number
             // of the unknown contact
            // this operation is done after insertMessage transaction to ensurance the dialog info coming from server
            // updates db after creating the basic dialog entry related to this message
            TelegramService *service = static_cast<TelegramService *>(mServiceRef);
            service->messagesGetDialogs(0, dialogId, 1);
        }
    }

    Q_FOREACH (const SecretChatMessage &secretChatMessage, secretChatMessages) {

        if (secretChatMessage.isServiceMessage()) {
            processSecretChatServiceMessage(secretChatMessage);
            continue;
        } else if (!existsDialog(secretChatMessage.chatId())) {
            continue;
        }

        if (!existsSecretChatMessage(secretChatMessage)) {
            if (!insertSecretChatMessage(secretChatMessage)) {
                mDbManager.rollbackTransaction();
                qCCritical(TG_PLUGIN_LOGIC) << "Error inserting secret chat message";
                return;
            }

            qint32 chatId = secretChatMessage.chatId();
            if (!increaseUnreadCount(chatId, 1)) {
                mDbManager.rollbackTransaction();
                qCCritical(TG_PLUGIN_LOGIC) << "Could not increase unread count for dialog" << chatId;
                return;
            }

            qint32 messageId = secretChatMessage.decryptedMessage().randomId();
            MessageItem messageItem = getMessage(messageId);

            Q_EMIT messageAdded(messageItem);
        }
    }

    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned chat in local database";
            return;
        }
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned user in local database";
            return;
        }
    }

    if (!updateState(state)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating client state in local database";
        return;
    }

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);

    if (secretChatMessages.count() > 0) {
        qint32 qts = state.qts();
        service->messagesReceivedQueue(qts);
    }

    if (isIntermediateState) {
        service->updatesGetDifference(state.pts(), state.date(), state.qts());
    }

    mDbManager.finishTransaction();
    Q_FOREACH(const Chat &chat, chats) {
        Q_EMIT chatUpdated(chat.id());
    }

    qCDebug(TG_PLUGIN_PROFILING) << "onUpdatesGetDifferenceAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesGetDialogsAnswer(qint64, qint32 totalCount, const QList<Dialog> &dialogs, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesGetDialogsAnswer";

    mDbManager.beginTransaction();

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting dialogs mentioned users in local database";
            return;
        }
    }

    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting dialogs mentioned chats in local database";
            return;
        }
    }

    Q_FOREACH (const Message &message, messages) {
        if (!insertOrUpdateMessage(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting dialogs top messages in local database";
            return;
        }
    }

    Q_FOREACH (const Dialog &dialog, dialogs) {
        if (!insertOrUpdateDialog(dialog)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting dialogs main info in local database";
            return;
        }
    }

    QList<DialogItem> dlgs;

    // create maps for mentioned users and chats to be easily referenced when composing dialog items.
    // Afterwards, use this map creation to compose contacts list also
    QMap<qint32, const User*> mentionedUsers;
    for (qint32 i=0; i<users.size(); i++) {
        const User &user = users.at(i);
        mentionedUsers.insert(user.id(), &user);
    }

    QMap<qint32, const Chat*> mentionedChats;
    for (qint32 i=0; i<chats.size(); i++) {
        const Chat &chat = chats.at(i);
        mentionedChats.insert(chat.id(), &chat);
    }

    for (qint32 i = 0; i < dialogs.size(); ++i) {
        DialogItem item;
        Dialog thisdialog = dialogs.at(i);
        Message thismessage = messages.at(i);
        Peer peer = thisdialog.peer();
        FileLocation photoToDownload;
        bool isUserPhoto = false;
        if (peer.classType() == Peer::typePeerChat) {
            const Chat *c = mentionedChats.value(peer.chatId());
            item.setIsChat(true);
            item.setId(c->id());
            if (c->classType() != Chat::typeChatEmpty) {
                item.setName(c->title());
                if (c->classType() != Chat::typeChatForbidden) {
                    photoToDownload = c->photo().photoSmall();
                    isUserPhoto = false;
                }
            }
            item.setPeerType(InputPeer::typeInputPeerChat);
        } else if (peer.classType() == Peer::typePeerUser) {
            const User *user = mentionedUsers.value(peer.userId());
            item.setIsChat(false);
            item.setId(peer.userId());
            QString fullName = user->firstName() + " " + user->lastName();
            item.setName(fullName);
            item.setPeerType(user->classType());
            photoToDownload = user->photo().photoSmall();
            isUserPhoto = true;

            //TODO Roberto: this line is needed for this release to update the title of old secret chats
            // After this release, this line can be commented out or removed
            notifySecretChatsTitleUpdated(peer.userId(), fullName);
        }
        item.setDate(thismessage.date());
        item.setUnreadCount(thisdialog.unreadCount());
        item.setPeerId(item.id());

        MessageItem messageItem = getMessage(thismessage.id());
        item.setTopMessage(messageItem);

        QString thumbnail = getThumbnail(item.id(), isUserPhoto);
        if (!thumbnail.isEmpty()) {
            item.setThumbnail(thumbnail);
            if (isUserPhoto) {
                Q_EMIT contactThumbnailUpdated(item.id(), thumbnail);
            }
        }

        dlgs.append(item);
    }

    mDbManager.finishTransaction();

    Q_EMIT dialogsAdded(dlgs, totalCount);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesGetDialogsAnswer - elapsed time:" << time.elapsed(); "ms";
}

void Data::onUploadGetFileAnswer(qint64 fileId, const StorageFileType &type, qint32 mtime, QByteArray bytes, qint32 partId, qint32 downloaded, qint32 total) {

    IncomingFile *incomingFile = mFiles.value(fileId, 0);
    if (incomingFile) {

        qint32 id = incomingFile->id();
        IncomingFile::IncomingFileType incomingFileType = incomingFile->incomingFileType();

        QFile &file = incomingFile->file();
        if (!file.isOpen()) {

            QTime *time = new QTime;
            mFileTimes.insert(fileId, time);
            time->start();

            QString extension;
            switch (type.classType()) {
            case StorageFileType::typeStorageFileGif:
                extension = ".gif";
                break;
            case StorageFileType::typeStorageFileJpeg:
                extension = ".jpg";
                break;
            case StorageFileType::typeStorageFilePng:
                extension = ".png";
                break;
            case StorageFileType::typeStorageFileMov:
                extension = ".mov";
                break;
            case StorageFileType::typeStorageFileMp3:
                extension = ".mp3";
                break;
            case StorageFileType::typeStorageFileMp4:
                extension = ".mp4";
                break;
            case StorageFileType::typeStorageFilePdf:
                extension = ".pdf";
                break;
            case StorageFileType::typeStorageFileUnknown:
                extension = ".unk";
                break;
            case StorageFileType::typeStorageFileWebp:
                extension = ".webp";
                break;
            }

            // if type of file is partial, the previously provided extension (before starting download) will be used
            if (extension.length() > 0) {
                incomingFile->setExtension(extension);
            }

            incomingFile->setModificationTime(mtime);

            // filePath is calculated like: <downloads_folder>/<id>_<mtime>_<size_type>.<extension>
            // (you can check size_type values at https://core.telegram.org/constructor/photoSize, in "Thumbnail type and its sizes")
            file.setFileName(incomingFile->filePath());

            // write the file only if it doesn't exists (if exists, is the same version as previously downloaded, because
            // in its name is the id + modificationTime)
            if (!file.exists()) {

                // create dir path if not exists yet
                QFileInfo fileInfo(file);
                QDir fileDir = fileInfo.absoluteDir();
                QString dirPath = fileDir.absolutePath();
                if (!fileDir.exists()) {
                    if (QDir::root().mkpath(dirPath)) {
                        qCDebug(TG_PLUGIN_LOGIC) << dirPath << "downloads directory created";
                    } else {
                        qCWarning(TG_PLUGIN_LOGIC) << "error creating downloads directory: " << dirPath;
                        return;
                    }
                }

                if (incomingFileType == IncomingFile::UserThumbnail || incomingFileType == IncomingFile::ChatThumbnail ) {
                    // remove any previous profile file found at this path for this id to avoid
                    // having garbage profile photos in folder
                    QStringList filters;
                    filters << QString::number(id) + "_*";
                    Q_FOREACH (QString fileNameToDelete, fileDir.entryList(filters, QDir::Files)) {
                        QFile fileToDelete(dirPath + QDir::separator() + fileNameToDelete);
                        if (fileToDelete.remove()) {
                            qCDebug(TG_PLUGIN_LOGIC) << "removed previous profile photo at " << fileToDelete.fileName();
                        }
                    }
                }

                file.open(QIODevice::WriteOnly);
            }
        }

        // if file has been open now or in a previous download step, write bytes and check if it must be closed
        if (file.isOpen()) {

            file.write(bytes);

            qCDebug(TG_PLUGIN_LOGIC) << "-----------";
            qCDebug(TG_PLUGIN_LOGIC) << "fileId:" << fileId;
            qCDebug(TG_PLUGIN_LOGIC) << "modificationTime:" << mtime;
            qCDebug(TG_PLUGIN_LOGIC) << "partId:" << partId;
            qCDebug(TG_PLUGIN_LOGIC) << "size:" << bytes.length();
            qCDebug(TG_PLUGIN_LOGIC) << "downloaded:" << downloaded;
            qCDebug(TG_PLUGIN_LOGIC) << "total:" << total;
            qCDebug(TG_PLUGIN_LOGIC) << "expected Size:" << incomingFile->expectedSize();

            if (total > 0) {
                qint8 percentage = (qint64)downloaded * 100 / total;
                qCDebug(TG_PLUGIN_LOGIC) << "calculated percentage:" << percentage << "%";

                if (!incomingFile->isThumbnail()) {
                    Q_EMIT downloadedPercentageUpdated(id, percentage);
                }
            }

            if (total > 0 && file.size() >= total) {
                file.close();
                qCDebug(TG_PLUGIN_LOGIC)  << "file" << file.fileName() << "write completed";
            }

            qCDebug(TG_PLUGIN_LOGIC) << "-----------";
        }

        // if file has just been closed or never has been open (because already existed), update db and emit signals
        if (!file.isOpen()) {

            FileLocation serverLocation = incomingFile->serverLocation();
            QString filePath = incomingFile->filePath();
            bool thumbnail = incomingFile->isThumbnail();
            qint32 fileLocationId = incomingFile->fileLocationId();

            switch (incomingFileType) {
            case IncomingFile::UserThumbnail: {
                if (insertOrUpdateUserThumbnail(id, filePath, serverLocation) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update thumbnail local path" << filePath << "for user" << id;
                    break;
                }
                Q_EMIT dialogThumbnailUpdated(id, filePath);
                Q_EMIT contactThumbnailUpdated(id, filePath);

                notifySecretChatsThumbnailUpdated(id, filePath);

                break;
            }
            case IncomingFile::UserPhoto: {
                if (insertOrUpdateUserBigPhoto(id, filePath, serverLocation) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update big photo local path" << filePath << "for user" << id;
                    break;
                }
                Q_EMIT contactBigPhotoUpdated(id, filePath);
                break;
            }
            case IncomingFile::ChatThumbnail: {
                if (insertOrUpdateChatThumbnail(id, filePath, serverLocation) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update thumbnail local path" << filePath << "for chat" << id;
                    break;
                }
                Q_EMIT dialogThumbnailUpdated(id, filePath);
                break;
            }
            case IncomingFile::ChatPhoto: {
                if (insertOrUpdateChatBigPhoto(id, filePath, serverLocation) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update big photo local path" << filePath << "for chat" << id;
                    break;
                }
                Q_EMIT chatBigPhotoUpdated(id, filePath);
                break;
            }
            case IncomingFile::MessageActionPhoto: {

                // check if must execute the message action
                if (mMessageIdsToUpdatePhoto.take(id)) {
                    qint32 dialogId = getDialogIdFromMessageId(id);
                    if (dialogId == 0 || insertOrUpdateChatThumbnail(dialogId, filePath, serverLocation) <= 0) {
                        qCCritical(TG_PLUGIN_LOGIC) << "Could not update thumbnail local path" << filePath << "for dialog" << dialogId;
                        break;
                    }
                    Q_EMIT dialogThumbnailUpdated(dialogId, filePath);
                }
                // Fall down to message photo
            }
            case IncomingFile::MessagePhoto: {
                if (updateFileLocation(fileLocationId, filePath) <= 0){
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update local path" << filePath << "for photo in message" << id;
                    break;
                }

                QList<qint32> messageIds = getMediaMessages(incomingFile->mediaId());
                Q_FOREACH (qint32 messageId, messageIds) {
                    Q_EMIT messagePhotoUpdated(messageId, filePath, thumbnail);
                    if (!incomingFile->isThumbnail()) {
                        Q_EMIT downloadedPercentageUpdated(messageId, 100);
                    }
                }
                break;
            }
            case IncomingFile::EncryptedMessagePhoto: {
                qCDebug(TG_PLUGIN_LOGIC) << "Encrypted message photo written in path" << filePath;
                if (updateFileLocation(fileLocationId, filePath) <= 0){
                   qCCritical(TG_PLUGIN_LOGIC) << "Could not update local path" << filePath << "for photo in message" << id;
                   break;
                }

                Q_EMIT messagePhotoUpdated(id, filePath, thumbnail);
                if (!incomingFile->isThumbnail()) {
                   Q_EMIT downloadedPercentageUpdated(id, 100);
                }
                break;
            }
            case IncomingFile::EncryptedMessageVideo: {
               qCDebug(TG_PLUGIN_LOGIC) << "Encrypted message video written in path" << filePath;
               qint64 mediaId = incomingFile->mediaId();
               if (updateVideo(mediaId, filePath) <= 0) {
                   qCCritical(TG_PLUGIN_LOGIC) << "Could not update local path" << filePath << "for video in message" << id;
                   break;
               }

               Q_EMIT messageVideoUpdated(id, filePath);
               if (!incomingFile->isThumbnail()) {
                   Q_EMIT downloadedPercentageUpdated(id, 100);
               }
               break;
            }
            case IncomingFile::MessageVideo: {
                qint64 mediaId = incomingFile->mediaId();
                if (updateVideo(mediaId, filePath) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update local path" << filePath << "for video in message" << id;
                    break;
                }

                QList<qint32> messageIds = getMediaMessages(incomingFile->mediaId());
                Q_FOREACH (qint32 messageId, messageIds) {
                    Q_EMIT messageVideoUpdated(messageId, filePath);
                    if (!incomingFile->isThumbnail()) {
                        Q_EMIT downloadedPercentageUpdated(messageId, 100);
                    }
                }
                break;
            }
            }

            Q_EMIT downloadingUpdated(id, false);

            mFiles.remove(fileId);
            mActiveDownloads.remove(id);
            delete incomingFile;
            incomingFile = 0;

            //measure time spent in having file ready
            QTime *time = mFileTimes.take(fileId);
            if (time) {
                qint32 elapsed = time->elapsed();
                qCDebug(TG_PLUGIN_PROFILING) << "File with id" << fileId << "downloaded in" << elapsed << "milliseconds";
                delete time;

                mFilesCount++;
                mFilesTotalTime += elapsed;
                mFilesAvgTime = mFilesTotalTime / mFilesCount;
                qCDebug(TG_PLUGIN_PROFILING) << "number of files:" << mFilesCount;
                qCDebug(TG_PLUGIN_PROFILING) << "total files download time:" << mFilesTotalTime << "ms";
                qCDebug(TG_PLUGIN_PROFILING) << "file average download time:" << mFilesAvgTime << "ms";
                qCDebug(TG_PLUGIN_PROFILING) << "-----------";
            }
        }
    }
}

void Data::onUploadSendFileAnswer(qint64 fileId, qint32 partId, qint32 downloaded, qint32 total) {

    qint32 id = mActiveUploads.value(fileId);
    if (id) {
        qint8 percentage = downloaded * 100 / total;
        Q_EMIT downloadedPercentageUpdated(id, percentage);
        if (percentage >= 100) {
            mActiveUploads.remove(fileId);
        }
    }
}

void Data::onUsersGetUsersAnswer(qint64, QList<User> users) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onUsersGetUsersAnswer";

    // though only requested user data, try to insert it in db if not exists
    QList<ContactItem> contactItems;

    mDbManager.beginTransaction();

    Q_FOREACH (const User &user, users) {

        // check previous username before update user db register
        bool changedUsername = hasChangedUsername(user);

        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error adding new contact to local database";
            return;
        }

        ContactItem contactItem(user);

        FileLocation photoToDownload = user.photo().photoSmall();
        bool newPhotoFetched = getProfilePhotoFileFromServer(photoToDownload, user.id(), true, true);
        if (!newPhotoFetched && (photoToDownload.classType() != FileLocation::typeFileLocationUnavailable)){
            QString thumbnail = getThumbnail(user.id(), true);
            contactItem.setThumbnail(thumbnail);
        }

        if (user.classType() == User::typeUserContact) {
            contactItems.append(contactItem);
        }


        if (changedUsername) {
            QString fullName = user.firstName() + " " + user.lastName();
            Q_EMIT dialogTitleUpdated(user.id(), fullName);
            notifySecretChatsTitleUpdated(user.id(), fullName);
        }

        Q_EMIT userInfo(contactItem);
    }

    mDbManager.finishTransaction();

    Q_EMIT contactsAdded(contactItems);

    qCDebug(TG_PLUGIN_PROFILING) << "onUsersGetUsersAnswer - end:" << time.elapsed() << "ms";
}

// TODO karni: Refactor other methods to re-use these and catch exceptions.

void Data::updateChats(const QList<Chat> &chats) {
    Q_FOREACH(const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            throw "Could not update chats.";
        }
    }
}

void Data::updateUsers(const QList<User> &users) {
    Q_FOREACH(const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            throw "Could not update users.";
        }
    }
}

void Data::updateChatPhoto(qint32 chatId, const Photo &photo) {
    if (!hasThumbnailDownloaded(photo.id())) {
        downloadThumbnail(chatId, photo, IncomingFile::ChatThumbnail);
    } else {
        QString thumbnail = getThumbnail(chatId, false);
        Q_EMIT dialogThumbnailUpdated(chatId, thumbnail);
    }
}

void Data::updateChatFull(const ChatFull &chatFull) {
    updateChatPhoto(chatFull.id(), chatFull.chatPhoto());
    updateChatParticipants(chatFull.participants());
}

void Data::onMessagesGetFullChatAnswer(qint64, ChatFull chatFull, QList<Chat> chats, QList<User> users) {
    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesGetFullChatAnswer";
    QTime time;
    time.start();

    mDbManager.beginTransaction();
    try {
        updateUsers(users);
        updateChats(chats);
        updateChatFull(chatFull);
        mDbManager.finishTransaction();
        Q_EMIT chatUpdated(chatFull.id());
    } catch (char const* e) {
        mDbManager.rollbackTransaction();
    }
    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesGetFullChatAnswer - elapsed time" << time.elapsed() << "ms";
}

void Data::onContactsImportContactsAnswer(qint64, QList<ImportedContact>, QList<qint64>, QList<User> users) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onContactsImportContactsAnswer";

    // emit signal with updated contacts
    QList<ContactItem> contactItems;

    mDbManager.beginTransaction();

    Q_FOREACH (const User &user, users) {

        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error adding new contact to local database";
            return;
        }

        ContactItem contactItem(user);

        FileLocation photoToDownload = user.photo().photoSmall();
        bool newPhotoFetched = getProfilePhotoFileFromServer(photoToDownload, user.id(), true, true);
        if (!newPhotoFetched && (photoToDownload.classType() != FileLocation::typeFileLocationUnavailable)){
            QString thumbnail = getThumbnail(user.id(), true);
            contactItem.setThumbnail(thumbnail);
        }

        if (user.classType() == User::typeUserContact) {
            contactItems.append(contactItem);
        }
    }

    mDbManager.finishTransaction();

    Q_EMIT contactsAdded(contactItems);

    qCDebug(TG_PLUGIN_PROFILING) << "onContactsImportContactAnswer - end:" << time.elapsed() << "ms";
}

void Data::onContactsGetContactsAnswer(qint64, bool modified, const QList<Contact> &, const QList<User> &users) {

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onContactsGetContactsAnswer";

    if (modified) {

        QList<ContactItem> contactItems;

        mDbManager.beginTransaction();

        Q_FOREACH (const User &user, users) {
            if (!insertOrUpdateUser(user)) {
                qCWarning(TG_PLUGIN_LOGIC) << "Error inserting contact in local database. Continue with the rest of contacts.";
            }

            ContactItem contactItem(user);

            FileLocation photoToDownload = user.photo().photoSmall();
            bool newPhotoFetched = getProfilePhotoFileFromServer(photoToDownload, user.id(), true, true);
            if (!newPhotoFetched && (photoToDownload.classType() != FileLocation::typeFileLocationUnavailable)){
                QString thumbnail = getThumbnail(user.id(), true);
                contactItem.setThumbnail(thumbnail);
            }

            contactItems.append(contactItem);
        }

        mDbManager.finishTransaction();

        Q_EMIT contactsAdded(contactItems);
    }

    qCDebug(TG_PLUGIN_PROFILING) << "onContactsGetContactsAnswer - elapsed time:" << time.elapsed() << "ms";
}

void Data::onMessagesGetHistoryAnswer(qint64, qint32 totalCount, const QList<Message> &messages, const QList<Chat> &chats, const QList<User> &users) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesGetHistoryAnswer";

    QList<MessageItem> messageItems;

    mDbManager.beginTransaction();

    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned chat in local database";
            return;
        }
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned user in local database";
            return;
        }
    }

    Q_FOREACH (const Message &message, messages) {
        if (!insertOrUpdateMessage(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting dialogs top messages in local database";
            return;
        }

        MessageItem messageItem = getMessage(message.id());
        messageItems.append(messageItem);
    }

    mDbManager.finishTransaction();

    Q_EMIT messagesAdded(messageItems, totalCount);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesGetHistoryAnswer - elapsed time" << time.elapsed() << "ms";
}

void Data::onMessagesReadHistoryAnswer(qint64, qint32 pts, qint32 seq, qint32 offset) {

    //TODO Roberto: implement offset management for affected history

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesReadHistoryAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating client state in local database after reading dialog";
        return;
    }

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesReadHistoryAnswer - elapsed time" << time.elapsed() << "ms";
}

void Data::onMessagesDeleteHistoryAnswer(qint64 id, qint32 pts, qint32 seq, qint32 offset) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesDeleteHistoryAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    qint32 chatId = mChatsToClearHistory.take(id);
    if (chatId == 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not complete dialog history deletion!";
        return;
    }
    deleteMessages(chatId);

    qint32 deleteChatId = mChatsToDelete.take(id);
    if (deleteChatId != 0) {
        qCDebug(TG_PLUGIN_LOGIC) << "deleting chat:" << deleteChatId;
        deleteDialog(deleteChatId);
    }

    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state";
        mDbManager.rollbackTransaction();
        return;
    }

    mDbManager.finishTransaction();

    Q_EMIT chatHistoryDeleted(chatId);
    if (deleteChatId != 0) {
        Q_EMIT chatDeleted(deleteChatId);
    }
}

void Data::onMessagesSendEncryptedAnswer(qint64 requestId, qint32 date, EncryptedFile encryptedFile) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesSendEncryptedAnswer";

    qint32 messageId = getMessageId(requestId);

    if (!updateMessage(messageId, messageId, date)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating database after a sent text message to a secret chat";
        return;
    }

    MessageItem messageItem = getMessage(messageId);
    Q_EMIT messageUpdated(messageId, messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesSendEncryptedAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesSendEncryptedFileAnswer(qint64 requestId, qint32 date, EncryptedFile encryptedFile) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesSendEncryptedFileAnswer";

    qint32 messageId = getMessageId(requestId);

    if (!updateMessage(messageId, messageId, date)) {
        qCCritical(TG_PLUGIN_LOGIC) << "onMessagesSendEncryptedFileAnswer: failed to update cached message";
        return;
    }

    MessageItem messageItem = getMessage(messageId);
    Q_EMIT messageUpdated(messageId, messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesSendEncryptedFileAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesSendMessageAnswer(qint64 requestId, qint32 msgId, qint32 date, qint32 pts, qint32 seq, const QList<ContactsLink> &links) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesSendMessageAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    // generate message identifier stored in db
    qint32 oldMsgId = Tools::requestIdToMsgId(requestId);
    if (!updateMessage(oldMsgId, msgId, date)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating database after a sent text message";
        return;
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link in database after a sent text message";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    clientState.setDate(date);
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after sending a text message";
        return;
    }

    MessageItem messageItem = getMessage(msgId);

    mDbManager.finishTransaction();

    Q_EMIT messageUpdated(oldMsgId, messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesSendMessageAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesDeleteMessagesAnswer(qint64, const QList<qint32> msgIds) {
    deleteMessages(msgIds);
}

void Data::onMessagesSendPhotoAnswer(qint64 requestId, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesSendPhotoAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned chat in local database";
            return;
        }
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned user in local database";
            return;
        }
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    // generate message identifier stored in db
     qint32 oldMsgId = Tools::requestIdToMsgId(requestId);
    if (!updateMessage(oldMsgId, message)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update already sent photo message";
        return;
    }

    clientState.setDate(message.date());
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after sending a photo message";
        return;
    }

    MessageItem messageItem = getMessage(message.id());

    mDbManager.finishTransaction();

    Q_EMIT downloadingUpdated(oldMsgId, false);
    Q_EMIT messageUpdated(oldMsgId, messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesSendPhotoAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesSendVideoAnswer(qint64 requestId, const Message &message, const QList<Chat> &chats, const QList<User> &users, const QList<ContactsLink> &links, qint32 pts, qint32 seq) {

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesSendVideoAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    qint32 oldMsgId = Tools::requestIdToMsgId(requestId);
    if (!updateMessage(oldMsgId, message)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update database message after sending a video";
        return;
    }

    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned chat in local database";
            return;
        }
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error inserting or updating mentioned user in local database";
            return;
        }
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    clientState.setDate(message.date());
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after sending a video message";
        return;
    }

    MessageItem messageItem = getMessage(message.id());

    mDbManager.finishTransaction();

    Q_EMIT downloadingUpdated(oldMsgId, false);
    Q_EMIT messageUpdated(oldMsgId, messageItem);
}

void Data::onMessagesCreateChatAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesCreateChatAnswer";

    if (chats.length() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Fatal error when creating a chat. The number of server returned chats parameter is" << chats.length();
        return;
    }

    const Chat &newChat = chats.at(0);

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    if (!insertChat(newChat)) {
        qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to insert the new chat in database";
        mDbManager.rollbackTransaction();
        return;
    }

    Dialog newDialog;
    Peer peer(Peer::typePeerChat);
    peer.setChatId(newChat.id());
    newDialog.setPeer(peer);
    newDialog.setUnreadCount(0);

    if (!insertDialog(newDialog)) {
        qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to insert the new dialog in database";
        mDbManager.rollbackTransaction();
        return;
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to insert in database the users information of chat participants";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    if (!insertMessage(message)) {
        qCCritical(TG_PLUGIN_LOGIC) << "An error has happened when trying to insert first top message of just created dialog";
        mDbManager.rollbackTransaction();
        return;
    }

    clientState.setDate(message.date());
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after creating a chat";
        return;
    }


    // get database just created data and emit signal to update models
    QList<DialogItem> dialogs;

    DialogItem dialogItem = getDialog(newChat.id());
    dialogItem.setName(newChat.title());
    FileLocation photoToDownload = newChat.photo().photoSmall();
    dialogItem.setPeerType(InputPeer::typeInputPeerChat);

    if (message.classType() != Message::typeMessageEmpty) {
        MessageItem messageItem = getMessage(message.id());
        dialogItem.setTopMessage(messageItem);
    }

    bool newPhotoFetched = getProfilePhotoFileFromServer(photoToDownload, dialogItem.id(), false, true);
    if (!newPhotoFetched && (photoToDownload.classType() != FileLocation::typeFileLocationUnavailable)){
        QString thumbnail = getThumbnail(dialogItem.id(), false);
        dialogItem.setThumbnail(thumbnail);
    }

    dialogs << dialogItem;

    mDbManager.finishTransaction();

    Q_EMIT dialogsAdded(dialogs);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesCreateChatAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesAddChatUserAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesAddChatUserAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    if (!insertOrUpdateMessage(message)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error inserting add chat user generated message";
        return;
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating added to group user in database";
            return;
        }
    }

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating modified chat in database";
            return;
        }
        // Update chat participants.
        service->messagesGetFullChat(chat.id());
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    clientState.setDate(message.date());
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after adding a user to a chat";
        return;
    }

    // send signal with the action related message
    MessageItem messageItem = getMessage(message.id());

    mDbManager.finishTransaction();

    Q_FOREACH (const Chat &chat, chats) {
        Q_EMIT chatUpdated(chat.id());
    }

    Q_EMIT messageAdded(messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesAddChatUsersAnswer - end:" << time.elapsed() << "ms";
}

void Data::onMessagesDeleteChatUserAnswer(qint64 id, Message message, QList<Chat> chats, QList<User> users, QList<ContactsLink> links, qint32 pts, qint32 seq) {

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onMessagesDeleteChatUserAnswer";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    if (!insertOrUpdateMessage(message)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error inserting delete chat user generated message";
        return;
    }

    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating added to group user in database";
            return;
        }
    }

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating modified chat in database";
            return;
        }
        // Update chat participants.
        service->messagesGetFullChat(chat.id());
    }

    Q_FOREACH(const ContactsLink &contactsLink, links) {
        if (!updateContactsLink(contactsLink)) {
            qCCritical(TG_PLUGIN_LOGIC) << "An error happened when trying to update a contact link";
            mDbManager.rollbackTransaction();
            return;
        }
    }

    clientState.setDate(message.date());
    clientState.setPts(pts);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after adding a user to a chat";
        return;
    }

    // send signal with the action related message
    MessageItem messageItem = getMessage(message.id());

    mDbManager.finishTransaction();

    Q_FOREACH (const Chat &chat, chats) {
        Q_EMIT chatUpdated(chat.id());
    }

    Q_EMIT messageAdded(messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onMessagesDeleteChatUsersAnswer - end:" << time.elapsed() << "ms";
}

bool Data::updateContactsLink(const ContactsLink &link) {
    return updateContactsLink(link.user().id(), link.myLink(), link.foreignLink());
}

bool Data::updateContactsLink(qint32 userId, const ContactsMyLink &myLink, const ContactsForeignLink &foreignLink) {

    bool hasPhone = foreignLink.hasPhone();

    if (myLink.classType() == ContactsMyLink::typeContactsMyLinkContact) {
        // userId is a contact
        return updateUserIsContact(userId, true, hasPhone);
    } else if (myLink.classType() == ContactsMyLink::typeContactsMyLinkRequested) {
        if (myLink.contact()) {
            // userId is a contact
            return updateUserIsContact(userId, true, hasPhone);
        } else {
             // userId is not a contact
            return updateUserIsContact(userId, false, hasPhone);
        }
    }
    // any other case is not known relationship with the user, so do nothing but return no error (true)
    return true;
}

void Data::onUpdateShortMessage(qint32 id, qint32 fromId, const QString &message, qint32 pts, qint32 date, qint32 seq) {
    qCDebug(TG_PLUGIN_LOGIC) << "onUpdateShortMessage";

    onUpdateShortChatMessage(id, fromId, 0, message, pts, date, seq);
}

void Data::onUpdateShortChatMessage(qint32 id, qint32 fromId, qint32 chatId, const QString &message, qint32 pts, qint32 date, qint32 seq) {

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdateShortChatMessage";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    // here we check if fromId is a contact, in other case ask for its phone number or any other data
    qint32 dialogId = chatId ? chatId : fromId;

    // search in db dialog related to message, to be evaluated after message insertion
    bool newDialog = !existsDialog(dialogId);

    Message msg(Message::typeMessage);

    msg.setId(id);
    msg.setFromId(fromId);
    msg.setMessage(message);
    msg.setOut(false);
    msg.setUnread(true);
    msg.setDate(date);
    MessageMedia media(MessageMedia::typeMessageMediaEmpty);
    msg.setMedia(media);

    if (chatId) {
        Peer peer(Peer::typePeerChat);
        peer.setChatId(chatId);
        msg.setToId(peer);
    } else {
        // in case receiving chatId == 0, this is a message set to us. Set the toId to ourId
        Peer peer(Peer::typePeerUser);
        peer.setUserId(Settings::getInstance()->ourId());
        msg.setToId(peer);
    }

    if (!insertMessage(msg)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert received short message in database";
        return;
    }

    clientState.setPts(pts);
    clientState.setDate(date);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after receiving a new short message";
        return;
    }

    const ContactItem &sender = getUser(msg.fromId());
    MessageItem messageItem(msg, sender);

    if (newDialog) {
        // Ask telegram servers for info about this just created dialog where it will be included the phone number
        // of the unknown contact
        // this operation is done after insertMessage transaction to ensurance the dialog info coming from server
        // updates db after creating the basic dialog entry related to this message
        TelegramService *service = static_cast<TelegramService *>(mServiceRef);
        service->messagesGetDialogs(0, dialogId, 1);
    }

    mDbManager.finishTransaction();

    Q_EMIT messageAdded(messageItem);
}

void Data::onUpdateShort(const Update &update, qint32 date) {
    qCDebug(TG_PLUGIN_LOGIC) << "onUpdateShort";

    mDbManager.beginTransaction();

    if (!processUpdate(update)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not process update";
        return;
    }

    UpdatesState clientState = getState();
    clientState.setDate(date);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after receiving a short update";
        return;
    }

    mDbManager.finishTransaction();
}

void Data::onUpdatesCombined(const QList<Update> &updates, const QList<User> &users, const QList<Chat> &chats, qint32 date, qint32 seqStart, qint32 seq) {
    qCDebug(TG_PLUGIN_LOGIC) << "onUpdatesCombined";

    UpdatesState clientState = getState();
    // seqStart must be exactly mState.seq() + 1 to process updates
    if (seqStart == clientState.seq() + 1) {
        onUpdates(updates, users, chats, date, seq);
    }
}

void Data::onUpdates(const QList<Update> &updates, const QList<User> &users, const QList<Chat> &chats, qint32 date, qint32 seq) {

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdates";

    UpdatesState clientState = getState();
    if (needUpdateDifference(seq, clientState)) {
        return;
    }

    mDbManager.beginTransaction();

    Q_FOREACH (const Update &update, updates) {
        if (!processUpdate(update, users, chats)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "An error has happened when processing an update";
            return;
        }
    }

    clientState.setDate(date);
    clientState.setSeq(seq);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after creating a chat";
        return;
    }

    mDbManager.finishTransaction();
}

void Data::onUpdatesTooLong() {
    UpdatesState state = getState();
    TelegramService *service = static_cast<TelegramService*>(mServiceRef);
    service->updatesGetDifference(state.pts(), state.date(), state.qts());
}

bool Data::appendOutgoingDecryptedMessage(qint64 requestId, qint32 chatId, const DecryptedMessage &message, const QString &mediaLocalPath, const QString &thumbnailLocalPath) {

    qint32 date = QDateTime::currentDateTime().toTime_t();

    MessageMedia::MessageMediaType mediaType = Tools::toMessageMediaType(message.media().classType());
    qint64 mediaId = 0;
    if (mediaType != MessageMedia::typeMessageMediaEmpty) {
        mediaId = message.randomId();
    }

    mDbManager.beginTransaction();

    if (!insertDecryptedMessage(chatId, message, date, true, mediaId, requestId)) {
        mDbManager.rollbackTransaction();
        return false;
    }

    if (!insertLocalMedia(message.randomId(), mediaType, mediaLocalPath, thumbnailLocalPath)) {
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();

    MessageItem messageItem = getMessage(message.randomId());
    Q_EMIT outgoingMessageAdded(chatId, messageItem);

    // 'local' is the size for sent photos until having a server location and different sizes
    // if there is a mediaLocalPath and media type is photo ro video, emit photo updated signal. Same for video
    if (!mediaLocalPath.isEmpty()) {
        switch (mediaType) {
        case MessageMedia::typeMessageMediaPhoto:
            Q_EMIT messagePhotoUpdated(message.randomId(), mediaLocalPath, false);
            break;
        case MessageMedia::typeMessageMediaVideo:
            Q_EMIT messagePhotoUpdated(message.randomId(), thumbnailLocalPath, true);
            Q_EMIT messageVideoUpdated(message.randomId(), mediaLocalPath);
            break;
        }

        mActiveUploads.insert(requestId, message.randomId());

        Q_EMIT downloadingUpdated(message.randomId(), true);
    }

    return true;
}

QList<qint32> Data::selectDecryptedMessagesToMarkAsRead(qint32 chatId, qint32 maxDate, bool outgoing) {

    QList<qint32> msgIds;

    QSqlQuery query(mDbManager.database());
    query.prepare("SELECT id FROM messages WHERE unread = 1 AND dialogId=:chatId AND out=:outgoing AND date<=:maxDate");
    query.bindValue(":chatId", chatId);
    query.bindValue(":outgoing", outgoing);
    query.bindValue(":maxDate", maxDate);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error selecting messages to mark as read for dialog" << chatId
                                   << query.lastError() << query.lastQuery();
        return msgIds;
    }
    while (query.next()) {
        qint32 messageId = query.value("id").toInt();
        msgIds << messageId;
    }
    return msgIds;
}

qint32 Data::getDecryptedMessagesMaxDate(qint32 chatId, bool outgoing) {

    QSqlQuery query(mDbManager.database());
    query.prepare("SELECT max(date) AS date FROM messages WHERE dialogId=:chatId AND out=:outgoing");
    query.bindValue(":chatId", chatId);
    query.bindValue(":outgoing", outgoing);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error gettings max date for messages in dialog" << chatId
                                   << query.lastError() << query.lastQuery();
        return -1;
    }

    if (!query.next()) {
        qCDebug(TG_PLUGIN_LOGIC) << "Not found any message related to dialog" << chatId;
        return -1;
    }

    return query.value("date").toInt();
}

void Data::deleteSecretChatHistory(qint32 dialogId, bool deleteChat) {
    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "deleteSecretChatHistory";

    mDbManager.beginTransaction();

    if (dialogId == 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not complete dialog history deletion!";
        return;
    }
    deleteMessages((quint32)dialogId);

    if (deleteChat) {
        this->deleteDialog(dialogId);
    }

    mDbManager.finishTransaction();

    Q_EMIT chatHistoryDeleted(dialogId);
    if (deleteChat) {
        Q_EMIT chatDeleted(dialogId);
    }
}

bool Data::markSecretChatAsRead(qint32 chatId) {

    mDbManager.beginTransaction();
    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE messages SET unread=0, readTime=:now WHERE dialogId=:chatId AND out=0 AND readTime = 0");
    query.bindValue(":now", QDateTime::currentDateTime().toTime_t());
    query.bindValue(":chatId", chatId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error marking as read incoming messages for secret chat" << chatId
                                   << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    if (!markChatAsRead(chatId)) {
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();
    return true;
}

bool Data::existsSecretChatMessage(const SecretChatMessage &secretChatMessage) {
    qint32 messageId = secretChatMessage.decryptedMessage().randomId();
    QSqlQuery query (mDbManager.database());
    query.prepare("SELECT id FROM messages WHERE id=:messageId");
    query.bindValue(":messageId", messageId);
    if (!query.exec()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Error checking if exists secret chat message:" << messageId
                                   << query.lastError() << query.lastQuery();
        return false;
    }

    return query.next();
}

bool Data::insertDecryptedMessage(qint32 chatId, const DecryptedMessage &message, qint32 date, bool outgoing, qint64 mediaId, qint64 requestId) {

    DialogItem dialog = getDialog(chatId);

    qint32 fromId = outgoing ? Settings::getInstance()->ourId() : dialog.peerId();
    qint32 toId = outgoing ? dialog.peerId() : Settings::getInstance()->ourId();
    qint32 messageId = message.randomId(); // loses precission but needed to be used in qml since qml cannot use long long numbers

    MessageMedia::MessageMediaType mediaType = Tools::toMessageMediaType(message.media().classType());

    QSqlQuery query(mDbManager.database());
    query.prepare("INSERT INTO messages (id, dialogId, randomId, fromId, toId, toIsChat, out, unread, readTime, ttl, date, text, mediaType, mediaId, sent) "
                  "VALUES (:id, :dialogId, :randomId, :fromId, :toId, :toIsChat, :out, :unread, :readTime, :ttl, :date, :text, :mediaType, :mediaId, :sent)");
    query.bindValue(":id", messageId);
    query.bindValue(":dialogId", chatId);
    query.bindValue(":randomId", requestId);
    query.bindValue(":fromId", fromId);
    query.bindValue(":toId", toId);
    query.bindValue(":toIsChat", false);
    query.bindValue(":out", outgoing);
    query.bindValue(":unread", true); //TODO Roberto: check if needed to provide this value externally
    query.bindValue(":readTime", outgoing ? QDateTime::currentDateTime().toTime_t() : 0);
    query.bindValue(":ttl", message.ttl());
    query.bindValue(":date", date);
    query.bindValue(":text", message.message());
    query.bindValue(":mediaType", mediaType);
    query.bindValue(":mediaId", mediaId);
    query.bindValue(":sent", !outgoing);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Trying to insert an existing message in db:" << message.randomId()
                                   << query.lastError() << query.lastQuery();
        return false;
    }

    return true;
}

bool Data::insertLocalMedia(qint64 mediaId, MessageMedia::MessageMediaType mediaType, const QString &mediaLocalPath, const QString &thumbnailLocalPath) {

    if (mediaType != MessageMedia::typeMessageMediaEmpty && mediaLocalPath.length() > 0) {

        mDbManager.beginTransaction();

        QSqlQuery query(mDbManager.database());
        qint32 fileLocationId;

        switch (mediaType) {
        case MessageMedia::typeMessageMediaPhoto:
            query.prepare("INSERT INTO mediaPhotos (id) VALUES (:id)");
            query.bindValue(":id", mediaId);
            if (!query.exec()) {
                qCWarning(TG_PLUGIN_LOGIC) << "Error when inserting message media photo:" << query.lastError() << query.lastQuery();
                mDbManager.rollbackTransaction();
                return false;
            }

            fileLocationId = insertFileLocation(mediaLocalPath);
            if (fileLocationId <= 0) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not insert local path in a new fileLocations register";
                mDbManager.rollbackTransaction();
                return false;
            }

            // temporally set 'local' size for outgoing photos until getting real sizes
            query.prepare("INSERT INTO photoSizes (photoId, type, fileLocationId) VALUES (:photoId, :type, :fileLocationId)");
            query.bindValue(":photoId", mediaId);
            query.bindValue(":type", MEDIA_LOCAL_FILES_TYPE);
            query.bindValue(":fileLocationId", fileLocationId);
            if (!query.exec()) {
                qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new size for the photo media:" << query.lastError() << query.lastQuery();
                mDbManager.rollbackTransaction();
                return false;
            }

            break;

        case MessageMedia::typeMessageMediaVideo:
            query.prepare("INSERT INTO mediaVideos (id, localPath) VALUES (:id, :localPath)");
            query.bindValue(":id", mediaId);
            query.bindValue(":localPath", mediaLocalPath);
            if (!query.exec()) {
                qCWarning(TG_PLUGIN_LOGIC) << "Error when inserting message media video:" << query.lastError() << query.lastQuery();
                mDbManager.rollbackTransaction();
                return false;
            }

            if (thumbnailLocalPath.length()) {
                fileLocationId = insertFileLocation(thumbnailLocalPath);
                if (fileLocationId <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert local path in a new fileLocations register for the video thumbnail";
                    mDbManager.rollbackTransaction();
                    return false;
                }

                // set 'local' size for thumbnail
                query.prepare("INSERT INTO photoSizes (photoId, type, fileLocationId) VALUES (:photoId, :type, :fileLocationId)");
                query.bindValue(":photoId", mediaId);
                query.bindValue(":type", MEDIA_LOCAL_FILES_TYPE);
                query.bindValue(":fileLocationId", fileLocationId);
                if (!query.exec()) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new photo size for the video thumbnail:" << query.lastError() << query.lastQuery();
                    mDbManager.rollbackTransaction();
                    return false;
                }
            }

            break;
        }

        mDbManager.finishTransaction();
    }

    return true;
}

qint32 Data::insertAttachment(qint32 messageId, const EncryptedFile &attachment, const DecryptedMessageMedia &media) {

    mDbManager.beginTransaction();

    // save the thumbnail included in the message
    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    QString downloadsPath = service->config().downloadsPath();
    qint32 date = QDateTime::currentDateTime().toTime_t();
    qint32 fileLocationId;
    QString thumbnailName;
    QByteArray bytes = media.thumb();
    if (bytes.length() > 0) {
        thumbnailName = downloadsPath + QDir::separator() + PHOTOS_MEDIA_FOLDER + QDir::separator() + "small" + QDir::separator() +
                QString::number(attachment.id()) + "_" + QString::number(date);

        QFileInfo fileInfo(thumbnailName);
        QDir fileDir = fileInfo.absoluteDir();
        QString dirPath = fileDir.absolutePath();
        if (!fileDir.exists()) {
           if (QDir::root().mkpath(dirPath)) {
               qCDebug(TG_PLUGIN_LOGIC) << dirPath << "downloads directory created";
           } else {
               qCWarning(TG_PLUGIN_LOGIC) << "error creating downloads directory: " << dirPath;
               return -1;
           }
        }

        QFile file(thumbnailName);
        file.open(QIODevice::WriteOnly);
        file.write(bytes);
        file.close();

        // Thumbnail location and photo size registers
        fileLocationId = insertFileLocation(thumbnailName);
        if (fileLocationId <= 0) {
            mDbManager.rollbackTransaction();
            return -1;
        }

        PhotoSize thumbPhotoSize;
        thumbPhotoSize.setW(media.thumbW());
        thumbPhotoSize.setH(media.thumbH());
        thumbPhotoSize.setSize(bytes.length());
        thumbPhotoSize.setType("small");
        if (insertPhotoSize(attachment.id(), thumbPhotoSize, fileLocationId) <= 0) {
            mDbManager.rollbackTransaction();
            return -1;
        }
    }

    switch (media.classType()) {
    case DecryptedMessageMedia::typeDecryptedMessageMediaPhoto: {
        QSqlQuery query(mDbManager.database());
        query.prepare("INSERT INTO mediaPhotos (id, accessHash, date, key, iv) VALUES (:id, :accessHash, :date, :key, :iv)");
        query.bindValue(":id", attachment.id());
        query.bindValue(":accessHash", attachment.accessHash());
        query.bindValue(":date", date);
        query.bindValue(":key", media.key().toBase64());
        query.bindValue(":iv", media.iv().toBase64());
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not insert photo attachment in database for media with id:"
                                        << attachment.id() << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }

        // Big photo location and photo size
        FileLocation serverLocation(FileLocation::typeFileLocation);
        serverLocation.setDcId(attachment.dcId());
        fileLocationId = insertFileLocation(serverLocation);
        if (fileLocationId <= 0) {
            mDbManager.rollbackTransaction();
            return -1;
        }

        PhotoSize bigPhotoSize;
        bigPhotoSize.setW(media.w());
        bigPhotoSize.setH(media.h());
        bigPhotoSize.setSize(attachment.size());
        bigPhotoSize.setType("big");
        if (insertPhotoSize(attachment.id(), bigPhotoSize, fileLocationId) <= 0) {
            mDbManager.rollbackTransaction();
            return -1;
        }

        if (thumbnailName.length() > 0) {
            Q_EMIT messagePhotoUpdated(messageId, thumbnailName, true);
        }

        break;
    }
    case DecryptedMessageMedia::typeDecryptedMessageMediaVideo_layer8:
    case DecryptedMessageMedia::typeDecryptedMessageMediaVideo: {
        QSqlQuery query(mDbManager.database());
        query.prepare("INSERT INTO mediaVideos (id, accessHash, date, mimeType, size, dcId, width, height, key, iv) "
                      "VALUES (:id, :accessHash, :date, :mimeType, :size, :dcId, :width, :height, :key, :iv)");
        query.bindValue(":id", attachment.id());
        query.bindValue(":accessHash", attachment.accessHash());
        query.bindValue(":date", date);
        query.bindValue(":mimeType", media.mimeType());
        query.bindValue(":size", attachment.size());
        query.bindValue(":dcId", attachment.dcId());
        query.bindValue(":width", media.w());
        query.bindValue(":height", media.h());
        query.bindValue(":key", media.key().toBase64());
        query.bindValue(":iv", media.iv().toBase64());
        if (!query.exec()) {
            qCWarning(TG_PLUGIN_LOGIC) << "Could not insert video attachment in database for media with id:"
                                       << attachment.id() << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }

        //assign a fake value for fileLocationId > 0 to indicate the operation has been valid
        fileLocationId = 1;

        if (thumbnailName.length() > 0) {
            Q_EMIT messagePhotoUpdated(messageId, thumbnailName, true);
        }

        break;
    }
    }

    mDbManager.finishTransaction();

    // return the fileLocationId for the big photo
    return fileLocationId;
}

void Data::downloadPhotoAttachment(qint32 messageId, qint64 mediaId, qint32 date, bool thumbnail, const QString &type,
                             qint32 size, qint32 dcId, qint64 accessHash, const QByteArray &key, const QByteArray &iv,
                             qint32 fileLocationId) {

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    QString downloadsPath = service->config().downloadsPath();
    IncomingFile *incomingFile = new IncomingFile(downloadsPath);
    incomingFile->setId(messageId);
    incomingFile->setMediaId(mediaId);
    incomingFile->setMediaType(MessageMedia::typeMessageMediaPhoto);
    incomingFile->setIncomingFileType(IncomingFile::EncryptedMessagePhoto);
    incomingFile->setSizeType(type);
    incomingFile->setIsThumbnail(thumbnail);
    incomingFile->setFileLocationId(fileLocationId);

    QString fileName = QString::number(messageId) + "_" + QString::number(date);
    incomingFile->setFilename(fileName);
    incomingFile->setExpectedSize(size);

    InputFileLocation serverLocation(InputFileLocation::typeInputEncryptedFileLocation);
    serverLocation.setId(mediaId);
    serverLocation.setAccessHash(accessHash);
    qint64 fileId = service->uploadGetFile(serverLocation, size, dcId, key, iv);

    mFiles.insert(fileId, incomingFile);
    mActiveDownloads.insert(messageId, fileId);

    Q_EMIT downloadingUpdated(messageId, true);
}

bool Data::insertSecretChatMessage(const SecretChatMessage &secretChatMessage) {
    qint32 chatId = secretChatMessage.chatId();
    qint32 date = secretChatMessage.date();
    const DecryptedMessage &decryptedMessage = secretChatMessage.decryptedMessage();
    const EncryptedFile &attachment = secretChatMessage.attachment();

    //TODO Roberto: not process service messages by the moment.
    if (decryptedMessage.classType() == DecryptedMessage::typeDecryptedMessageService ||
            decryptedMessage.classType() == DecryptedMessage::typeDecryptedMessageService_level8) {
        return false;
    }

    qint32 messageId = decryptedMessage.randomId();
    //TODO Roberto: sanity check to avoid managing void incoming messages
    if (messageId == 0) {
        return false;
    }

    mDbManager.beginTransaction();

    if (!insertDecryptedMessage(chatId, decryptedMessage, date, false, attachment.id())) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert received new message with id" << messageId <<
                                       "in secret chat" << chatId;
        mDbManager.rollbackTransaction();
        return false;
    }

    if (attachment.classType() == EncryptedFile::typeEncryptedFile) {
        DecryptedMessageMedia media = decryptedMessage.media();
        switch (media.classType()) {
        case DecryptedMessageMedia::typeDecryptedMessageMediaPhoto: {

            if (existsPhoto(attachment.id())) {
                qCDebug(TG_PLUGIN_LOGIC) << "Photo already exists. No need to be downloaded again";
            } else {
                qint32 fileLocationId = insertAttachment(messageId, attachment, media);
                if (fileLocationId <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert photo attachment for message" << messageId;
                    mDbManager.rollbackTransaction();
                    return false;
                }
                downloadPhoto(messageId);
            }
            break;
        }
        case DecryptedMessageMedia::typeDecryptedMessageMediaVideo_layer8:
        case DecryptedMessageMedia::typeDecryptedMessageMediaVideo: {

            if (existsVideo(attachment.id())) {
                qCDebug(TG_PLUGIN_LOGIC) << "Video already exists. No need to be downloaded again";
            } else {
                qint32 result = insertAttachment(messageId, attachment, media);
                if (result <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert video attachment for message" << messageId;
                    mDbManager.rollbackTransaction();
                    return false;
                }
            }
            break;
        }
        }
    }

    mDbManager.finishTransaction();
    return true;
}

void Data::setSecretChatMessageTTL(qint32 chatId, qint32 ttlSeconds) {
    qCDebug(TG_PLUGIN_LOGIC) << "setSecretChatMessageTTL" << chatId << ttlSeconds;
    QSqlQuery query;
    query.prepare("UPDATE secretChats SET ttl = :ttl WHERE id = :chatId");
    query.bindValue(":ttl", ttlSeconds);
    query.bindValue(":chatId", chatId);
    if (!query.exec() || query.numRowsAffected() == 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update secret chat TTL" << query.lastError();
    } else {
        qCDebug(TG_PLUGIN_LOGIC) << "Updated secret chat TTL";
        Q_EMIT secretChatUpdated(chatId);
    }
}

void Data::processSecretChatServiceMessage(const SecretChatMessage &secretChatMessage) {
    DecryptedMessage message = secretChatMessage.decryptedMessage();
    DecryptedMessageAction action = message.action();

    if (action.classType() == DecryptedMessageAction::typeDecryptedMessageActionSetMessageTTL) {
        setSecretChatMessageTTL(secretChatMessage.chatId(), action.ttlSeconds());
    } // TODO: other DecryptedMessageAction types here...
}

void Data::onUpdateSecretChatMessage(const SecretChatMessage &secretChatMessage, qint32 qts) {

    QTime time;
    time.start();

    qCDebug(TG_PLUGIN_LOGIC) << "onUpdateSecretChatMessage";

    if (secretChatMessage.isServiceMessage()) {
        processSecretChatServiceMessage(secretChatMessage);
        return;
    } else if (!existsDialog(secretChatMessage.chatId())) {
        return;
    }

    mDbManager.beginTransaction();

    UpdatesState clientState = getState();
    clientState.setQts(qts);
    if (!insertOrUpdateState(clientState)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after receiving a decrypted message in a secret chat";
        return;
    }

    if (!insertSecretChatMessage(secretChatMessage)) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert secret chat message";
        mDbManager.rollbackTransaction();
        return;
    }

    qint32 chatId = secretChatMessage.chatId();
    if (!increaseUnreadCount(chatId, 1)) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Could not increase unread count for dialog" << chatId;
        return;
    }

    qint32 messageId = secretChatMessage.decryptedMessage().randomId();
    MessageItem messageItem = getMessage(messageId);

    mDbManager.finishTransaction();

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    service->messagesReceivedQueue(qts);

    Q_EMIT messageAdded(messageItem);

    qCDebug(TG_PLUGIN_PROFILING) << "onUpdateSecretChatMessage - end:" << time.elapsed() << "ms";
}

void Data::updateChatParticipantsFromUpdate(const Update &update) {
    qCWarning(TG_PLUGIN_LOGIC) << "typeUpdateChatParticipants";
    ChatParticipants participants = update.participants();
    updateChatParticipants(participants);
    mDbManager.finishTransaction(); // Finish before emitting chatUpdated!
    Q_EMIT chatUpdated(update.chatId());
}

void Data::processUpdateUserName(const Update &update) {
    qint32 userId = update.userId();
    QString firstName = update.firstName();
    QString lastName = update.lastName();
    if (updateUserName(userId, firstName, lastName)) {
        Q_EMIT contactNameUpdated(userId, firstName, lastName);

        QString fullName = firstName + " " + lastName;
        Q_EMIT dialogTitleUpdated(userId, fullName);
        notifySecretChatsTitleUpdated(userId, fullName);
    }
}

void Data::processUpdateUserPhoto(const Update &update) {
    qint32 userId = update.userId();
    const FileLocation &photoToDownload = update.photo().photoSmall();
    bool isUserPhoto = true;
    bool isThumbnail = true;
    getProfilePhotoFileFromServer(photoToDownload, userId, isUserPhoto, isThumbnail);
}

bool Data::processUpdate(const Update &update, const QList<User> &users, const QList<Chat> &chats) {
    // FIXME Users and Chats code blocks repeated in onUpdatesGetDifferenceAnswer.
    mDbManager.beginTransaction();
    Q_FOREACH (const User &user, users) {
        if (!insertOrUpdateUser(user)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating user in database while processing an update";
            return false;
        }
        // request user photo if it has changed
        getProfilePhotoFileFromServer(user.photo().photoSmall(), user.id(), true, true);
    }
    Q_FOREACH (const Chat &chat, chats) {
        if (!insertOrUpdateChat(chat)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating chat in database while processing an update";
            return false;
        }
    }

    try {
    //TODO COMPLETE UPDATES
    switch (update.classType()) {
    case Update::typeUpdateMessageID: {
        qint64 randomId = update.randomId();
        qint32 messageId = update.id();
        qint32 oldMessageId = getMessageId(randomId);

        if (oldMessageId && updateMessageId(randomId, messageId)) {

            qCDebug(TG_PLUGIN_LOGIC) << "updated message" << oldMessageId << "with randomId"
                                     << randomId << "to id" << messageId;
            MessageItem message = getMessage(messageId);
            Q_EMIT messageUpdated(oldMessageId, message);
        } else {
            qCDebug(TG_PLUGIN_LOGIC) << "Could not update message related to randomId" << randomId;
        }

        break;
    }
    case Update::typeUpdateUserTyping: {
        if (update.action().classType() == SendMessageAction::typeSendMessageTypingAction) {
            Q_EMIT userTyping(update.userId());
        }
        break;
    }
    case Update::typeUpdateChatUserTyping: {
        if (update.action().classType() == SendMessageAction::typeSendMessageTypingAction) {
            qint32 dialogId = update.chatId();
            qint32 userId = update.userId();
            const ContactItem &user = getUser(userId);
            QString fullName = user.firstName();
            if (!user.lastName().isEmpty()) {
                fullName += " " + user.lastName();
            }
            Q_EMIT userTyping(dialogId, fullName);
        }
        break;
    }
    case Update::typeUpdateEncryptedChatTyping: {
        Q_EMIT userTyping(update.chatId());
        break;
    }
    case Update::typeUpdateUserName:
        processUpdateUserName(update);
        break;
    case Update::typeUpdateUserPhoto:
        processUpdateUserPhoto(update);
        break;
    case Update::typeUpdateNewMessage: {

        Message message = update.message();
        qint32 dialogId = Tools::getBelongingDialogId(message);

        // search in db dialog related to just received message
        bool newDialog = !existsDialog(dialogId);

        if (!insertMessage(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Impossible insert new received message";
            return false;
        }

        if (!executeAction(message)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not execute message related action";
            return false;
        }

        UpdatesState clientState = getState();
        clientState.setPts(update.pts());
        if (!insertOrUpdateState(clientState)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after receiving a new message";
            return false;
        }

        MessageItem messageItem = getMessage(message.id());
        Q_EMIT messageAdded(messageItem);

        if (newDialog) {
            // Ask telegram servers for info about this just created dialog where it will be included the phone number
            // of the unknown contact
            // this operation is done after insertMessage transaction to ensurance the dialog info coming from server
            // updates db after creating the basic dialog entry related to this message
            TelegramService *service = static_cast<TelegramService *>(mServiceRef);
            service->messagesGetDialogs(0, dialogId, 1);
        }

        break;
    }
    case Update::typeUpdateReadMessages: {
        if (markMessagesAsRead(update.messages())) {
            Q_EMIT messagesMarkedAsRead(update.messages());
        }

        UpdatesState clientState = getState();
        clientState.setPts(update.pts());
        if (!insertOrUpdateState(clientState)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after reading history messages";
            return false;
        }

        break;
    }
    case Update::typeUpdateDeleteMessages: {
        if (deleteMessages(update.messages())) {
            Q_EMIT messagesDeleted(update.messages());
        }

        UpdatesState clientState = getState();
        clientState.setPts(update.pts());
        if (!insertOrUpdateState(clientState)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update client state after deleting messages";
            return false;
        }

        break;
    }
    case Update::typeUpdateContactLink: {
        qint32 userId = update.userId();
        ContactsMyLink myLink = update.myLink();
        ContactsForeignLink foreignLink = update.foreignLink();
        if (!updateContactsLink(userId, myLink, foreignLink)) {
            mDbManager.rollbackTransaction();
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update contacts links";
            return false;
        }
        break;
    }

    // updateChatParticipants#7761198 participants:ChatParticipants = Update;
    case Update::typeUpdateChatParticipants: {
        qCDebug(TG_PLUGIN_LOGIC) << "typeUpdateChatParticipants";
        updateChatParticipantsFromUpdate(update);
        break;
    }

    // updateChatParticipantAdd#3a0eeb22 chat_id:int user_id:int inviter_id:int version:int = Update;
    case Update::typeUpdateChatParticipantAdd: {
        // NOTE: This is not called when participant is added.
        qCDebug(TG_PLUGIN_LOGIC) << "typeUpdateChatParticipantAdd";
        break;
    }

    // updateChatParticipantDelete#6e5f8c22 chat_id:int user_id:int version:int = Update;
    case Update::typeUpdateChatParticipantDelete: {
        // NOTE: This is not called when participant is deleted.
        qCDebug(TG_PLUGIN_LOGIC) << "typeUpdateChatParticipantDelete";
        break;
    }

    case Update::typeUpdateNotifySettings: {
        NotifyPeer notifyPeer = update.peer();
        PeerNotifySettings settings = update.notifySettings();
        qint32 muteUntil = settings.muteUntil();

        bool muted = muteUntil > QDateTime::currentDateTime().toTime_t() ? true : false;

        if (notifyPeer.classType() == NotifyPeer::typeNotifyPeer) {
            Peer peer = notifyPeer.peer();
            qint32 id = peer.classType() == Peer::typePeerUser ? peer.userId() : peer.chatId();
            if (muteDialog(id, muted)) {
                Q_EMIT dialogMuted(id, muted);
            }
        } else if (muteDialogs(notifyPeer.classType(), muted)) {
            Q_EMIT dialogsMuted(notifyPeer.classType(), muted);
        }
        break;
    }

    case Update::typeUpdateEncryptedMessagesRead: {
        qint32 chatId = update.chatId();
        qint32 maxDate = update.maxDate();
        bool selectOutgoingMessages = true;
        QList<qint32> msgIds = selectDecryptedMessagesToMarkAsRead(chatId, maxDate, selectOutgoingMessages);
        if (markMessagesAsRead(msgIds)) {
            Q_EMIT messagesMarkedAsRead(msgIds);
        }
        break;
    }
    }

        // FIXME for those other updates that don't call it yet:
        mDbManager.finishTransaction();
    } catch (const char* e) {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "failed update" << e;
    }
    return true;
}

bool Data::needUpdateDifference(qint32 seq, UpdatesState clientState) {
    if (seq != clientState.seq() + 1) {
        qCDebug(TG_PLUGIN_LOGIC) << "needUpdateDifference: " << clientState.seq();
        TelegramService *service = static_cast<TelegramService *>(mServiceRef);
        service->updatesGetDifference(clientState.pts(), clientState.date(), clientState.qts());
        return true;
    } else {
        return false;
    }
}

void Data::notifySecretChatsThumbnailUpdated(qint32 userId, const QString &filePath) {
    QList<qint32> userSecretChats = getUserSecretChats(userId);
    Q_FOREACH (qint32 secretChatId, userSecretChats) {
        Q_EMIT dialogThumbnailUpdated(secretChatId, filePath);
    }
}

void Data::notifySecretChatsTitleUpdated(qint32 userId, const QString &title) {
    QList<qint32> userSecretChats = getUserSecretChats(userId);
    Q_FOREACH (qint32 secretChatId, userSecretChats) {
        Q_EMIT dialogTitleUpdated(secretChatId, title);
    }
}

bool Data::downloadUserBigPhoto(qint32 userId) {
    return downloadProfilePhoto(userId, true, false);
}

bool Data::downloadChatBigPhoto(qint32 chatId) {
    return downloadProfilePhoto(chatId, false, false);
}

bool Data::downloadProfilePhoto(qint32 id, bool isUser, bool isThumbnail) {

    QPair<QString, FileLocation> fullFileLocation = getFullFileLocation(id, isUser, isThumbnail);

    QString localPath = fullFileLocation.first;
    if (!localPath.isEmpty()) {
        qCWarning(TG_PLUGIN_LOGIC) << "profile photo for" << id << "has been previously downloaded. Skipping";
        return false;
    }

    FileLocation serverLocation = fullFileLocation.second;
    return downloadPhotoFileFromServer(serverLocation, id, isUser, isThumbnail);
}

QPair<QString, FileLocation> Data::getFullFileLocation(qint32 id, bool isUser, bool isThumbnail) {

    QPair<QString, FileLocation> locationPair;

    QString queryString("SELECT localPath, dcId, localId, secret, volumeId FROM fileLocations,%1 "
                        "WHERE %1.%2=fileLocations.rowid AND %1.id=:id");

    QString table = isUser ? "users" : "chats";
    QString field = isThumbnail ? "thumbnail" : "bigPhoto";

    queryString = queryString.arg(table, field);

    QSqlQuery query(mDbManager.database());
    query.prepare(queryString);
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get" << field << "for" << table << " register with id" << id
                                    << query.lastError() << query.lastQuery();
        return locationPair;
    }

    if (!query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not find" << field << "for a register in table" << table << "with id" << id;
        return locationPair;
    }

    QString localPath = query.value("localPath").toString();

    qint32 dcId = query.value("dcId").toInt();
    qint32 localId = query.value("localId").toInt();
    qint64 secret = query.value("secret").toLongLong();
    qint64 volumeId = query.value("volumeId").toLongLong();

    FileLocation serverLocation(FileLocation::typeFileLocation);
    serverLocation.setDcId(dcId);
    serverLocation.setLocalId(localId);
    serverLocation.setSecret(secret);
    serverLocation.setVolumeId(volumeId);

    locationPair = QPair<QString, FileLocation>(localPath, serverLocation);

    return locationPair;
}

bool Data::getProfilePhotoFileFromServer(const FileLocation &serverLocation, qint32 id, bool isUser, bool isThumbnail) {

    if (serverLocation.classType() != FileLocation::typeFileLocationUnavailable) {
        // Search in database if same location parameter are stored for thumbnail that
        // received ones. In that case don't download cos it's the same file
        QPair<QString, FileLocation> dbFullFileLocation = getFullFileLocation(id, isUser, isThumbnail);
        QString localPath = dbFullFileLocation.first;
        FileLocation dbLocation = dbFullFileLocation.second;

        if (!localPath.isEmpty() && dbLocation == serverLocation) {
            return false;
        }

        return downloadPhotoFileFromServer(serverLocation, id, isUser, isThumbnail);
    }
    return false;
}

bool Data::downloadPhotoFileFromServer(const FileLocation &serverLocation, qint32 id, bool isUser, bool isThumbnail) {

    if (serverLocation.classType() != FileLocation::typeFileLocationUnavailable) {
        // create object with the data of the file to be downloaded
        TelegramService *service = static_cast<TelegramService *>(mServiceRef);
        QString downloadsPath = service->config().downloadsPath();
        IncomingFile *incomingFile = new IncomingFile(downloadsPath);
        incomingFile->setId(id);

        IncomingFile::IncomingFileType incomingFileType;
        if (isUser) {
            if (isThumbnail) {
                incomingFileType = IncomingFile::UserThumbnail;
            } else {
                incomingFileType = IncomingFile::UserPhoto;
            }
        } else {
            if (isThumbnail) {
                incomingFileType = IncomingFile::ChatThumbnail;
            } else {
                incomingFileType = IncomingFile::ChatPhoto;
            }
        }

        incomingFile->setIncomingFileType(incomingFileType);
        incomingFile->setMediaType(MessageMedia::typeMessageMediaPhoto);
        incomingFile->setIsThumbnail(isThumbnail);
        // default extension for photos is jpeg. If incoming file is enough big, response will not inform us
        // about the real type of the incoming photo, so use this extension
        incomingFile->setExtension(".jpeg");
        incomingFile->setServerLocation(serverLocation);

        InputFileLocation inputFileLocation(InputFileLocation::typeInputFileLocation);
        inputFileLocation.setLocalId(serverLocation.localId());
        inputFileLocation.setSecret(serverLocation.secret());
        inputFileLocation.setVolumeId(serverLocation.volumeId());
        qint64 fileId = service->uploadGetFile(inputFileLocation, 0, serverLocation.dcId());

        mFiles.insert(fileId, incomingFile);
        mActiveDownloads.insert(id, fileId);

        Q_EMIT downloadingUpdated(id, true);

        return true;
    }
    return false;
}

bool Data::downloadPhoto(qint32 messageId) {

    bool downloading = false;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT isSecret FROM dialogs WHERE dialogs.id IN (SELECT dialogId FROM messages WHERE messages.id=:messageId)");
    query.bindValue(":messageId", messageId);
    if (!query.exec() || !query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not check if the message" << messageId << "belongs to a secret chat";
        return false;
    }

    bool isSecret = query.value("isSecret").toBool();

    query.prepare("SELECT mediaId, dcId, localId, secret, volumeId, size, type, fileLocationId, "
                        "mediaPhotos.date AS date, accessHash, key, iv, localPath "
                  "FROM messages, mediaPhotos, photoSizes, fileLocations "
                  "WHERE mediaPhotos.id = photoSizes.photoId "
                  "     AND photoSizes.fileLocationId = fileLocations.rowid "
                  "     AND messages.mediaId = mediaPhotos.id "
                  "     AND messages.id = :messageId "
                  "ORDER BY photoSizes.size DESC "
                  "LIMIT 2");
    query.bindValue(":messageId", messageId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get photo metadata for downloading photo file for message" << messageId
                                       << query.lastError() << query.lastQuery();
        return downloading;
    }

    // we have selected two sizes, starting with the bigger.
    // First one will be downloaded as big photo, second one as thumbnail
    bool thumbnail = false;
    while (query.next()) {

        qint64 mediaId = query.value("mediaId").toLongLong();
        qint32 dcId = query.value("dcId").toInt();
        qint32 localId = query.value("localId").toInt();
        qint64 secret = query.value("secret").toLongLong();
        qint64 volumeId = query.value("volumeId").toLongLong();
        qint32 size = query.value("size").toInt();
        QString type = query.value("type").toString();
        qint32 fileLocationId = query.value("fileLocationId").toInt();
        qint32 date = query.value("date").toInt();
        qint64 accessHash = query.value("accessHash").toLongLong();
        QByteArray key = QByteArray::fromBase64(query.value("key").toByteArray());
        QByteArray iv = QByteArray::fromBase64(query.value("iv").toByteArray());
        QString localPath = query.value("localPath").toString();

        if (localPath.isEmpty()) {
            if (isSecret) {
                downloadPhotoAttachment(messageId, mediaId, date, thumbnail, type, size, dcId, accessHash, key, iv, fileLocationId);
            } else {
                downloadPhotoFile(messageId, mediaId, date, IncomingFile::MessagePhoto, thumbnail, type, size, dcId, localId, secret,
                                  volumeId, fileLocationId);
            }
        }

        thumbnail = true;
        downloading = true;
    }

    return downloading;
}

bool Data::downloadVideo(qint32 messageId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT isSecret FROM dialogs WHERE dialogs.id IN (SELECT dialogId FROM messages WHERE messages.id=:messageId)");
    query.bindValue(":messageId", messageId);
    if (!query.exec() || !query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not check if the message" << messageId << "belongs to a secret chat";
        return false;
    }

    bool isSecret = query.value("isSecret").toBool();

    query.prepare("SELECT id, accessHash, mimeType, size, dcId, localPath, key, iv FROM mediaVideos WHERE id IN "
                  "(SELECT mediaId FROM messages WHERE messages.id=:messageId)");
    query.bindValue(":messageId", messageId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get video info from database for message" << messageId
                                       << query.lastError() << query.lastQuery();
        return false;
    }

    if (!query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get video info from database for message" << messageId;
        return false;
    }

    qint64 mediaId = query.value("id").toLongLong();
    qint64 accessHash = query.value("accessHash").toLongLong();
    QString mimeType = query.value("mimeType").toString();
    qint32 size = query.value("size").toInt();
    qint32 dcId = query.value("dcId").toInt();
    QString localPath = query.value("localPath").toString();
    QByteArray key = QByteArray::fromBase64(query.value("key").toByteArray());
    QByteArray iv = QByteArray::fromBase64(query.value("iv").toByteArray());

    if (!localPath.isEmpty()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Video has already been downloaded previously";
        return false;
    }

    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    QString downloadsPath = service->config().downloadsPath();
    IncomingFile *incomingFile = new IncomingFile(downloadsPath);
    incomingFile->setId(messageId);
    incomingFile->setMediaId(mediaId);
    incomingFile->setMediaType(MessageMedia::typeMessageMediaVideo);

    if (isSecret) {
        incomingFile->setIncomingFileType(IncomingFile::EncryptedMessageVideo);
    } else {
        incomingFile->setIncomingFileType(IncomingFile::MessageVideo);
    }

    QString fileExtension = Tools::getExtensionFromMimeType(mimeType);
    //set default extension to .mp4. This is needed for incoming iOS videos that doesn't have a mime type
    if (fileExtension.isEmpty()) {
        fileExtension = ".mp4";
    }
    incomingFile->setExtension(fileExtension);
    incomingFile->setExpectedSize(size);

    InputFileLocation serverLocation;
    if (isSecret) {
        serverLocation.setClassType(InputFileLocation::typeInputEncryptedFileLocation);
    } else {
        serverLocation.setClassType(InputFileLocation::typeInputVideoFileLocation);
    }
    serverLocation.setId(mediaId);
    serverLocation.setAccessHash(accessHash);

    // key and iv won't be taken into account if the message belongs to a regular chat
    qint64 fileId = service->uploadGetFile(serverLocation, size, dcId, key, iv);

    mFiles.insert(fileId, incomingFile);
    mActiveDownloads.insert(messageId, fileId);

    Q_EMIT downloadingUpdated(messageId, true);

    return true;
}

// Next, you'll find database direct operations methods. These could be, later, included in a base class,
// to split incoming messages management from database management operations.
QList<DialogItem> Data::getDialogs() {

    QList<DialogItem> result;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id, isChat, unreadCount FROM dialogs");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "error retrieving dialogs list from database";
        return result;
    }

    while (query.next()) {
        DialogItem dialog;
        dialog.setId(query.value("id").toInt());
        dialog.setIsChat(query.value("isChat").toBool());
        dialog.setUnreadCount(query.value("unreadCount").toInt());

        if (dialog.isChat()) {
            DialogItem chat = getChat(dialog.id());
            dialog.setName(chat.name());
            dialog.setMembers(chat.members());
            dialog.setThumbnail(chat.thumbnail());
        } else {
            ContactItem user = getUser(dialog.id());
            dialog.setName(user.firstName() + " " + user.lastName());
            dialog.setThumbnail(user.thumbnail());
            dialog.setUserOnline(user.status()->classType() == UserStatus::typeUserStatusOnline);
            dialog.setLastSeenOnline(user.status()->wasOnline());
        }

        MessageItem message = topMessage(dialog.id());
        dialog.setTopMessage(message);
        dialog.setDate(message.date());
        result.append(dialog);
    }
    return result;
}

bool Data::existsDialog(qint32 dialogId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM dialogs WHERE id=:id");
    query.bindValue(":id", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not check if dialog exists" << query.lastError() << query.lastQuery();
        return false;
    }

    return query.next();
}

DialogItem Data::getDialog(qint32 dialogId) {

    DialogItem dialog;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT isChat, unreadCount, date, isSecret FROM dialogs WHERE id=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when getting dialog from db with id" << dialogId
                                    << query.lastError() << query.lastQuery();
        return dialog;
    }

    if (query.next()) {
        dialog.setId(dialogId);
        dialog.setIsChat(query.value("isChat").toBool());
        dialog.setIsSecret(query.value("isSecret").toBool());
        dialog.setUnreadCount(query.value("unreadCount").toInt());
        dialog.setDate(query.value("date").toInt());
        dialog.setPeerId(dialogId);

        if (dialog.isChat()) {
            DialogItem chat = getChat(dialogId);
            dialog.setName(chat.name());
            dialog.setMembers(chat.members());
            dialog.setThumbnail(chat.thumbnail());
        } else {

            if (dialog.isSecret()) {

                DialogItem secretChat = getSecretChat(dialogId);
                if (dialog.date() == 0) {
                    dialog.setDate(secretChat.date());
                }
                dialog.setPeerId(secretChat.peerId());
                dialog.setState(secretChat.state());
            }

            ContactItem user = getUser(dialog.peerId());
            dialog.setName(user.firstName() + " " + user.lastName());
            dialog.setThumbnail(user.thumbnail());
            dialog.setUserOnline(user.status()->classType() == UserStatus::typeUserStatusOnline);
            dialog.setLastSeenOnline(user.status()->wasOnline());
            dialog.setPeerType(user.classType());
        }

        //TODO Roberto: see how to get secret chat messages
        MessageItem message = topMessage(dialog.id());
        if (!message.isEmpty()) {
            dialog.setTopMessage(message);
        }
    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "queried dialog with id" << dialogId << "has not been found";
    }

    return dialog;
}

qint32 Data::getDialogIdFromMessageId(qint32 msgId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT dialogId FROM messages WHERE id=:id");
    query.bindValue(":id", msgId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return 0;
    }

    if (query.next()) {
        return query.value("dialogId").toInt();
    } else {
        return 0;
    }
}

bool Data::dialogIsEmpty(qint32 dialogId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT count(id) as count FROM messages WHERE dialogId=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    if (query.next()) {
        return query.value("count").toInt() == 0;
    } else {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when checking if dialog hasn't messages." << query.lastQuery();
        return false;
    }
}

bool Data::decreaseUnreadCount(qint32 dialogId, qint32 count) {
    return increaseUnreadCount(dialogId, -count);
}

bool Data::increaseUnreadCount(qint32 dialogId, qint32 count) {

    QSqlQuery query(mDbManager.database());

    // for avoiding weird cases that suppose having a dialog with a number of unread messages < 0, check that
    // count is not greater than current unreadCount in case unreadCount must be decreased
    if (count < 0) {
        query.prepare("SELECT unreadCount FROM dialogs WHERE id=:id");
        query.bindValue(":id", dialogId);
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error checking dialog with id" << dialogId << "unread count:" << query.lastError() << query.lastQuery();
            return false;
        }

        if (query.next()) {
            qint32 unreadCount = query.value("unreadCount").toInt();
            if (qAbs(count) > unreadCount) {
                count = -unreadCount;
            }
        } else {
            qCWarning(TG_PLUGIN_LOGIC) << "Dialog with id" << dialogId << "to be updated its unread count has not been found";
        }
    }

    query.prepare("UPDATE dialogs SET unreadCount=unreadCount + :count WHERE id=:id");
    query.bindValue(":id", dialogId);
    query.bindValue(":count", count);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update unreadCount for dialog" << dialogId << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

qint32 Data::getUnreadCount(qint32 dialogId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT unreadCount FROM dialogs WHERE id=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return -1;
    }
    if (query.next()) {
        return query.value("unreadCount").toInt();
    } else {
        return -1;
    }
}

MessageItem Data::topMessage(qint32 dialogId) {

    MessageItem message;

    QString sql = QString("SELECT messages.id, fromId, firstName AS fromFirstName, lastName AS fromLastName, "
                          "phone AS fromPhone, toId, toIsChat, out, unread, sent, date, text, mediaType, "
                          "(SELECT type FROM messageActions WHERE messageId=messages.id) AS actionType, "
                          "(SELECT title FROM messageActions WHERE messageId=messages.id) AS actionTitle, "
                          "(SELECT coalesce(users.firstName,'') || ' ' || coalesce(users.lastName,'') FROM users WHERE users.id IN "
                                "(SELECT userId FROM messageActions WHERE messageId=messages.id)) AS actionUser "
                          "FROM messages, users "
                          "WHERE dialogId=:dialogId AND fromId=users.id "
                          "ORDER BY datetime(date) DESC LIMIT 1");

    QSqlQuery query(mDbManager.database());
    query.prepare(sql);
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried top message in dialog with id" << dialogId << "has not been found:" << query.lastError() << query.lastQuery();;
        return message;
    }

    if (query.next()) {
        message.setId(query.value("id").toInt());
        message.setDialogId(dialogId);
        message.setFromId(query.value("fromId").toInt());
        message.setFromFirstName(query.value("fromFirstName").toString());
        message.setFromLastName(query.value("fromLastName").toString());
        message.setFromPhone(query.value("fromPhone").toString());
        Peer p(Peer::typePeerUser);
        if (query.value("toIsChat").toBool()) {
            p.setChatId(query.value("toId").toInt());
        } else {
            p.setUserId(query.value("toId").toInt());
        }
        message.setToId(p);
        message.setOut(query.value("out").toBool());
        message.setUnread(query.value("unread").toBool());
        message.setSent(query.value("sent").toBool());
        message.setDate(query.value("date").toInt());
        message.setMessage(query.value("text").toString());
        MessageMedia media((MessageMedia::MessageMediaType)query.value("mediaType").toInt());
        message.setMedia(media);

        quint32 actionType = query.value("actionType").toInt();
        if (actionType) {
            MessageAction messageAction((MessageAction::MessageActionType)actionType);
            messageAction.setTitle(query.value("actionTitle").toString());
            message.setAction(messageAction);
            message.setActionUser(query.value("actionUser").toString());
        }

    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "Not found top message for dialog" << dialogId << ";returning empty object";
    }

    return message;
}

bool Data::insertOrUpdateState(const UpdatesState &state) {

    if (!updateState(state)) {
        return insertState(state);
    }
    return true;
}

bool Data::insertState(const UpdatesState &state) {

    qint32 pts = state.pts();
    qint32 qts = state.qts();
    qint32 date = state.date();
    qint32 seq = state.seq();
    qint32 unreadCount = state.unreadCount();

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO state (name, value) VALUES (:name, :value)");

    mDbManager.beginTransaction();

    query.bindValue(":name", "pts");
    query.bindValue(":value", pts);
    if (!query.exec() || query.numRowsAffected() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting pts value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    query.bindValue(":name", "qts");
    query.bindValue(":value", qts);
    if (!query.exec() || query.numRowsAffected() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting qts value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    query.bindValue(":name", "date");
    query.bindValue(":value", date);
    if (!query.exec() || query.numRowsAffected() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting date value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    query.bindValue(":name", "seq");
    query.bindValue(":value", seq);
    if (!query.exec() || query.numRowsAffected() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting seq value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    query.bindValue(":name", "unreadCount");
    query.bindValue(":value", unreadCount);
    if (!query.exec() || query.numRowsAffected() != 1) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting unreadCount value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();
    return true;
}

bool Data::updateState(const UpdatesState &state) {

    qint32 rowsAffected = 0;

    qint32 pts = state.pts();
    qint32 qts = state.qts();
    qint32 date = state.date();
    qint32 seq = state.seq();
    qint32 unreadCount = state.unreadCount();

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE state SET value=:value WHERE name=:name");

    mDbManager.beginTransaction();

    query.bindValue(":name", "pts");
    query.bindValue(":value", pts);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating pts value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    rowsAffected += query.numRowsAffected();

    query.bindValue(":name", "qts");
    query.bindValue(":value", qts);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating qts value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    rowsAffected += query.numRowsAffected();

    query.bindValue(":name", "date");
    query.bindValue(":value", date);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating date value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    rowsAffected += query.numRowsAffected();

    query.bindValue(":name", "seq");
    query.bindValue(":value", seq);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating seq value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    rowsAffected += query.numRowsAffected();

    query.bindValue(":name", "unreadCount");
    query.bindValue(":value", unreadCount);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating unreadCount value:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    rowsAffected += query.numRowsAffected();

    mDbManager.finishTransaction();

    if (rowsAffected > 0) {
        return true;
    }
    return false;
}

UpdatesState Data::getState() {

    UpdatesState updatesState;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT name, value FROM state");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return updatesState;
    }

    qint32 seq = 0;
    qint32 pts = 0;
    qint32 qts = 0;
    qint32 date = 0;
    qint32 unreadCount = 0;

    while (query.next()) {
        QString nextName = query.value("name").toString();
        QString value = query.value("value").toString();

        if (nextName == "seq") {
            seq = value.toInt();
        } else if (nextName == "pts") {
            pts = value.toInt();
        } else if (nextName == "qts") {
            qts = value.toInt();
        } else if (nextName == "date") {
            date = value.toInt();
        } else if (nextName == "unreadCount") {
            unreadCount = value.toInt();
        }
    }

    updatesState.setSeq(seq);
    updatesState.setPts(pts);
    updatesState.setQts(qts);
    updatesState.setDate(date);
    updatesState.setUnreadCount(unreadCount);

    return updatesState;
}



bool Data::insertOrUpdateDialog(const Dialog &dialog) {

    if (!updateDialog(dialog)) {
        return insertDialog(dialog);
    }
    return true;
}

bool Data::insertDialog(const Dialog &dialog) {

    bool isChat = dialog.peer().classType() == Peer::typePeerChat;
    qint32 dialogId = isChat ? dialog.peer().chatId() : dialog.peer().userId();

    QSqlQuery query(mDbManager.database());

    MessageItem messageItem = topMessage(dialogId);
    qint32 date = messageItem.isEmpty() ? 0 : messageItem.date();

    query.prepare("INSERT INTO dialogs (id, isChat, unreadCount, date) VALUES (:id, :isChat, :unreadCount, :date)");
    query.bindValue(":id", dialogId);
    query.bindValue(":isChat", isChat);
    query.bindValue(":unreadCount", dialog.unreadCount());
    query.bindValue(":date", date);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }
    return query.numRowsAffected() > 0;
}

bool Data::updateDialog(const Dialog &dialog) {

    bool isChat = dialog.peer().classType() == Peer::typePeerChat;
    qint32 dialogId = isChat ? dialog.peer().chatId() : dialog.peer().userId();

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE dialogs SET isChat=:isChat, unreadCount=:unreadCount WHERE id=:id");
    query.bindValue(":id", dialogId);
    query.bindValue(":isChat", isChat);
    query.bindValue(":unreadCount", dialog.unreadCount());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

bool Data::deleteDialog(qint32 dialogId) {

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());
    query.prepare("DELETE FROM dialogs WHERE id=:id");
    query.bindValue(":id", dialogId);

    if (query.exec()) {
        deleteMessages(dialogId);
    } else {
        mDbManager.rollbackTransaction();
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    if (!dropMessagesView(dialogId)) {
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();
    return true;
}

bool Data::dropMessagesView(qint32 dialogId) {

    QSqlQuery query(mDbManager.database());

    QString dropSql = QString("DROP VIEW IF EXISTS messagesView_%1").arg((quint32) dialogId);
    query.prepare(dropSql);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error dropping messages view related to dialog" << dialogId << query.lastError() << query.lastQuery();
        return false;
    }
    return true;
}

// MESSAGEs operations
MessageItem Data::getMessage(qint32 messageId) {
    MessageItem messageItem;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT dialogId, fromId, toId, toIsChat, out, sent, unread, date, ttl, text, mediaType, "
                  "(SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                      "photoId=mediaId AND (type='s' OR type='m') ORDER BY size DESC LIMIT 1) AS photoSmall, "
                  "(SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                      "photoId=mediaId AND type='local' ORDER BY size DESC LIMIT 1) AS photoBigLocal, "
                  "(SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                      "photoId=mediaId ORDER BY size DESC LIMIT 1) AS photoBigRemote, "
                  "(SELECT localPath FROM mediaVideos WHERE id=mediaId) AS video, "
                  "(SELECT firstName FROM users WHERE id=fromId) AS fromFirstName, "
                  "(SELECT lastName FROM users WHERE id=fromId) AS fromLastName, "
                  "(SELECT phone FROM users WHERE id=fromId) AS fromPhone, "
                  "(SELECT type FROM users WHERE id=fromId) AS fromUserType, "
                  "(SELECT localPath FROM fileLocations WHERE rowid=(SELECT thumbnail FROM users WHERE id=fromId)) AS fromThumbnail, "
                  "(SELECT type FROM messageActions WHERE messageId=messages.id) AS actionType, "
                  "(SELECT title FROM messageActions WHERE messageId=messages.id) AS actionTitle, "
                  "(SELECT coalesce(users.firstName,'') || ' ' || coalesce(users.lastName,'') FROM users WHERE id IN "
                      "(SELECT userId FROM messageActions WHERE messageId=messages.id)) AS actionUser "
                  "FROM messages WHERE id=:id");
    query.bindValue(":id", messageId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error getting message with id" << messageId << "Error:" << query.lastError() << query.lastQuery();
        return messageItem;
    }

    if (query.next()) {
        qint32 dialogId = query.value("dialogId").toInt();
        qint32 fromId = query.value("fromId").toInt();
        QString fromFirstName = query.value("fromFirstName").toString();
        QString fromLastName = query.value("fromLastName").toString();
        QString fromPhone = query.value("fromPhone").toString();
        QString fromThumbnail = query.value("fromThumbnail").toString();
        qint32 fromUserType = query.value("fromUserType").toInt();
        qint32 toId = query.value("toId").toInt();
        bool toIsChat = query.value("toIsChat").toBool();
        bool out = query.value("out").toBool();
        bool sent = query.value("sent").toBool();
        bool unread = query.value("unread").toBool();
        qint32 ttl = query.value("ttl").toInt();
        qint32 date = query.value("date").toInt();
        QString text = query.value("text").toString();
        qint32 mediaType = query.value("mediaType").toInt();
        QString photoSmall = query.value("photoSmall").toString();
        QString photoBigLocal = query.value("photoBigLocal").toString();
        QString photoBigRemote = query.value("photoBigRemote").toString();
        QString video = query.value("video").toString();
        quint32 actionType = query.value("actionType").toInt();
        QString actionTitle = query.value("actionTitle").toString();
        QString actionUser = query.value("actionUser").toString();

        QString photoBig = photoBigLocal.length() > 0 ? photoBigLocal : photoBigRemote;

        messageItem.setId(messageId);
        messageItem.setDialogId(dialogId);
        messageItem.setFromId(fromId);
        messageItem.setFromFirstName(fromFirstName);
        messageItem.setFromLastName(fromLastName);
        messageItem.setFromPhone(fromPhone);
        messageItem.setFromThumbnail(fromThumbnail);
        messageItem.setFromUserType(fromUserType);
        Peer peer(Peer::typePeerUser);
        if (toIsChat) {
            peer.setClassType(Peer::typePeerChat);
            peer.setChatId(toId);
        } else {
            peer.setUserId(toId);
        }
        messageItem.setToId(peer);
        messageItem.setOut(out);
        messageItem.setSent(sent);
        messageItem.setUnread(unread);
        messageItem.setTtl(ttl);
        messageItem.setDate(date);
        messageItem.setMessage(text);
        MessageMedia media((MessageMedia::MessageMediaType)mediaType);
        messageItem.setMedia(media);
        messageItem.setThumbnail(photoSmall);
        messageItem.setPhoto(photoBig);
        messageItem.setVideo(video);

        //Evaluate if downloaded
        bool downloaded = false;
        if (out) {
            downloaded = sent;
        } else {
            switch (mediaType){
            case MessageMedia::typeMessageMediaPhoto:
                downloaded = !photoBig.isEmpty();
                break;
            case MessageMedia::typeMessageMediaVideo:
                downloaded = !video.isEmpty();
                break;
            }
        }
        downloaded ? messageItem.setDownloadedPercentage(100) : messageItem.setDownloadedPercentage(0);
        messageItem.setDownloading(false);

        if (actionType) {
            MessageAction messageAction((MessageAction::MessageActionType)actionType);
            messageAction.setTitle(actionTitle);
            messageItem.setAction(messageAction);
            messageItem.setActionUser(actionUser);
        }

    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "queried message with id" << messageId << "has not been found";
    }

    return messageItem;
}

bool Data::insertOrUpdateMessage(const Message &message) {

    if (!updateMessage(message)) {
        return insertMessage(message);
    }
    return true;
}

bool Data::updateMessage(const Message &message) {
    return updateMessage(message.id(), message);
}

bool Data::updateMessage(qint32 msgOldId, const Message &message) {

    // first of all let's do an advanced query to see if register exists. This is done so here to avoid performing
    // all update message related operations in database in case exists media data (most of times updateMessage will
    // be called before insertMessage)
    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM messages WHERE id=:id");
    query.bindValue(":id", msgOldId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error checking if message exists:" << query.lastError() << query.lastQuery();
        return false;
    }

    if (!query.next()) {
        return false;
    }

    bool toIsChat = message.toId().classType() == Peer::typePeerChat;
    qint32 toId = toIsChat ? message.toId().chatId() : message.toId().userId();
    qint32 mediaType = message.media().classType();
    qint64 mediaId = 0;

    mDbManager.beginTransaction();

    // insert media if media type is not null
    if (mediaType != MessageMedia::typeMessageMediaEmpty) {

        switch (mediaType) {
        case MessageMedia::typeMessageMediaPhoto: {
            Photo photo = message.media().photo();
            mediaId = photo.id();

            if (updatePhoto(message.id(), msgOldId, photo) < 0) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not update message photo";
                mDbManager.rollbackTransaction();
                return false;
            }

            // if already not downloaded any photo size to local path, download it from server
            if (!hasThumbnailDownloaded(photo.id())) {
                downloadThumbnail(message.id(), photo, IncomingFile::MessagePhoto);
                downloadPhoto(message.id(), photo, IncomingFile::MessagePhoto);
            }
            break;
        }
        case MessageMedia::typeMessageMediaVideo:{
            Video video = message.media().video();
            mediaId = video.id();

            // First of all, let's see if the definitive media id exists.
            // Don't add media videos register if media exists (it could exists due to a message forward from another dialog, for instance)
            bool exists = existsVideo(mediaId) > 0;
            if (exists == -1) {
                mDbManager.rollbackTransaction();
                return false;
            }

            // if not exists, we should update or insert video and update or insert its thumbnail
            if (!exists) {

                qint32 affectedRows = updateVideo(msgOldId, video);
                if (affectedRows == -1) {
                    mDbManager.rollbackTransaction();
                    return false;
                }

                if (!affectedRows) {
                    if (insertVideo(video) == -1) {
                        mDbManager.rollbackTransaction();
                        return false;
                    }
                }

                // insert or update video thumbnail image
                qint32 fileLocationId = insertOrUpdatePhotoSize(msgOldId, mediaId, video.thumb());
                if (fileLocationId < 0){
                    mDbManager.rollbackTransaction();
                    return false;
                }

                // if inserted a thumbnail, check if downloaded and do it if not
                if (fileLocationId > 0 && !hasThumbnailDownloaded(video.id())) {
                    downloadPhotoSize(message.id(), video.id(), video.date(), IncomingFile::MessagePhoto, true,
                                      video.thumb(), fileLocationId);
                }
            }

            break;
        }
        }
    }

    // finally, regardless of having updated/inserted or nothing with media, update message
    if (message.action().classType() != MessageAction::typeMessageActionEmpty) {
        if (updateMessageAction(message) == -1) {
            mDbManager.rollbackTransaction();
            return false;
        }
    }

    query.prepare("UPDATE messages "
                  "SET id=:id, fromId=:fromId, toId=:toId, toIsChat=:toIsChat, out=:out, unread=:unread, date=:date, "
                  "text=:text, mediaType=:mediaType, mediaId=:mediaId, sent=1 "
                  "WHERE id=:oldId");
    query.bindValue(":id", message.id());
    query.bindValue(":fromId", message.fromId());
    query.bindValue(":toId", toId);
    query.bindValue(":toIsChat", toIsChat);
    query.bindValue(":out", message.out());
    query.bindValue(":unread", message.unread());
    query.bindValue(":date", message.date());
    query.bindValue(":text", message.message());
    query.bindValue(":mediaType", mediaType);
    query.bindValue(":mediaId", mediaId);
    query.bindValue(":oldId", msgOldId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();

    return query.numRowsAffected() > 0;
}

qint32 Data::existsVideo(qint64 videoId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM mediaVideos WHERE id=:id");
    query.bindValue(":id", videoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when checking if media video exists:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.next() ?  1 : 0;
}

qint32 Data::insertVideo(const Video &video) {

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO mediaVideos (id, accessHash, userId, date, caption, duration, mimeType, size, dcId, width, height) "
                  "VALUES (:id, :accessHash, :userId, :date, :caption, :duration, :mimeType, :size, :dcId, :width, :height)");
    query.bindValue(":id", video.id());
    query.bindValue(":accessHash", video.accessHash());
    query.bindValue(":userId", video.userId());
    query.bindValue(":date", video.date());
    query.bindValue(":caption", video.caption());
    query.bindValue(":duration", video.duration());
    query.bindValue(":mimeType", video.mimeType());
    query.bindValue(":size", video.size());
    query.bindValue(":dcId", video.dcId());
    query.bindValue(":width", video.w());
    query.bindValue(":height", video.h());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting media video:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::updateVideo(qint64 oldVideoId, const Video &video) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE mediaVideos "
                  "SET id=:id, accessHash=:accessHash, userId=:userId, date=:date, caption=:caption, duration=:duration, "
                  "mimeType=:mimeType ,size=:size, dcId=:dcId, width=:width, height=:height "
                  "WHERE id=:msgOldId");
    query.bindValue(":id", video.id());
    query.bindValue(":accessHash", video.accessHash());
    query.bindValue(":userId", video.userId());
    query.bindValue(":date", video.date());
    query.bindValue(":caption", video.caption());
    query.bindValue(":duration", video.duration());
    query.bindValue(":mimeType", video.mimeType());
    query.bindValue(":size", video.size());
    query.bindValue(":dcId", video.dcId());
    query.bindValue(":width", video.w());
    query.bindValue(":height", video.h());
    query.bindValue(":msgOldId", oldVideoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating media video previous register:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::updateVideo(qint64 mediaId, const QString &localPath) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE mediaVideos SET localPath=:localPath WHERE id=:mediaId");
    query.bindValue(":localPath", localPath);
    query.bindValue(":mediaId", mediaId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update video" << mediaId << "local path" << localPath
                                    << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

QString Data::getVideoLocalPath(qint64 videoId) {

    QString localPath;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT localPath FROM mediaVideos WHERE id=:videoId");
    query.bindValue(":videoId", videoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get video local path for videoId" << videoId
                                    << query.lastError() << query.lastQuery();
        return localPath;
    }

    if (!query.next()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Could not find any downloaded video for videoId" << videoId;
        return localPath;
    }

    return query.value("localPath").toString();
}

QString Data::getThumbnailLocalPath(qint64 photoId) {
    return getPhotoLocalPath(photoId, true);
}

QString Data::getPhotoLocalPath(qint64 photoId, bool isThumbnail) {

    QString localPath;

    QString thumbnailCondition = isThumbnail ? "AND (type='s' OR type='m' OR type='a' OR type='b' OR type='local')" : "";

    QSqlQuery query(mDbManager.database());

    QString sql = QString("SELECT localPath FROM fileLocations, photoSizes WHERE fileLocations.rowid=fileLocationId AND "
                  "photoId=:photoId %1 ORDER BY size DESC LIMIT 1").arg(thumbnailCondition);
    query.prepare(sql);
    query.bindValue(":photoId", photoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get photo local path for media" << photoId <<
                                       query.lastError() << query.lastQuery();
        return localPath;
    }

    if (!query.next()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Could not find any local path for photo" << photoId;
        return localPath;
    }

    return query.value("localPath").toString();
}

QList<qint32> Data::getMediaMessages(qint64 mediaId) {

    QList<qint32> messages;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM messages WHERE mediaId=:mediaId");
    query.bindValue(":mediaId", mediaId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get messages for media" << mediaId <<
                                       query.lastError() << query.lastQuery();
        return messages;
    }

    while (query.next()) {
        messages << query.value("id").toInt();
    }

    return messages;
}


qint32 Data::insertOrUpdatePhotoSize(qint64 oldMediaId, qint64 newMediaId, const PhotoSize &photoSize) {

    qint32 fileLocationId = 0;

    if (photoSize.classType() != PhotoSize::typePhotoSizeEmpty)  {

        qint32 affectedRows = updatePhotoSize(oldMediaId, newMediaId, photoSize);

        if (affectedRows == -1) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update old photo size";
            mDbManager.rollbackTransaction();
            return -1;
        }

        if (affectedRows > 0) {

            fileLocationId = getPhotoSizeFileLocationId(newMediaId, photoSize.type());
            if (fileLocationId != -1) {
                if (updateFileLocation(fileLocationId, photoSize.location()) == -1) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update old photo size file location";
                    mDbManager.rollbackTransaction();
                    return -1;
                }
            } else {
                fileLocationId = insertFileLocation(photoSize.location());
                if (fileLocationId == -1) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert old photo size new file location";
                    mDbManager.rollbackTransaction();
                    return -1;
                }

                if (updatePhotoSizeLocation(newMediaId, photoSize.type(), fileLocationId) == -1) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not update old photo size with new file location id";
                    mDbManager.rollbackTransaction();
                    return -1;
                }
            }

        } else {

            fileLocationId = insertFileLocation(photoSize.location());
            if (fileLocationId == -1) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new file location";
                mDbManager.rollbackTransaction();
                return -1;
            }

            affectedRows = insertPhotoSize(newMediaId, photoSize, fileLocationId);
            if (affectedRows == -1) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new photo size";
                mDbManager.rollbackTransaction();
                return -1;
            }
        }
    }
    return fileLocationId;
}

qint32 Data::insertFileLocation(const QString &localPath) {

    if (localPath.isEmpty()) {
        return 0;
    }

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO fileLocations (localPath) VALUES (:localPath)");
    query.bindValue(":localPath", localPath);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new file location:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.lastInsertId().toInt();
}

qint32 Data::insertFileLocation(const FileLocation &fileLocation) {

    if (fileLocation.classType() == FileLocation::typeFileLocationUnavailable) {
        return 0;
    }

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO fileLocations (dcId, localId, secret, volumeId) "
                  "VALUES (:dcId, :localId, :secret, :volumeId)");
    query.bindValue(":dcId", fileLocation.dcId());
    query.bindValue(":localId", fileLocation.localId());
    query.bindValue(":secret", fileLocation.secret());
    query.bindValue(":volumeId", fileLocation.volumeId());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new file location:" << query.lastError() << query.lastQuery();
        return -1;
    }

    if (query.numRowsAffected() > 0) {
        return query.lastInsertId().toInt();
    }
    return -1;
}

qint32 Data::updateFileLocation(qint32 fileLocationId, const FileLocation &fileLocation) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE fileLocations "
                  "SET dcId=:dcId, localId=:localId, secret=:secret, volumeId=:volumeId "
                  "WHERE rowid=:fileLocationId");
    query.bindValue(":fileLocationId", fileLocationId);
    query.bindValue(":dcId", fileLocation.dcId());
    query.bindValue(":localId", fileLocation.localId());
    query.bindValue(":secret", fileLocation.secret());
    query.bindValue(":volumeId", fileLocation.volumeId());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when updating file location:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::insertPhotoSize(qint64 mediaId, const PhotoSize &photoSize, qint32 fileLocationId) {

    QString type = photoSize.type().isEmpty() ? "local" : photoSize.type();

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO photoSizes (photoId, type, width, height, size, fileLocationId) "
                  "VALUES (:photoId, :type, :width, :height, :size, :fileLocationId)");
    query.bindValue(":photoId", mediaId);
    query.bindValue(":type", type);
    query.bindValue(":width", photoSize.w());
    query.bindValue(":height", photoSize.h());
    query.bindValue(":size", photoSize.size());
    query.bindValue(":fileLocationId", fileLocationId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new photo size for the thumbnail:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::updatePhotoSize(qint64 oldPhotoId, qint64 newPhotoId, const PhotoSize &photoSize) {

    QString type = photoSize.type().isEmpty() ? "local" : photoSize.type();

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE photoSizes SET photoId=:photoId, width=:width, height=:height, size=:size "
                          "WHERE photoId=:oldPhotoId AND type=:type");
    query.bindValue(":photoId", newPhotoId);
    query.bindValue(":width", photoSize.w());
    query.bindValue(":height", photoSize.h());
    query.bindValue(":size", photoSize.size());
    query.bindValue(":oldPhotoId", oldPhotoId);
    query.bindValue(":type", type);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update photo size:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::updatePhotoSizeLocation(qint64 photoId, const QString &type, const qint32 fileLocationId) {

    QString typeParam = type.isEmpty() ? "local" : type;

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE photoSizes SET fileLocationId=:fileLocationId WHERE photoId=:photoId AND type=:type");
    query.bindValue(":fileLocationId", fileLocationId);
    query.bindValue(":photoId", photoId);
    query.bindValue(":type", typeParam);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update photo size file location id value:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::getPhotoSizeFileLocationId(qint64 photoId, const QString &type) {

    QString typeParam = type.isEmpty() ? "local" : type;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT fileLocationId FROM photoSizes WHERE photoId=:photoId AND type=:type");
    query.bindValue(":photoId", photoId);
    query.bindValue(":type", typeParam);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when reading fileLocationId from photoSize:" << query.lastError() << query.lastQuery();
        return -1;
    }

    if (query.next()) {
        return query.value("fileLocationId").toInt();
    }

    return -1;
}

bool Data::appendOutgoingMessage(qint32 recipientId, qint64 randomId, const Message &message, const QString &mediaLocalPath, qint64 uploadFileId, const QString &thumbnailLocalPath) {
    if (message.id() == 0) return false; // Sanity check, this has happened before.

    // search for message related dialog
    bool newDialog = !existsDialog(recipientId);

    if (!insertMessage(message, randomId, false, mediaLocalPath, thumbnailLocalPath)) {
        return false;
    }

    if (newDialog) {
        DialogItem dialogItem = getDialog(recipientId);
        QList<DialogItem> dialogs;
        dialogs << dialogItem;
        Q_EMIT dialogsAdded(dialogs);
    }

    const ContactItem &sender = getUser(message.fromId());
    MessageItem messageItem(message, sender, false);
    Q_EMIT outgoingMessageAdded(recipientId, messageItem);

    // 'local' is the size for sent photos until having a server location and different sizes
    // if there is a mediaLocalPath and media type is photo ro video, emit photo updated signal. Same for video
    if (!mediaLocalPath.isEmpty()) {
        switch (message.media().classType()) {
        case MessageMedia::typeMessageMediaPhoto:
            Q_EMIT messagePhotoUpdated(message.id(), mediaLocalPath, MEDIA_LOCAL_FILES_TYPE);
            break;
        case MessageMedia::typeMessageMediaVideo:
            Q_EMIT messagePhotoUpdated(message.id(), thumbnailLocalPath, MEDIA_LOCAL_FILES_TYPE);
            Q_EMIT messageVideoUpdated(message.id(), mediaLocalPath);
            break;
        }

        mActiveUploads.insert(uploadFileId, message.id());

        Q_EMIT downloadingUpdated(message.id(), true);
    }

    return true;
}

/**
 * Inserts a message, creates related dialog if not existing and increases its unreadCount if needed
 * */
bool Data::insertMessage(const Message &message, qint64 randomId, bool sent, const QString &mediaLocalPath, const QString &thumbnailLocalPath) {

    // There must be a dialog related to the message to add. If not, create new one
    bool toIsChat = message.toId().classType() == Peer::typePeerChat;
    qint32 toId = toIsChat ? message.toId().chatId() : message.toId().userId();

    qint32 dialogId = message.fromId();
    if (toIsChat || message.out()){
        dialogId = toId;
    }

    bool isAction = message.action().classType() != MessageAction::typeMessageActionEmpty;
    MessageMedia::MessageMediaType mediaType =
            message.action().classType() == MessageAction::typeMessageActionChatEditPhoto ?
                MessageMedia::typeMessageMediaPhoto :
                message.media().classType();

    bool mustIncreaseUnreadCount = !message.out() && message.unread();

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id, date FROM dialogs WHERE id=:id");
    query.bindValue(":id", dialogId);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    mDbManager.beginTransaction();

    if (!query.next()) {
        query.prepare("INSERT INTO dialogs (id, isChat, unreadCount, date) VALUES (:id, :isChat, :unreadCount, :date)");
        query.bindValue(":id", dialogId);
        query.bindValue(":isChat", toIsChat);
        mustIncreaseUnreadCount ? query.bindValue(":unreadCount", 1) : query.bindValue(":unreadCount", 0);
        query.bindValue(":date", message.date());

        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return false;
        }
    } else {
        qint32 dialogDate = query.value("date").toInt();
        bool mustUpdateDate = message.date() > dialogDate;

        if (mustIncreaseUnreadCount || mustUpdateDate) {
            // dialog already exists, update its date and increase unread count if needed.
            QString arg1 = mustIncreaseUnreadCount ? "unreadCount=unreadCount+1" : "";
            QString arg2 = mustUpdateDate ? "date=:date" : "";
            // add comma in the middle if both args are not empty
            if (arg1.length() > 0 && arg2.length() > 0) {
                arg1 += ",";
            }
            QString sql = QString("UPDATE dialogs SET %1 %2 WHERE id=:id").arg(arg1, arg2);

            query.prepare(sql);
            query.bindValue(":id", dialogId);
            if (mustUpdateDate) {
                query.bindValue(":date", message.date());
            }

            if (!query.exec()) {
                qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
                mDbManager.rollbackTransaction();
                return false;
            }
        }
    }

    qint64 mediaId = 0;

    // insert media if media type is not null
    if (mediaType != MessageMedia::typeMessageMediaEmpty) {

        switch (mediaType) {
        case MessageMedia::typeMessageMediaPhoto: {
            Photo photo = isAction ? message.action().photo() : message.media().photo();
            mediaId = photo.id();

            // check mediaLocalPath. If not empty, it means this is a outgoing media file, so insert temporal registers
            // in message, fileLocations, mediaPhotos and photoSizes tables
            if (mediaLocalPath.length()) {

                query.prepare("INSERT INTO mediaPhotos (id) VALUES (:id)");
                query.bindValue(":id", photo.id());
                if (!query.exec()) {
                    qCWarning(TG_PLUGIN_LOGIC) << "Error when inserting message media photo:" << query.lastError() << query.lastQuery();
                    mDbManager.rollbackTransaction();
                    return false;
                }

                qint32 fileLocationId = insertFileLocation(mediaLocalPath);
                if (fileLocationId <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert local path in a new fileLocations register";
                    mDbManager.rollbackTransaction();
                    return false;
                }

                // temporally set 'local' size for outgoing photos until getting real sizes
                query.clear();
                query.prepare("INSERT INTO photoSizes (photoId, type, fileLocationId) VALUES (:photoId, :type, :fileLocationId)");
                query.bindValue(":photoId", photo.id());
                query.bindValue(":type", MEDIA_LOCAL_FILES_TYPE);
                query.bindValue(":fileLocationId", fileLocationId);
                if (!query.exec()) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new size for the photo media:" << query.lastError() << query.lastQuery();
                    mDbManager.rollbackTransaction();
                    return false;
                }

            } else {

                qint32 exists = existsPhoto(photo.id());
                if (exists == -1) {
                    mDbManager.rollbackTransaction();
                    return false;
                }

                if (exists) {
                    QString photoLocalPath = getPhotoLocalPath(photo.id());
                    QString thumbnailLocalPath = getThumbnailLocalPath(photo.id());

                    if (!photoLocalPath.isEmpty()) {
                        Q_EMIT messagePhotoUpdated(message.id(), photoLocalPath, false);
                    }

                    if (!thumbnailLocalPath.isEmpty()) {
                        Q_EMIT messagePhotoUpdated(message.id(), thumbnailLocalPath, true);
                    }

                } else {
                    insertPhoto(photo);
                    IncomingFile::IncomingFileType incomingFileType = isAction ? IncomingFile::MessageActionPhoto : IncomingFile::MessagePhoto;
                    downloadThumbnail(message.id(), photo, incomingFileType);
                    downloadPhoto(message.id(), photo, incomingFileType);
                }

               }
            break;
        }
        case MessageMedia::typeMessageMediaVideo: {
            Video video = message.media().video();
            mediaId = video.id();
            // check mediaLocalPath. If not empty, it means this is a outgoing media file, so insert temporal registers
            // in message, fileLocations, mediaVideos and photoSize tables
            if (mediaLocalPath.length()) {

                query.prepare("INSERT INTO mediaVideos (id, duration, width, height, localPath) VALUES (:id, :duration, :width, :height, :localPath)");
                query.bindValue(":id", video.id());
                query.bindValue(":duration", video.duration());
                query.bindValue(":width", video.w());
                query.bindValue(":height", video.h());
                query.bindValue(":localPath", mediaLocalPath);
                if (!query.exec()) {
                    qCWarning(TG_PLUGIN_LOGIC) << "Error when inserting message media video:" << query.lastError() << query.lastQuery();
                    mDbManager.rollbackTransaction();
                    return false;
                }

                if (thumbnailLocalPath.length()) {
                    query.prepare("INSERT INTO fileLocations (localPath) VALUES (:localPath)");
                    query.bindValue(":localPath", thumbnailLocalPath);
                    if (!query.exec()) {
                        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new fileLocation for the video thumbnail media:" << query.lastError() << query.lastQuery();
                        mDbManager.rollbackTransaction();
                        return false;
                    }

                    qint32 fileLocationId = query.lastInsertId().toInt();

                    // set 'local' size for thumbnail
                    query.prepare("INSERT INTO photoSizes (photoId, type, fileLocationId) VALUES (:photoId, :type, :fileLocationId)");
                    query.bindValue(":photoId", video.id());
                    query.bindValue(":type", MEDIA_LOCAL_FILES_TYPE);
                    query.bindValue(":fileLocationId", fileLocationId);
                    if (!query.exec()) {
                        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting new photo size for the video thumbnail:" << query.lastError() << query.lastQuery();
                        mDbManager.rollbackTransaction();
                        return false;
                    }
                }

            } else {

                // First of all, let's see if the definitive media id exists.
                // Don't add media videos register if media exists (it could exists due to a message forward from another dialog, for instance)
                qint32 exists = existsVideo(mediaId);
                if (exists == -1) {
                    mDbManager.rollbackTransaction();
                    return false;
                }

                if (exists) {
                    QString videoLocalPath = getVideoLocalPath(mediaId);
                    QString thumbnailLocalPath = getThumbnailLocalPath(mediaId);

                    if (!thumbnailLocalPath.isEmpty()) {
                        Q_EMIT messagePhotoUpdated(message.id(), thumbnailLocalPath, true);
                    }

                    if (!videoLocalPath.isEmpty()) {
                        Q_EMIT messageVideoUpdated(message.id(), videoLocalPath);
                    }

                } else {
                    // if not exists, we should insert video and insert its thumbnail if not empty
                    if (insertVideo(video) == -1) {
                        mDbManager.rollbackTransaction();
                        return false;
                    }

                    PhotoSize thumb = video.thumb();
                    if (thumb.classType() != PhotoSize::typePhotoSizeEmpty) {
                        qint32 fileLocationId = insertFileLocation(thumb.location());
                        if (fileLocationId == -1) {
                            mDbManager.rollbackTransaction();
                            return false;
                        }

                        if (insertPhotoSize(mediaId, thumb, fileLocationId) == -1) {
                            mDbManager.rollbackTransaction();
                            return false;
                        }

                        downloadPhotoSize(message.id(), video.id(), video.date(), IncomingFile::MessagePhoto,
                                          true, video.thumb(), fileLocationId);
                    }
                }

            }
            break;
        }
        }
    }

    if (message.action().classType() != MessageAction::typeMessageActionEmpty) {
        if (insertMessageAction(message) <= 0) {
            mDbManager.rollbackTransaction();
            return false;
        }
    }

    query.prepare("INSERT INTO messages (id, dialogId, randomId, fromId, toId, toIsChat, out, unread, date, ttl, text, mediaType, mediaId, sent) "
                  "VALUES (:id, :dialogId, :randomId, :fromId, :toId, :toIsChat, :out, :unread, :date, :ttl, :text, :mediaType, :mediaId, :sent)");
    query.bindValue(":id", message.id());
    query.bindValue(":dialogId", dialogId);
    query.bindValue(":randomId", randomId);
    query.bindValue(":fromId", message.fromId());
    query.bindValue(":toId", toId);
    query.bindValue(":toIsChat", toIsChat);
    query.bindValue(":out", message.out());
    query.bindValue(":unread", message.unread());
    query.bindValue(":date", message.date());
    query.bindValue(":ttl", message.ttl());
    query.bindValue(":text", message.message());
    query.bindValue(":mediaType", mediaType);
    query.bindValue(":mediaId", mediaId);
    query.bindValue(":sent", sent);
    if (!query.exec()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Trying to insert an existing message in db:" << message.id()
                                   << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();

    surfaceUnreadCount();

    return true;
}

bool Data::updateMessage(qint32 oldMessageId, qint32 newMessageId, qint32 date) {

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());

    qint32 dialogId = getDialogIdFromMessageId(oldMessageId);
    if (dialogId == 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "can't update message, dialog does not exist";
        mDbManager.rollbackTransaction();
        return false;
    }

    query.prepare("SELECT date FROM dialogs WHERE id=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error getting date from dialog" << dialogId << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    if (!query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not find any related register to dialog" << dialogId;
        mDbManager.rollbackTransaction();
        return false;
    }

    qint32 dialogDate = query.value("date").toInt();

    if (date > dialogDate) {
        query.prepare("UPDATE dialogs SET date=:date WHERE id=:dialogId");
        query.bindValue(":date", date);
        query.bindValue(":dialogId", dialogId);
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error updating date for dialog" << dialogId << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return false;
        }
    }

    //when updating message date is because and update date is taken from server. Update sent to true also
    query.prepare("UPDATE messages SET date=:date, id=:newId, sent=1 WHERE id=:oldId");
    query.bindValue(":date", date);
    query.bindValue(":oldId", oldMessageId);
    query.bindValue(":newId", newMessageId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    if (query.numRowsAffected() == 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "No row was updated during this execution:" << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();

    return true;
}

bool Data::updateMessageId(qint64 randomId, qint32 id) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE messages SET id=:id, sent=1 WHERE randomId=:randomId");
    query.bindValue(":id", id);
    query.bindValue(":randomId", randomId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update message with randomId" << randomId
                                    << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

bool Data::deleteMessages(qint32 dialogId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT isChat FROM dialogs WHERE id=:dialogId");
    query.bindValue(":dialogId", dialogId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    if (query.next()) {
        query.clear();
        query.prepare("DELETE FROM messages WHERE dialogId=:dialogId");
        query.bindValue(":dialogId", dialogId);
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
            return false;
        }
    }

    return true;
}

bool Data::deleteMessages(const QList<qint32> msgIds) {
    qCDebug(TG_PLUGIN_LOGIC) << "Data::deleteMessages";

    if (!msgIds.count()) {
        qCDebug(TG_PLUGIN_LOGIC) << "Data::deleteMessages - no messages to delete";
        return false;
    }

    qint32 dialogId = getDialogIdFromMessageId(msgIds.at(0));
    if (dialogId == 0) {
        qCDebug(TG_PLUGIN_LOGIC) << "Data::deleteMessages - could not find dialog";
        return false;
    }

    qint32 amountToDecrease = 0;

    QSqlQuery deleteQuery(mDbManager.database());
    deleteQuery.prepare("DELETE FROM messages WHERE id=:id");

    mDbManager.beginTransaction();

    Q_FOREACH (qint32 msgId, msgIds) {

        if (isMessageIncomingAndUnread(msgId)) {
            amountToDecrease++;
        }

        // delete the message
        deleteQuery.bindValue(":id", msgId);
        if (!deleteQuery.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error:" << deleteQuery.lastError() << deleteQuery.lastQuery();
            mDbManager.rollbackTransaction();
            return false;
        } else {
            qCDebug(TG_PLUGIN_LOGIC) << "Data::deleteMessages - message deleted" << msgId;
        }
    }

    decreaseUnreadCount(dialogId, amountToDecrease);

    mDbManager.finishTransaction();

    Q_EMIT messagesDeleted(msgIds);
}

bool Data::isMessageIncomingAndUnread(qint32 msgId) {

    QSqlQuery unreadQuery(mDbManager.database());
    unreadQuery.prepare("SELECT unread, out FROM messages WHERE id=:id");

    unreadQuery.bindValue(":id", msgId);
    if (!unreadQuery.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << unreadQuery.lastError() << unreadQuery.lastQuery();
        return false;
    }

    if (unreadQuery.next()) {
        bool unread = unreadQuery.value("unread").toBool();
        bool out = unreadQuery.value("out").toBool();
        return unread && !out;
    } else {
        qCCritical(TG_PLUGIN_LOGIC) << "Not found any msg related to id" << msgId;
        return false;
    }
}

bool Data::markChatAsRead(qint32 chatId) {
    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE dialogs SET unreadCount=0 WHERE id=:chatId");
    query.bindValue(":chatId", chatId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when marking as read, chatId:" << chatId
                                    << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    query.prepare(("UPDATE messages SET unread = 0 WHERE dialogId=:dialogId AND out = 0 AND unread = 1"));
    query.bindValue(":dialogId", chatId);
    if (query.exec()) {
        surfaceUnreadCount();
    }

    Q_EMIT chatMarkedAsRead(chatId);

    return true;
}

bool Data::markMessagesAsRead(const QList<qint32> &messages) {

    mDbManager.beginTransaction();

    QSqlQuery markQuery(mDbManager.database());
    markQuery.prepare("UPDATE messages SET unread=0 WHERE id=:id");


    Q_FOREACH (qint32 msgId, messages) {

        bool incomingAndUnread = isMessageIncomingAndUnread(msgId);

        markQuery.bindValue(":id", msgId);
        if (!markQuery.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error when marking as read msgId:" << msgId << "Error:" << markQuery.lastError() << markQuery.lastQuery();
            mDbManager.rollbackTransaction();
            return false;
        }

        if (incomingAndUnread) {
            qint32 markedMessageCount = markQuery.numRowsAffected();

            if (markedMessageCount > 0) {
                qint32 dialogId = getDialogIdFromMessageId(msgId);
                if (dialogId <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Not found related dialog for msgId:" << msgId;
                    mDbManager.rollbackTransaction();
                    return false;
                }

                if (!decreaseUnreadCount(dialogId, markedMessageCount)) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Error decreasing unread count for dialog" << dialogId
                                                << "when marked as read msgId" << msgId;
                    mDbManager.rollbackTransaction();
                    return false;
                }
            }
        }
    }

    mDbManager.finishTransaction();

    surfaceUnreadCount();

    return true;
}

qint32 Data::getMessageId(qint64 randomId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM messages WHERE randomId=:randomId");
    query.bindValue(":randomId", randomId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get randomId message related id" << randomId
                                    << query.lastError() << query.lastQuery();
        return 0;
    }

    if (!query.next()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not find any message related to randomId" << randomId;
        return 0;
    }

    return query.value("id").toInt();
}

qint32 Data::insertOrUpdateMessageAction(const Message &message) {

    qint32 updated = updateMessageAction(message);
    switch (updated) {
    case -1 : return -1;
    case 0: return insertMessageAction(message);
    default: return updated;
    }
}

qint32 Data::insertMessageAction(const Message &message) {

    MessageAction action = message.action();

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO messageActions (messageId, type, title, userId) "
                  "VALUES (:messageId, :type, :title, :userId)");
    query.bindValue(":messageId", message.id());
    query.bindValue(":type", action.classType());
    query.bindValue(":title", action.title());
    query.bindValue(":userId", action.userId());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert message action" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    return query.numRowsAffected();
}

qint32 Data::updateMessageAction(const Message &message) {

    MessageAction action = message.action();

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE messageActions SET type=:type, title=:title, userId=:userId "
                  "WHERE messageId=:messageId");
    query.bindValue(":type", action.classType());
    query.bindValue(":title", action.title());
    query.bindValue(":userId", action.userId());
    query.bindValue(":messageId", message.id());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating message action with id" << message.id() << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

bool Data::executeAction(const Message &message) {

    qint32 chatId = message.toId().chatId();
    MessageAction::MessageActionType actionType = message.action().classType();

    switch (actionType) {
    case MessageAction::typeMessageActionChatDeletePhoto: {
        if (!deleteChatPhoto(chatId)) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not delete chat photo";
            return false;
        }
        Q_EMIT dialogThumbnailUpdated(chatId, QString());
        break;
    }
    case MessageAction::typeMessageActionChatEditTitle: {
        QString title = message.action().title();
        if (!updateChatTitle(chatId, title)) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update chat title";
            return false;
        }
        Q_EMIT dialogTitleUpdated(chatId, title);
        break;
    }
    case MessageAction::typeMessageActionChatEditPhoto: {
        mMessageIdsToUpdatePhoto.insert(message.id(), true);
        break;
    }
    case MessageAction::typeMessageActionChatAddUser:
        //TODO COMPLETE
        break;
    case MessageAction::typeMessageActionChatDeleteUser:
        //TODO COMPLETE
        break;
    default:
        break;
    }

    return true;
}


// USERs operations
ContactItem Data::getUser(qint32 id) {

    ContactItem contact;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT firstName, lastName, phone, lastSeenOnline, online, type, accessHash, "
                  "(SELECT localPath FROM fileLocations WHERE rowid=users.thumbnail) AS thumbnail, "
                  "(SELECT localPath FROM fileLocations WHERE rowid=users.bigPhoto) AS bigPhoto "
                  "FROM users "
                  "WHERE id=:id");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when queried user with id:" << id << query.lastError() << query.lastQuery();
        return contact;
    }

    if (query.next()) {

        contact.setId(id);
        contact.setFirstName(query.value("firstName").toString());
        contact.setLastName(query.value("lastName").toString());
        contact.setPhone(query.value("phone").toString());

        UserStatus us(UserStatus::typeUserStatusOffline);
        if (query.value("online").toBool()) {
            us.setClassType(UserStatus::typeUserStatusOnline);
        }
        us.setWasOnline(query.value("lastSeenOnline").toInt());
        contact.setClassType((User::UserType)query.value("type").toInt());
        contact.setAccessHash(query.value("accessHash").toLongLong());
        contact.setThumbnail(query.value("thumbnail").toString());
        contact.setBigPhoto(query.value("bigPhoto").toString());

        contact.setStatus(us);
    } else {
        qCWarning(TG_PLUGIN_LOGIC) << "queried user with id" << id << "has not been found";
    }
    return contact;
}

QList<ContactItem> Data::getUsers() {
    QList<ContactItem> result;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id, firstName, lastName, phone, lastSeenOnline, online, type, accessHash, "
                  "(SELECT localPath FROM fileLocations WHERE rowid=users.thumbnail) AS thumbnail, "
                  "(SELECT localPath FROM fileLocations WHERE rowid=users.bigPhoto) AS bigPhoto "
                  "FROM users");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "an error has happened when requesting complete contacts list";
        return result;
    }
    while (query.next()) {
        ContactItem contact;
        contact.setId(query.value("id").toInt());
        contact.setFirstName(query.value("firstName").toString());
        contact.setLastName(query.value("lastName").toString());
        contact.setPhone(query.value("phone").toString());
        contact.setThumbnail(query.value("thumbnail").toString());
        contact.setBigPhoto(query.value("bigPhoto").toString());
        UserStatus us(UserStatus::typeUserStatusOffline);
        if (query.value("online").toBool()) {
            us.setClassType(UserStatus::typeUserStatusOnline);
        }
        us.setWasOnline(query.value("lastSeenOnline").toInt());
        contact.setStatus(us);
        qint32 userType = query.value("type").toInt();
        contact.setClassType((User::UserType)userType);
        contact.setAccessHash(query.value("accessHash").toLongLong());

        result.append(contact);
    }
    return result;
}

QList<qint32> Data::getUserSecretChats(qint32 userId) {

    QList<qint32> secretChats;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM secretChats WHERE peerId=:userId");
    query.bindValue(":userId", userId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get user secret chats" << query.lastQuery() << query.lastError();
        return secretChats;
    }

    while (query.next()) {
        secretChats << query.value("id").toInt();
    }

    return secretChats;
}

bool Data::isContact(qint32 userId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT isContacts FROM users WHERE id=:userId");
}

bool Data::insertOrUpdateUser(const User &user) {

    if (!updateUser(user)) {
        return insertUser(user);
    }

    return true;
}

FileLocation Data::getFileLocation(qint32 fileLocationId) {

    FileLocation serverLocation;

    QSqlQuery query(mDbManager.database());
    query.prepare("SELECT dcId, localId, secret, volumeId FROM fileLocations WHERE rowid=:fileLocationId");
    query.bindValue(":fileLocationId", fileLocationId);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get file location for rowid" << fileLocationId;
    } else if (query.next()) {
        serverLocation.setClassType(FileLocation::typeFileLocation);
        serverLocation.setDcId(query.value("dcId").toInt());
        serverLocation.setLocalId(query.value("localId").toInt());
        serverLocation.setSecret(query.value("secret").toLongLong());
        serverLocation.setVolumeId(query.value("volumeId").toLongLong());
    }

    return serverLocation;
}

bool Data::deleteFileLocation(qint32 fileLocationId) {
    QSqlQuery query(mDbManager.database());
    query.prepare("DELETE FROM fileLocations WHERE rowid=:fileLocationId");
    query.bindValue(":fileLocationId", fileLocationId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not delete file location with id" << fileLocationId;
        return false;
    }
    return query.numRowsAffected();
}

qint32 Data::insertOrUpdateUserThumbnailFileLocation(const User &user) {
    return insertOrUpdateFileLocation(user.id(), user.photo().photoSmall(), true, true);
}

qint32 Data::insertOrUpdateUserBigPhotoFileLocation(const User &user) {
    return insertOrUpdateFileLocation(user.id(), user.photo().photoBig(), true, false);
}

qint32 Data::insertOrUpdateChatThumbnailFileLocation(const Chat &chat) {
    return insertOrUpdateFileLocation(chat.id(), chat.photo().photoSmall(), false, true);
}

qint32 Data::insertOrUpdateChatBigPhotoFileLocation(const Chat &chat) {
    return insertOrUpdateFileLocation(chat.id(), chat.photo().photoBig(), false, false);
}


qint32 Data::insertOrUpdateFileLocation(qint32 id, const FileLocation &serverLocation, bool isUser, bool isThumbnail) {
    QString field = isThumbnail ? "thumbnail" : "bigPhoto";
    QString table = isUser ? "users" : "chats";
    QString sql = QString("SELECT %1 FROM %2 WHERE id=:id").arg(field, table);

    QSqlQuery query(mDbManager.database());
    query.prepare(sql);
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get" << field << "file location rowid in table" << table << "for" << id
                                       << query.lastError() << query.lastQuery();
        return -1;
    }

    qint32 rowId = 0;
    if (query.next()) {
        rowId = query.value(field).toInt();
        if (rowId) {
            const FileLocation &dbLocation = getFileLocation(rowId);
            if (dbLocation == serverLocation) {
                return rowId;
            }
        }
    }

    rowId = insertFileLocation(serverLocation);

    if (rowId > 0 && isThumbnail) {
        getProfilePhotoFileFromServer(serverLocation, id, true, isThumbnail);
    }

    return rowId;
}

bool Data::existsUser(qint32 userId) {
    return exists(userId, true);
}

bool Data::exists(qint32 id, bool isUser) {

    QString table = isUser ? "users" : "chats";
    QString sql = QString("SELECT 1 FROM %1 WHERE id=:id").arg(table);

    QSqlQuery query(mDbManager.database());
    query.prepare(sql);
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not check if exists" << id << "in" << table
                                    << query.lastError() << query.lastQuery();
        return false;
    }

    return query.next();
}


bool Data::updateUser(const User &user) {

    if (!existsUser(user.id())) {
        return false;
    }

    bool online = user.status().classType() == UserStatus::typeUserStatusOnline;

    qint32 thumbnailRowId = insertOrUpdateUserThumbnailFileLocation(user);
    if (thumbnailRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update thumbnail for user" << user.id();
        return false;
    }

    qint32 bigPhotoRowId = insertOrUpdateUserBigPhotoFileLocation(user);
    if (bigPhotoRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update big photo for user" << user.id();
        if (thumbnailRowId > 0) {
            deleteFileLocation(thumbnailRowId);
        }
        return false;
    }

    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE users SET firstName=:firstName, lastName=:lastName, phone=:phone, lastSeenOnline=:lastSeenOnline, "
                  "online=:online, type=:type, accessHash=:accessHash, thumbnail=:thumbnail, bigPhoto=:bigPhoto WHERE id=:id");
    query.bindValue(":firstName", user.firstName());
    query.bindValue(":lastName", user.lastName());
    query.bindValue(":phone", user.phone());
    query.bindValue(":lastSeenOnline", user.status().wasOnline());
    query.bindValue(":online", online);
    query.bindValue(":type", user.classType());
    query.bindValue(":accessHash", user.accessHash());
    query.bindValue(":thumbnail", thumbnailRowId);
    query.bindValue(":bigPhoto", bigPhotoRowId);
    query.bindValue(":id", user.id());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update user" << user.id() << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

bool Data::insertUser(const User &user) {

    bool online = user.status().classType() == UserStatus::typeUserStatusOnline;

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());

    FileLocation thumbnailServerLocation = user.photo().photoSmall();
    qint32 thumbnailRowId = insertFileLocation(thumbnailServerLocation);
    if (thumbnailRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert thumbnail server location related to user" << user.id();
        mDbManager.rollbackTransaction();
        return false;
    }

    FileLocation photoBigServerLocation = user.photo().photoBig();
    qint32 bigPhotoRowId = insertFileLocation(photoBigServerLocation);
    if (bigPhotoRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert big photo server location related to user" << user.id();
        mDbManager.rollbackTransaction();
        return false;
    }

    // try to append user to database
    query.prepare("INSERT INTO users (id, firstName, lastName, phone, lastSeenOnline, online, type, accessHash, thumbnail, bigPhoto) "
                  "VALUES (:id, :firstName, :lastName, :phone, :lastSeenOnline, :online, :type, :accessHash, :thumbnail, :bigPhoto)");
    query.bindValue(":id", user.id());
    query.bindValue(":firstName", user.firstName());
    query.bindValue(":lastName", user.lastName());
    query.bindValue(":phone", user.phone());
    query.bindValue(":lastSeenOnline", user.status().wasOnline());
    query.bindValue(":online", online);
    query.bindValue(":type", user.classType());
    query.bindValue(":accessHash", user.accessHash());
    query.bindValue(":thumbnail", thumbnailRowId);
    query.bindValue(":bigPhoto", bigPhotoRowId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new user:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();

    if (thumbnailRowId > 0) {
        getProfilePhotoFileFromServer(thumbnailServerLocation, user.id(), true, true);
    }

    return query.numRowsAffected() > 0;
}


bool Data::updateUserName(qint32 userId, const QString &firstName, const QString &lastName) {

    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE users SET firstName=:firstName, lastName=:lastName WHERE id=:id");
    query.bindValue(":firstName", firstName);
    query.bindValue(":lastName", lastName);
    query.bindValue(":id", userId);

    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update user name for userId" << userId
                                    << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

bool Data::updateUserIsContact(qint32 userId, bool isContact, bool hasPhone) {

    qint32 type = User::typeUserContact;
    if (!isContact) {
        if (hasPhone) {
            type = User::typeUserRequest;
        } else {
            type = User::typeUserForeign;
        }
    }

    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE users SET type=:type WHERE id=:id");
    query.bindValue(":id", userId);
    query.bindValue(":type", type);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when trying to update contact=" << isContact << "for userId" << userId << ":" << query.lastError() << query.lastQuery();
        return false;
    }

    if (query.numRowsAffected() > 0) {

        ContactItem contactItem = getUser(userId);
        QList<ContactItem> contacts;
        contacts << contactItem;
        Q_EMIT contactsAdded(contacts);

    } else {

        // ask for user in the server if not found in db
        type = isContact ? InputUser::typeInputUserContact : InputUser::typeInputUserForeign;
        InputUser inputUser((InputUser::InputUserType)type);
        inputUser.setUserId(userId);
        QList<InputUser> users;
        users << inputUser;
        TelegramService *service = static_cast<TelegramService *>(mServiceRef);
        service->usersGetUsers(users);
    }
    return true;
}

bool Data::deleteUsers() {

    QSqlQuery query(mDbManager.database());

    query.prepare("DELETE FROM users");
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }
    return true;
}

bool Data::hasChangedUsername(const User &user) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT firstName, lastName FROM users WHERE id=:id");
    query.bindValue(":id", user.id());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not check if name has changed for user" << user.id()
                                    << query.lastError() << query.lastQuery();
        return false;
    }

    if (query.next()) {
        QString firstName = query.value("firstName").toString();
        QString lastName = query.value("lastName").toString();

        return user.firstName() != firstName || user.lastName() != lastName;
    }

    return false;
}

// CHATs operations
DialogItem Data::getChat(qint32 id) {
    DialogItem dialog;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT title, participantsCount, thumbnail FROM chats WHERE id=:id");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "queried chat with id" << id << "has not been found";
        return dialog;
    }
    query.next();

    dialog.setId(id);
    dialog.setName(query.value("title").toString());
    dialog.setMembers(query.value("participantsCount").toInt());
    dialog.setThumbnail(query.value("thumbnail").toString());
    return dialog;
}

bool Data::existsChat(qint32 chatId) {
    return exists(chatId, false);
}

bool Data::insertOrUpdateChat(const Chat &chat) {
    if (!updateChat(chat)) {
        return insertChat(chat);
    }
    Q_EMIT chatUpdated(chat.id());
    return true;
}

void bindChatValuesForQuery(QSqlQuery &query, const Chat &chat, qint32 thumbnailRowId, qint32 bigPhotoRowId) {
    query.bindValue(":id", chat.id());
    query.bindValue(":version", chat.version());
    query.bindValue(":title", chat.title());
    query.bindValue(":participantsCount", chat.participantsCount());
    query.bindValue(":thumbnail", thumbnailRowId);
    query.bindValue(":bigPhoto", bigPhotoRowId);
    query.bindValue(":date", chat.date());
    query.bindValue(":left", chat.left());
}

bool Data::insertChat(const Chat &chat) {

    mDbManager.beginTransaction();

    FileLocation thumbnailServerLocation = chat.photo().photoSmall();
    qint32 thumbnailRowId = insertFileLocation(thumbnailServerLocation);
    if (thumbnailRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert thumbnail server location related to chat" << chat.id();
        mDbManager.rollbackTransaction();
        return false;
    }

    FileLocation bigPhotoServerLocation = chat.photo().photoBig();
    qint32 bigPhotoRowId = insertFileLocation(bigPhotoServerLocation);
    if (bigPhotoRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert big photo server location related to chat" << chat.id();
        mDbManager.rollbackTransaction();
        return false;
    }

    QSqlQuery query(mDbManager.database());
    query.prepare("INSERT INTO chats (id, version, title, participantsCount, thumbnail, bigPhoto, date, left) "
                  "VALUES (:id, :version, :title, :participantsCount, :thumbnail, :bigPhoto, :date, :left)");
    bindChatValuesForQuery(query, chat, thumbnailRowId, bigPhotoRowId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    if (thumbnailRowId > 0) {
        getProfilePhotoFileFromServer(thumbnailServerLocation, chat.id(), false, true);
    }

    mDbManager.finishTransaction();
    return query.numRowsAffected() > 0;
}

bool Data::updateChat(const Chat &chat) {

    if (!existsChat(chat.id())) {
        return false;
    }

    qint32 thumbnailRowId = insertOrUpdateChatThumbnailFileLocation(chat);
    if (thumbnailRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update thumbnail for chat" << chat.id();
        return false;
    }

    qint32 bigPhotoRowId = insertOrUpdateChatBigPhotoFileLocation(chat);
    if (bigPhotoRowId < 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update big photo for chat" << chat.id();
        if (thumbnailRowId > 0) {
            deleteFileLocation(thumbnailRowId);
        }
        return false;
    }

    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE chats SET version=:version, title=:title, "
                  "participantsCount=:participantsCount, thumbnail=:thumbnail, bigPhoto=:bigPhoto, date=:date, left=:left "
                  "WHERE id=:id");
    bindChatValuesForQuery(query, chat, thumbnailRowId, bigPhotoRowId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

bool Data::updateChatTitle(qint32 chatId, const QString &newTitle) {
    QSqlQuery query(mDbManager.database());
    query.prepare("UPDATE chats SET title=:title WHERE id=:id");
    query.bindValue(":id", chatId);
    query.bindValue(":title", newTitle);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating title for chat" << chatId << query.lastError() << query.lastQuery();
        return false;
    }
    return query.numRowsAffected() > 0;
}

void Data::updateChatParticipants(const ChatParticipants &chatParticipants) {
    QSqlQuery query(mDbManager.database());
    query.prepare(QString("DELETE FROM participants WHERE chatId=:chatId"));
    query.bindValue(":chatId", chatParticipants.chatId());
    if (!query.exec()) {
        throw "Error updating chat participants (on removal)";
    }

    query.prepare(QString("INSERT INTO participants "
                  "(chatId, userId, inviterId, date, isAdmin, version) "
                  "VALUES (:chatId, :userId, :inviterId, :date, :isAdmin, :version)"));
    Q_FOREACH (const ChatParticipant &member, chatParticipants.participants()) {
        query.bindValue(":chatId", chatParticipants.chatId());
        query.bindValue(":userId", member.userId());
        query.bindValue(":inviterId", member.inviterId());
        query.bindValue(":date", member.date());
        query.bindValue(":isAdmin", chatParticipants.adminId() == member.userId());
        query.bindValue(":version", chatParticipants.version());
        if (!query.exec()) {
            throw "Error updating chat participants (on insert)";
        }
    }
}

bool Data::deleteChat(qint32 chatId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("DELETE FROM chats WHERE id=:id");
    query.bindValue(":id", chatId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
        return false;
    }
    return true;
}

bool Data::deleteChatPhoto(qint32 chatId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE chats SET thumbnail=0 WHERE id=:chatId");
    query.bindValue(":chatId", chatId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error deleting thumbnail for chat" << chatId << query.lastError() << query.lastQuery();
        return false;
    }
    return query.numRowsAffected() > 0;
}

qint32 Data::insertOrUpdateUserThumbnail(qint32 id, const QString &localPath, const FileLocation &serverLocation) {
    return insertOrUpdateProfilePhoto(id, true, true, localPath, serverLocation);
}

qint32 Data::insertOrUpdateChatThumbnail(qint32 id, const QString &localPath, const FileLocation &serverLocation) {
    return insertOrUpdateProfilePhoto(id, false, true, localPath, serverLocation);
}

qint32 Data::insertOrUpdateUserBigPhoto(qint32 id, const QString &localPath, const FileLocation &serverLocation) {
    return insertOrUpdateProfilePhoto(id, true, false, localPath, serverLocation);
}

qint32 Data::insertOrUpdateChatBigPhoto(qint32 id, const QString &localPath, const FileLocation &serverLocation) {
    return insertOrUpdateProfilePhoto(id, false, false, localPath, serverLocation);
}

qint32 Data::insertOrUpdateProfilePhoto(qint32 id, bool isUser, bool isThumbnail, const QString &localPath, const FileLocation &serverLocation) {

    QString queryString("UPDATE fileLocations SET localPath=:localPath, dcId=:dcId, localId=:localId, secret=:secret, volumeId=:volumeId "
                        "WHERE rowid IN (SELECT %1 FROM %2 WHERE id=:id)");

    QString table = isUser ? "users" : "chats";
    QString field = isThumbnail ? "thumbnail" : "bigPhoto";
    queryString = queryString.arg(field, table);

    mDbManager.beginTransaction();
    QSqlQuery query(mDbManager.database());
    query.prepare(queryString);
    query.bindValue(":localPath", localPath);
    query.bindValue(":dcId", serverLocation.dcId());
    query.bindValue(":localId", serverLocation.localId());
    query.bindValue(":secret", serverLocation.secret());
    query.bindValue(":volumeId", serverLocation.volumeId());
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update fileLocations table:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    if (query.numRowsAffected() == 0) {

        query.prepare("INSERT INTO fileLocations (localPath, dcId, localId, secret, volumeId) "
                      "VALUES (:localPath, :dcId, :localId, :secret, :volumeId)");
        query.bindValue(":localPath", localPath);
        query.bindValue(":dcId", serverLocation.dcId());
        query.bindValue(":localId", serverLocation.localId());
        query.bindValue(":secret", serverLocation.secret());
        query.bindValue(":volumeId", serverLocation.volumeId());
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not insert fileLocations new register:" << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }

        qint32 lastRowId = query.lastInsertId().toInt();

        queryString = QString("UPDATE %1 SET %2=:rowid WHERE id=:id").arg(table, field);
        query.prepare(queryString);
        query.bindValue(":rowid", lastRowId);
        query.bindValue(":id", id);
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Could not update" << field << "for" << table << "register"
                                        << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }
    }

    mDbManager.finishTransaction();
    return query.numRowsAffected();
}

QString Data::getThumbnail(qint32 id, bool isUser) {

    QString localPath;

    QSqlQuery query(mDbManager.database());

    QString queryString("SELECT localPath FROM fileLocations,%1 WHERE %1.thumbnail=fileLocations.rowid AND %1.id=:id");
    if (isUser) {
        queryString = queryString.arg("users");
    } else {
        queryString = queryString.arg("chats");
    }
    query.prepare(queryString);
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error:" << query.lastError() << query.lastQuery();
    }

    if (query.next()) {
        localPath = query.value("localPath").toString();
    }

    return localPath;
}

qint32 Data::updatePhoto(qint32 id, qint32 oldPhotoId, const Photo &photo) {

    qint32 affectedSizes = 0;

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());

    //don't add media photos register if photo exists (it could exists due to a message forward from another dialog, for instance)
    query.prepare("SELECT id FROM mediaPhotos WHERE id=:id");
    query.bindValue(":id", photo.id());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when checking if media photo exists:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    if (query.next()) {
        if (deletePhoto(oldPhotoId) == -1) {
            mDbManager.rollbackTransaction();
            return -1;
        }
    } else {
        // try to update media photos if present previously
        // (if this messages was outgoing, media id would be set temporally to message id and file location to local path)
        query.prepare("UPDATE mediaPhotos "
                      "SET id=:id, accessHash=:accessHash, userId=:userId, date=:date, caption=:caption, geoLat=:geoLat, geoLong=:geoLong "
                      "WHERE id=:oldPhotoId");
        query.bindValue(":id", photo.id());
        query.bindValue(":accessHash", photo.accessHash());
        query.bindValue(":userId", photo.userId());
        query.bindValue(":date", photo.date());
        query.bindValue(":caption", photo.caption());
        query.bindValue(":geoLat", photo.geo().lat());
        query.bindValue(":geoLong", photo.geo().longitude());
        query.bindValue(":oldPhotoId", oldPhotoId);
        if (!query.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error when updating media photos previous register:" << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }

        if (query.numRowsAffected() > 0) {
            // if affected rows, there were a previously stored media related to this message and its photo sizes must be updated
            if (oldPhotoId != photo.id()) {
                query.prepare("UPDATE photoSizes SET photoId=:photoId WHERE photoId=:oldPhotoId");
                query.bindValue(":photoId", photo.id());
                query.bindValue(":oldPhotoId", oldPhotoId);
                if (!query.exec()) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Error when updating local photo size previous register:" << query.lastError() << query.lastQuery();
                    mDbManager.rollbackTransaction();
                    return -1;
                }

                affectedSizes = query.numRowsAffected();

                // retrieved from server data updates photo sizes (or insert if not exists)
                PhotoSize thumbnail = selectThumbnailSize(photo.sizes());
                if (insertOrUpdatePhotoSize(oldPhotoId, photo.id(), thumbnail) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert or update thumbnail for updated media photo";
                    mDbManager.rollbackTransaction();
                    return -1;
                }

                PhotoSize bigPhoto = selectBigPhotoSize(photo.sizes());
                if (insertOrUpdatePhotoSize(oldPhotoId, photo.id(), bigPhoto) <= 0) {
                    qCCritical(TG_PLUGIN_LOGIC) << "Could not insert or update photo sizes for updated media photo";
                    mDbManager.rollbackTransaction();
                    return -1;
                }
            }

        } else {
            // if not affected any row, was no photo previously stored related to this message, so add it
            affectedSizes = insertPhoto(photo);
            if (affectedSizes <= 0) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not insert not existant photo";
                mDbManager.rollbackTransaction();
                return -1;
            }

            downloadThumbnail(id, photo, IncomingFile::MessagePhoto);
            downloadPhoto(id, photo, IncomingFile::MessagePhoto);
        }
    }

    mDbManager.finishTransaction();

    return affectedSizes;
}


PhotoSize Data::selectThumbnailSize(const QList<PhotoSize> &photoSizes) {
    return selectSize(photoSizes, true);
}

PhotoSize Data::selectBigPhotoSize(const QList<PhotoSize> &photoSizes) {
    return selectSize(photoSizes, false);
}

PhotoSize Data::selectSize(const QList<PhotoSize> &photoSizes, bool thumbnail) {

    PhotoSize selectedPhotoSize;
    qint8 currentPriority = 0;

    Q_FOREACH (PhotoSize photoSize, photoSizes) {

        qint8 thisSizePriority;
        if (thumbnail) {
            thisSizePriority = mPhotoSmallPriorities.value(photoSize.type(), 0);
        } else {
            thisSizePriority = mPhotoBigPriorities.value(photoSize.type(), 0);
        }

        if ((!currentPriority) || (thisSizePriority && thisSizePriority < currentPriority)) {
            currentPriority = thisSizePriority;
            selectedPhotoSize = photoSize;

            // in case we found the max priority exit from loop
            if (currentPriority == 1) {
                break;
            }
        }
    }

    return selectedPhotoSize;
}

qint32 Data::existsPhoto(qint64 photoId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT id FROM mediaPhotos WHERE id=:id");
    query.bindValue(":id", photoId);
    if (!query.exec()) {
        qCWarning(TG_PLUGIN_LOGIC) << "Could not check if media photo exists:" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.next() ? 1 : 0;
}

qint32 Data::insertPhoto(const Photo &photo) {

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());

    query.prepare("INSERT INTO mediaPhotos (id, accessHash, userId, date, caption, geoLat, geoLong) "
                  "VALUES (:id, :accessHash, :userId, :date, :caption, :geoLat, :geoLong)");

    query.bindValue(":id", photo.id());
    query.bindValue(":accessHash", photo.accessHash());
    query.bindValue(":userId", photo.userId());
    query.bindValue(":date", photo.date());
    query.bindValue(":caption", photo.caption());
    query.bindValue(":geoLat", photo.geo().lat());
    query.bindValue(":geoLong", photo.geo().longitude());
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when inserting media photo:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }


    PhotoSize thumbnail = selectThumbnailSize(photo.sizes());
    PhotoSize photoBig = selectBigPhotoSize(photo.sizes());

    // insert photo small
    qint32 fileLocationId = insertFileLocation(thumbnail.location());
    if (fileLocationId < 0) {
        mDbManager.rollbackTransaction();
        return -1;
    }

    qint32 result = insertPhotoSize(photo.id(), thumbnail, fileLocationId);
    if (result <= 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new photo size";
        mDbManager.rollbackTransaction();
        return -1;
    }

    // insert photo big
    fileLocationId = insertFileLocation(photoBig.location());
    if (fileLocationId < 0) {
        mDbManager.rollbackTransaction();
        return -1;
    }

    result = insertPhotoSize(photo.id(), photoBig, fileLocationId);
    if (result <= 0) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not insert new photo size";
        mDbManager.rollbackTransaction();
        return -1;
    }

    mDbManager.finishTransaction();

    return 1;
}

qint32 Data::deletePhoto(qint32 photoId) {

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());

    // delete register for media photos
    query.prepare("DELETE FROM mediaPhotos WHERE id=:photoId");
    query.bindValue(":photoId", photoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when deleting not needed media photos register:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    query.prepare("SELECT fileLocationId FROM photoSizes WHERE photoId=:photoId");
    query.bindValue(":photoId", photoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when getting photo locations for photo to delete:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    QSqlQuery subquery(mDbManager.database());
    subquery.prepare("DELETE FROM FileLocations WHERE rowid=:fileLocationId");

    while (query.next()) {
        qint32 fileLocationId = query.value("fileLocationId").toInt();

        subquery.bindValue(":fileLocationId", fileLocationId);
        if (!subquery.exec()) {
            qCCritical(TG_PLUGIN_LOGIC) << "Error when deleting photo locations for photo to delete:" << query.lastError() << query.lastQuery();
            mDbManager.rollbackTransaction();
            return -1;
        }
    }

    query.prepare(("DELETE FROM photoSizes WHERE photoId=:photoId"));
    query.bindValue(":photoId", photoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when deleting photo sizes:" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    mDbManager.finishTransaction();
}

bool Data::hasThumbnailDownloaded(qint32 photoId) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT localPath FROM fileLocations WHERE rowid IN (SELECT fileLocationId FROM photoSizes WHERE photoId=:photoId)");
    query.bindValue(":photoId", photoId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error when checking if needed to download thumbnail for photoId:" << photoId <<
                                       query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return -1;
    }

    while (query.next()) {
        QString localPath = query.value("localPath").toString();
        if (!localPath.isEmpty()) {
            return false;
        }
    }

    return true;
}

void Data::downloadThumbnail(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType) {
    selectSizeAndDownloadPhoto(id, photo, incomingFileType, true);
}

void Data::downloadPhoto(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType) {
    selectSizeAndDownloadPhoto(id, photo, incomingFileType, false);
}

void Data::selectSizeAndDownloadPhoto(qint32 id, const Photo &photo, IncomingFile::IncomingFileType incomingFileType, bool thumbnail) {

    PhotoSize photoSizeToDownload;
    qint32 fileLocationIdToDownload = 0;

    qint8 currentPriority = 0;

    Q_FOREACH (PhotoSize photoSize, photo.sizes()) {

        qint8 thisSizePriority;
        if (thumbnail) {
            thisSizePriority = mPhotoSmallPriorities.value(photoSize.type(), 0);
        } else {
            thisSizePriority = mPhotoBigPriorities.value(photoSize.type(), 0);
        }

        if ((!currentPriority) || (thisSizePriority && thisSizePriority < currentPriority)) {
            currentPriority = thisSizePriority;
            photoSizeToDownload = photoSize;
            fileLocationIdToDownload = getPhotoSizeFileLocationId(photo.id(), photoSize.type());
        }
    }

    if (fileLocationIdToDownload) {
        downloadPhotoSize(id, photo.id(), photo.date(), incomingFileType, thumbnail, photoSizeToDownload, fileLocationIdToDownload);
    }
}

void Data::downloadPhotoSize(qint32 id, qint64 mediaId, qint32 date, IncomingFile::IncomingFileType incomingFileType, bool thumbnail,
                       const PhotoSize &photoSize, qint32 fileLocationId) {

    if (fileLocationId && photoSize.classType() != PhotoSize::typePhotoSizeEmpty) {

        downloadPhotoFile(id, mediaId, date, incomingFileType, thumbnail,
                          photoSize.type(),
                          photoSize.size(),
                          photoSize.location().dcId(),
                          photoSize.location().localId(),
                          photoSize.location().secret(),
                          photoSize.location().volumeId(),
                          fileLocationId);
    }
}

void Data::downloadPhotoFile(qint32 id, qint64 mediaId, qint32 date, IncomingFile::IncomingFileType incomingFileType, bool thumbnail,
                             const QString &type, qint32 size, qint32 dcId, qint32 localId, qint64 secret, qint64 volumeId,
                             qint32 fileLocationId) {
    // create object with the data of the file to be downloaded
    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    QString downloadsPath = service->config().downloadsPath();
    IncomingFile *incomingFile = new IncomingFile(downloadsPath);
    incomingFile->setId(id);
    incomingFile->setMediaId(mediaId);
    incomingFile->setIncomingFileType(incomingFileType);
    incomingFile->setMediaType(MessageMedia::typeMessageMediaPhoto);
    incomingFile->setSizeType(type);
    incomingFile->setIsThumbnail(thumbnail);
    incomingFile->setFileLocationId(fileLocationId);
    // file name is composed of <mediaId>_<creation date>_<type>.<extension>
    QString fileName = QString::number(mediaId) + "_" + QString::number(date) + "_" + type;
    incomingFile->setFilename(fileName);
    incomingFile->setExpectedSize(size);

    // create input needed data to download the file from telegram servers
    InputFileLocation inputFileLocation(InputFileLocation::typeInputFileLocation);
    inputFileLocation.setLocalId(localId);
    inputFileLocation.setSecret(secret);
    inputFileLocation.setVolumeId(volumeId);
    qint64 fileId = service->uploadGetFile(inputFileLocation, size, dcId);

    // correlate api provided fileId with the incoming file created object
    mFiles.insert(fileId, incomingFile);
    mActiveDownloads.insert(id, fileId);

    Q_EMIT downloadingUpdated(id, true);
}

qint32 Data::updateFileLocation(qint32 id, const QString &localPath) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE fileLocations SET localPath=:localPath WHERE rowid=:id");
    query.bindValue(":localPath", localPath);
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error updating" << localPath << "file location with id" <<
                                       id << ":" << query.lastError() << query.lastQuery();
        return -1;
    }

    return query.numRowsAffected();
}

PhotoSize Data::getPhotoSize(qint32 id) {

    PhotoSize photoSize(PhotoSize::typePhotoSizeEmpty);

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT type, width, height, size FROM photoSizes WHERE fileLocationId=:id");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error getting photo size related to fileLocation" << id << ":" <<
                                       query.lastError() << query.lastQuery();
        return photoSize;
    }

    if (query.next()) {
        QString type = query.value("type").toString();
        qint32 width = query.value("width").toInt();
        qint32 height = query.value("height").toInt();
        qint32 size = query.value("size").toInt();
        photoSize.setClassType(PhotoSize::typePhotoSize);
        photoSize.setType(type);
        photoSize.setW(width);
        photoSize.setH(height);
        photoSize.setSize(size);
    }
    return photoSize;
}

qint32 Data::getPhotoMsgId(qint32 id) {

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT messages.id FROM messages WHERE mediaId IN (SELECT mediaPhotos.id FROM "
                  "mediaPhotos WHERE mediaPhotos.id IN (SELECT photoId FROM photoSizes WHERE fileLocationId=:id))");
    query.bindValue(":id", id);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Error getting msgId related to a photo file with id" << id << ":" <<
                                       query.lastError() << query.lastQuery();
        return -1;
    }

    if (query.next()) {
        return query.value(0).toInt();
    }

    return -1;
}


qint64 Data::getDownloadingFileId(qint32 id) {
    return mActiveDownloads.value(id);
}

void Data::onUploadCancelFileAnswer(qint64 fileId, bool cancelled) {
    if (cancelled) {
        IncomingFile * incomingFile = mFiles.take(fileId);
        if (incomingFile) {
            // remove already downloaded parts
            QFile &qFile = incomingFile->file();
            if (!qFile.remove()) {
                qCCritical(TG_PLUGIN_LOGIC) << "Could not remove already downloaded file parts when cancel operation requested";
            }

            qint32 id = incomingFile->id();
            mActiveDownloads.take(id);
        }
    }
}

void Data::setMutingInProgress(qint64 requestId, qint32 peerId, bool muting) {
    if (muting) {
        mMuteInProgress.insert(requestId, peerId);
    } else {
        mUnmuteInProgress.insert(requestId, peerId);
    }
}

void Data::onAccountUpdateNotifySettingsAnswer(qint64 requestId, bool ok) {
    bool mute = true;
    qint32 peerId = mMuteInProgress.take(requestId);
    if (peerId == 0) {
        mute = false;
        peerId = mUnmuteInProgress.take(requestId);
    }
    if (ok && peerId != 0) {
        muteDialog(peerId, mute);
        Q_EMIT dialogMuted(peerId, mute);
    }
}

bool Data::muteDialog(qint32 dialogId, bool muted) {
    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE dialogs SET muted=:muted WHERE id=:id");
    query.bindValue(":id", dialogId);
    query.bindValue(":muted", muted);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update mute to" << muted
                << "for dialog" << dialogId
                <<  query.lastError() << query.lastQuery();
        return false;
    }
    return query.numRowsAffected() > 0;
}

bool Data::muteDialogs(NotifyPeer::NotifyPeerType peerType, bool muted) {
    QSqlQuery query(mDbManager.database());

    QString condition;
    switch (peerType) {
    case NotifyPeer::typeNotifyUsers:
        condition = " WHERE isChat=0";
        break;
    case NotifyPeer::typeNotifyChats:
        condition = " WHERE isChat=1";
        break;
    case NotifyPeer::typeNotifyAll:
        break;
    default:
        // for individual peers there is another method
        return false;
    }

    QString sql = QString("UPDATE dialogs SET muted=:muted").arg(condition);
    query.prepare(sql);
    query.bindValue(":muted", muted);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update mute to" << muted
                << "for dialogs" << query.lastError() << query.lastQuery();
        return false;
    }

    return query.numRowsAffected() > 0;
}

void Data::onMessagesCreateEncryptedChatAnswer(qint32 chatId, qint32 date, qint32 peerId, qint64) {
    if (insertSecretChat(chatId, date, peerId, SecretChat::Init)) {
        DialogItem dialog = getDialog(chatId);
        QList<DialogItem> dialogs;
        dialogs << dialog;
        Q_EMIT dialogsAdded(dialogs);
    }
}

void Data::onMessagesEncryptedChatRequested(qint32 chatId, qint32 date, qint32 creatorId, qint64 accessHash) {
    TelegramService *service = static_cast<TelegramService *>(mServiceRef);
    service->acceptSecretChat(chatId);
}

void Data::onMessagesEncryptedChatCreated(qint32 chatId, qint32 date, qint32 peerId, qint64 accessHash) {
    if (insertOrUpdateSecretChat(chatId, date, peerId, SecretChat::Accepted)) {
        DialogItem dialog = getDialog(chatId);
        QList<DialogItem> dialogs;
        dialogs << dialog;
        Q_EMIT dialogsAdded(dialogs);
        Q_EMIT secretChatCreated(chatId);
    }
}

void Data::onMessagesEncryptedChatDiscarded(qint32 chatId) {
    if (updateSecretChat(chatId, SecretChat::Cancelled)) {
        DialogItem dialog = getDialog(chatId);
        QList<DialogItem> dialogs;
        dialogs << dialog;
        Q_EMIT dialogsAdded(dialogs);
        Q_EMIT secretChatDiscarded(chatId);
    }
}

bool Data::insertOrUpdateSecretChat(qint32 chatId, qint32 date, qint32 peerId, qint32 state) {
    if (!updateSecretChat(chatId, state)) {
        return insertSecretChat(chatId, date, peerId, state);
    }
    return true;
}

bool Data::insertSecretChat(qint32 chatId, qint32 date, qint32 peerId, qint32 state) {

    mDbManager.beginTransaction();

    QSqlQuery query(mDbManager.database());
    query.prepare("INSERT INTO secretChats (id, date, peerId, state) VALUES (:chatId, :date, :peerId, :state)");
    query.bindValue(":chatId", chatId);
    query.bindValue(":date", date);
    query.bindValue(":peerId", peerId);
    query.bindValue(":state", state);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not add new secret chat" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    query.prepare("INSERT INTO dialogs (id, isChat, unreadCount, muted, date, isSecret) "
                  "VALUES (:chatId, 0, 0, 0, :date, 1)");
    query.bindValue(":chatId", chatId);
    query.bindValue(":date", date);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not add new secret chat related dialog" << query.lastError() << query.lastQuery();
        mDbManager.rollbackTransaction();
        return false;
    }

    mDbManager.finishTransaction();
    return true;
}

bool Data::updateSecretChat(qint32 chatId, qint32 state) {

    QSqlQuery query(mDbManager.database());

    query.prepare("UPDATE secretChats SET state=:state WHERE id=:chatId");
    query.bindValue(":chatId", chatId);
    query.bindValue(":state", state);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not update secret chat state" << query.lastError() << query.lastQuery();
        return false;
    }
    return query.numRowsAffected() > 0;
}

DialogItem Data::getSecretChat(qint32 chatId) {

    DialogItem dialog;

    QSqlQuery query(mDbManager.database());

    query.prepare("SELECT peerId, date, state FROM secretChats WHERE id=:chatId");
    query.bindValue(":chatId", chatId);
    if (!query.exec()) {
        qCCritical(TG_PLUGIN_LOGIC) << "Could not get secret chat with id" << chatId
                                    << query.lastError() << query.lastQuery();
        return dialog;
    }

    if (query.next()) {
        dialog.setId(chatId);
        dialog.setPeerId(query.value("peerId").toInt());
        dialog.setDate(query.value("date").toInt());
        dialog.setState(query.value("state").toInt());
    }

    return dialog;
}

void Data::surfaceUnreadCount() {
    QSqlQuery query;
    query.prepare(QString("SELECT SUM(unreadCount) AS unreadCount FROM dialogsView"));

    if (query.exec() && query.next()) {
        qint32 unreadCount = query.value("unreadCount").toInt();
        if (mLastUnreadCount != unreadCount) {
            qCDebug(TG_PLUGIN_LOGIC) << "Updating badge" << unreadCount;
            mLastUnreadCount = unreadCount;
            Q_EMIT unreadCountChanged(unreadCount);
        }
    }
}
